#include "businterface.h"

#include <ds/log.h>

#include "model/apartment.h"
#include "vdc-element-reader.h"
#include "vdc-messages.pb.h"

DS_REFLECT_DEFINE_STRUCT((dss), DSMeterSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), GroupSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), ClusterSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceBinaryInputSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceStateSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), CircuitPowerStateSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceSensorSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceSensorValue_t);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceSpec);
DS_REFLECT_DEFINE_STRUCT((dss), DSMeterHash_t);
DS_REFLECT_DEFINE_STRUCT((dss), MeteringConfigurationSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), ZoneHeatingConfigSpec);
DS_REFLECT_DEFINE_STRUCT((dss), ZoneHeatingInternalsSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), ZoneHeatingStateSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), ZoneHeatingOperationModeSpec_t);
DS_REFLECT_DEFINE_STRUCT((dss), Caller);

namespace dss {

static ds::LogChannel dsLogChannel("busInterface");

std::ostream& operator<<(std::ostream& ostream, const StructureModifyingBusInterface& x) {
  return ostream << x.name();
}

vdcapi::Message StructureQueryBusInterface::getVdcProperty(const dsuid_t& vdcId, const dsuid_t& targetId,
                                                           const std::vector<std::string>& keys) {
  char dsuid_string[DSUID_STR_LEN];
  dsuid_to_string(&targetId, dsuid_string);

  vdcapi::Message message;
  message.set_type(vdcapi::VDSM_REQUEST_GET_PROPERTY);
  vdcapi::vdsm_RequestGetProperty* getPropertyRequest = message.mutable_vdsm_request_get_property();
  getPropertyRequest->set_dsuid(dsuid_string);
  auto&& query = *getPropertyRequest->mutable_query();
  for (auto&& key : keys) {
    query.Add()->set_name(key);
  }

  uint8_t arrayOut[REQUEST_LEN];
  if (!message.SerializeToArray(arrayOut, sizeof(arrayOut))) {
    throw std::runtime_error("SerializeToArray failed");
  }

  uint8_t arrayIn[RESPONSE_LEN];
  uint16_t arrayInSize;
  protobufMessageRequest(vdcId, message.ByteSizeLong(), arrayOut, &arrayInSize, arrayIn);

  if (!message.ParseFromArray(arrayIn, arrayInSize)) {
    throw std::runtime_error("ParseFromArray failed");
  }
  if (message.type() != vdcapi::VDC_RESPONSE_GET_PROPERTY) {
    throw std::runtime_error("Invalid vdc response");
  }
  return message;
}

struct StructureQueryBusInterface::VdcInputDesc StructureQueryBusInterface::getVdcInputDesc(dsuid_t _vdsm,
                                                                                            dsuid_t _device) {
  struct StructureQueryBusInterface::VdcInputDesc iDesc;
  vdcapi::Message message = getVdcProperty(_vdsm, _device, {"sensorDescriptions", "binaryInputDescriptions"});

  DS_DEBUG(message);
  vdcapi::vdc_ResponseGetProperty response = message.vdc_response_get_property();
  VdcElementReader reader(message.vdc_response_get_property().properties());

  auto&& sensorList = iDesc.m_sensorInputs;
  auto&& sensorDescriptionsReader = reader["sensorDescriptions"];
  for (const auto& sensorReader : sensorDescriptionsReader) {
    SensorDesc sDesc;
    if (sensorReader["dsIndex"].isValid()) {
      // vdc-api v3 format: "techn. name" : { dsIndex: 7 }
      sDesc.sensorName = sensorReader.getName();
      sDesc.index = sensorReader["dsIndex"].getValueAsInt(255);
    } else {
      // legacy format: "7" : { name: "techn. name" }
      sDesc.sensorName = sensorReader["name"].getValueAsString();
      try {
        sDesc.index = strToInt(sensorReader.getName());
      } catch (std::exception&) {
        sDesc.index = 255;
      }
    }
    sDesc.sensorType =
        static_cast<SensorType>(sensorReader["sensorType"].getValueAsInt(static_cast<int>(SensorType::UnknownType)));
    sDesc.sensorUsage = sensorReader["sensorUsage"].getValueAsInt();
    sDesc.min = sensorReader["min"].getValueAsDouble();
    sDesc.max = sensorReader["max"].getValueAsDouble();
    sDesc.resolution = sensorReader["resolution"].getValueAsDouble();
    sDesc.updateInterval = sensorReader["updateInterval"].getValueAsDouble();

    DS_DEBUG(_device, sDesc.sensorName, sDesc.index, sDesc.sensorType);
    sensorList[sDesc.sensorName] = sDesc;
  }

  auto&& binputList = iDesc.m_binaryInputs;
  auto&& binputDescriptionsReader = reader["binaryInputDescriptions"];
  for (const auto& inputReader : binputDescriptionsReader) {
    BinaryInputDesc bDesc;
    if (inputReader["dsIndex"].isValid()) {
      // vdc-api v3 format: "techn. name" : { dsIndex: 7 }
      bDesc.inputName = inputReader.getName();
      bDesc.dsIndex = inputReader["dsIndex"].getValueAsInt(255);
    } else {
      // legacy format: "7" : { name: "techn. name" }
      bDesc.inputName = inputReader["name"].getValueAsString();
      try {
        bDesc.dsIndex = strToInt(inputReader.getName());
      } catch (std::exception&) {
        bDesc.dsIndex = 255;
      }
    }
    bDesc.sensorFunction = static_cast<BinaryInputType>(
        inputReader["sensorFunction"].getValueAsInt(static_cast<int>(BinaryInputType::AppMode)));
    bDesc.updateInterval = inputReader["updateInterval"].getValueAsDouble();

    DS_DEBUG(_device, bDesc.inputName, bDesc.dsIndex, bDesc.sensorFunction);
    binputList[bDesc.inputName] = bDesc;
  }

  return iDesc;
}

} // namespace dss
