/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "ds485types.h"
#include "hal/hardware-abstraction.h"
#include "model/set.h"

namespace dss {

class Apartment;

struct CacheScenarioState {
  int sceneId;
  int zoneId;
  int groupId;
  dsModel::ScenarioStatusStatus status;
  ds::DateTime lastChanged;
};

struct ScenarioCache {
  std::map<std::string, CacheScenarioState> scenarios;
};

struct ParameterCacheStatus {
  dsModel::ParameterCacheAttributes cache{};
  std::vector<hal::DeviceCacheState> devices{};
};

struct ZoneStateCache {
  std::optional<int> zoneState;
  std::optional<int> area1State;
  std::optional<int> area2State;
  std::optional<int> area3State;
  std::optional<int> area4State;
};

struct DsmZoneStateCache {
  //zoneId, groupId
  std::map<std::pair<int, int>, ZoneStateCache> dsmZoneStateCache;
};

class ApartmentProxy {
public:
  virtual ~ApartmentProxy() = default;

  virtual void callScene(const Dsuid& device, int sceneId, bool forced, ds::DateTime time) const = 0;
  virtual void callScene(int zoneId, int groupId, int sceneId, bool forced, ds::DateTime time) = 0;
  virtual void saveScene(const Dsuid& device, int sceneId) const = 0;
  virtual void callSceneMin(const Dsuid& device, int sceneId, ds::DateTime time) const = 0;
  virtual void callSceneMin(int zoneId, int groupId, int sceneId, ds::DateTime time) const = 0;
  virtual void undoScene(const Dsuid& device, int sceneId, ds::DateTime time) const = 0;
  virtual void undoScene(int zoneId, int groupId, int sceneId, ds::DateTime time) = 0;
  virtual void setOutputValue(const Dsuid& device, uint8_t value, ds::DateTime time) const = 0;
  virtual void setOutputValue(int zoneId, int groupId, uint8_t value, ds::DateTime time) const = 0;
  virtual void setOutputValue(const Dsuid& device, ChannelType channel, double value, ds::DateTime time) const = 0;
  virtual void setSceneLocalPrio(int zoneId, int groupId, int sceneId) const = 0;
  virtual void setOperationLock(const Dsuid& device, bool lock) const = 0;
  virtual void stopDimming(int zoneId, int groupId, int sceneId, ds::DateTime time) const = 0;
  virtual void stopDimming(const Dsuid& device) const = 0;
  virtual void completeMoving(const Dsuid& device) const = 0;
  virtual void stopDsmTimer() = 0;
  virtual void updateDSMState(Group& group, int zoneState, int area1State, int area2State, int area3State,
                              int area4State) const = 0;
  virtual void updateDSMState(Cluster& cluster, int state) const = 0;
  virtual void syncWithDsm(int zoneId, int groupId) const = 0;
  virtual void syncWithDsm(std::shared_ptr<Cluster> cluster) const = 0;

  virtual dsModel::ScenarioStatus zoneScenarioStatus(const dsModel::ScenarioId&) const = 0;
  virtual dsModel::DSDeviceStatus dsDeviceStatus(const dsModel::DSDeviceId& id) const = 0;
  virtual dsModel::LightsStatus lightsStatus(const dsModel::ZoneId&) const = 0;
  virtual dsModel::ShadesStatus shadesStatus(const dsModel::ZoneId&) const = 0;
  virtual dsModel::AudioStatus audioStatus(const dsModel::ZoneId&) const = 0;
  virtual dsModel::VentilationStatus ventilationUnitStatus() const = 0;
  virtual dsModel::VentilationStatus ventilationStatus(const dsModel::ZoneId& zoneId) const = 0;
  virtual dsModel::VentilationStatus recirculationStatus(const dsModel::ZoneId& zoneId) const = 0;
  virtual std::vector<dsModel::GroupStatus> groupsStatus() const = 0;
  virtual std::vector<std::string> scenarioDeviceList(Group& group, int scene) const = 0;

  virtual void updateSceneConfig(const DeviceSceneChannelId& deviceSceneChannelId, double value) const = 0;
  virtual void updateDeviceSceneChannelProperty(const DeviceSceneChannelPropertyId& deviceSceneChannelPropertyId,
                                                bool value) const = 0;
  virtual void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const = 0;
  virtual std::vector<double> getDeviceThresholds(const Dsuid& dsuid) const = 0;

  virtual dsModel::Ventilation ventilationUnitLevels() const = 0;
  virtual dsModel::Ventilation ventilationLevels(const dsModel::ZoneId& zoneId) const = 0;
  virtual dsModel::Ventilation recirculationLevels(const dsModel::ZoneId& zoneId) const = 0;

  virtual void completeProcessing() const = 0;

  virtual ParameterCacheStatus getParameterCacheStatus() const = 0;
  virtual void updateParameterCacheStatus() const = 0;
  virtual bool getEmptyClickPreventionActive() const = 0;
  virtual void setEmptyClickPreventionActive(bool active) = 0;
};

std::unique_ptr<ApartmentProxy> makeApartmentProxy(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db,
                                                   EventInterpreter* pInterpreter, const ds::Conf& conf);
std::unique_ptr<ApartmentProxy> makeNullApartmentProxy();

} /* namespace dss */

DS_JSON_DECLARE_STRUCT((dss), ScenarioCache, (scenarios));
DS_JSON_DECLARE_STRUCT((dss), CacheScenarioState, (sceneId)(zoneId)(groupId)(status)(lastChanged));
