
/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "cache-updater.h"

#include <stdio.h>

#include <regex>
#include <trompeloeil.hpp>

#include "businterface-mock.h"
#include "device-cache.h"
#include "ds-api/route-segments.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "dsm-api-mock.h"
#include "event-mock.h"
#include "hal/hal-device.h"
#include "hal/hardware-abstraction-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/modelmaintenance-mock.h"
#include "propertysystem-mock.h"
#include "propertysystem_common_paths.h"

static ds::LogChannel dsLogChannel("cacheUpdaterTest", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][cacheUpdater]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;
namespace {


void updateSceneFlags(CacheDb<DeviceCache>& db, const DeviceSceneId& adr, const SceneFlags& rflags) {
  auto devCache = db.tryLookup(adr.dsuid);
  devCache.scene.sceneTable[adr.sceneId].flags = rflags;

  auto& lflags = devCache.scene.sceneTable[adr.sceneId].flags;
  // cache-update requires all flags to be set... only dontCare affects readout
  if (!lflags.dontCare) {
    lflags.dontCare = false;
  }
  if (!lflags.ignoreLocalPrio) {
    lflags.ignoreLocalPrio = false;
  }
  if (!lflags.specialHandling) {
    lflags.specialHandling = false;
  }
  if (!lflags.blink) {
    lflags.blink = false;
  }
  if (!lflags.ledConfig) {
    lflags.ledConfig = LedConfig::ledCon0;
  }
  if (!lflags.dimTime) {
    lflags.dimTime = DimTime::dimTime2;
  }
  db.update(adr.dsuid, devCache);
}

void updateChannelDontCare(CacheDb<DeviceCache>& db, const DeviceSceneChannelId& id, bool dontCare) {
  auto devCache = db.tryLookup(id.scene.dsuid);
  devCache.scene.sceneTable[id.scene.sceneId].channels[id.type].dontCare = dontCare;
  db.update(id.scene.dsuid, devCache);
}

void updateChannelValue(CacheDb<DeviceCache>& db, const DeviceSceneChannelId& id, double value) {
  auto devCache = db.tryLookup(id.scene.dsuid);
  devCache.scene.sceneTable[id.scene.sceneId].channels[id.type].value = value;
  db.update(id.scene.dsuid, devCache);
}

void updateDeviceProperties(CacheDb<DeviceCache>& db, const Dsuid& dsuid, DeviceProperty property) {
  auto devCache = db.tryLookup(dsuid);

  switch (property) {
    case DeviceProperty::outputMode:
    case DeviceProperty::switchedThreshold:
    case DeviceProperty::switchedThresholdTwoStep0:
    case DeviceProperty::switchedThresholdTwoStep1:
    case DeviceProperty::switchedThreshold3Step0:
    case DeviceProperty::switchedThreshold3Step1:
    case DeviceProperty::switchedThreshold3Step2: devCache.property[property] = 1; break;

    case DeviceProperty::dimTime0Up:
    case DeviceProperty::dimTime0Down:
    case DeviceProperty::dimTime1Up:
    case DeviceProperty::dimTime1Down:
    case DeviceProperty::dimTime2Up:
    case DeviceProperty::dimTime2Down:
    case DeviceProperty::dimStepUp:
    case DeviceProperty::dimStepDown:
    case DeviceProperty::dimStepNum:
    case DeviceProperty::dimStepLen:
    case DeviceProperty::dimStepPause: devCache.property[property] = 1; break;

    case DeviceProperty::posTimeUp:
    case DeviceProperty::posTimeDown:
    case DeviceProperty::minDim: devCache.property[property] = 1; break;

    default: DS_FAIL_REQUIRE(property);
  }
  db.update(dsuid, devCache);
}

SceneEntry getSceneEntry(CacheDb<DeviceCache>& db, const DeviceSceneId& adr) {
  auto devCache = db.tryLookup(adr.dsuid);
  return devCache.scene.sceneTable[adr.sceneId];
}

DeviceCache getDeviceEntry(CacheDb<DeviceCache>& db, const Dsuid& dsuid) {
  return db.tryLookup(dsuid);
}

void fillCacheForDevice(HardwareAbstraction& hal, const Dsuid& dsuid) {
  hal.populateCache(dsuid); // pre-allocate entries in internal maps, vectors

  auto& db = hal.getCache();
  auto devCache = getDeviceEntry(db, dsuid);

  for (auto& propertyPair : devCache.property) {
    updateDeviceProperties(db, dsuid, propertyPair.first);
  }

  for (auto& scenePair : devCache.scene.sceneTable) {
    auto id = DeviceSceneId{.dsuid = dsuid, .sceneId = scenePair.first};
    updateSceneFlags(db, id, {.dontCare = false});

    for (auto& channelPair : scenePair.second.channels) {
      auto channelId = DeviceSceneChannelId{.scene = id, .type = channelPair.first};
      updateChannelDontCare(db, channelId, false);
      updateChannelValue(db, channelId, 40.0);
    }
    // DS_NOTICE(adr, getSceneEntry(db, id));
  }
}

class CacheUpdatePublic : public CacheUpdatePlugin {
public:
  CacheUpdatePublic(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db, EventInterpreter& interpreter,
                    std::shared_ptr<PropertySystem> propSystem, std::shared_ptr<EventQueue> eventQueue,
                    std::shared_ptr<EventRunner> eventRunner)
      : CacheUpdatePlugin(apt, hal, db, interpreter, propSystem, eventQueue, eventRunner) {}

  using CacheUpdatePlugin::getMetaInfo;
  using CacheUpdatePlugin::handleModelReady;
  using CacheUpdatePlugin::m_deviceCacheVersion;
  using CacheUpdatePlugin::scanDevice;
  using CacheUpdatePlugin::scanStart;
  using CacheUpdatePlugin::scanStep;
  using CacheUpdatePlugin::updateMetaInfo;
};

SCENARIO("A device gets added to cache", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;

  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;
  std::shared_ptr<PropertySystem> propSystem(new PropertySystem);

  EventInterpreter interpreter(nullptr);
  std::shared_ptr<EventQueue> queue(new EventQueue(&interpreter));
  std::shared_ptr<EventRunner> runner(new EventRunner(&interpreter));

  interpreter.initialize();
  interpreter.setEventQueue(queue.get());
  interpreter.setEventRunner(runner.get());
  queue->setEventRunner(runner.get());
  runner->setEventQueue(queue.get());

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  HardwareAbstractionMock hal{};
  ds::sqlite3::Database db(":memory:");
  auto deviceCache = CacheDb<DeviceCache>(&db);

  REQUIRE_CALL(hal, getCache()).LR_RETURN(deviceCache);
  CacheUpdatePublic interface(&apt, &hal, &db, interpreter, propSystem, queue, runner);

  REQUIRE_CALL(hal, getCache()).LR_RETURN(deviceCache);
  interface.handleModelReady();

  // giving up on restricting them...
  ALLOW_CALL(hal, deviceCacheable(_)).RETURN(true);
  ALLOW_CALL(hal, getCache()).LR_RETURN(deviceCache);

  auto populateCache = [&deviceCache, &db](const auto& device) {
    auto stateCache = CacheDb<DeviceInternalState>(&db);
    auto halLightDevice = hal::HalLightDevice(deviceCache, stateCache);
    auto halShadeDevice = hal::HalShadeDevice(deviceCache, stateCache);
    switch (device.getDeviceClass()) {
      case DeviceClasses_t::DEVICE_CLASS_INVALID: break;
      case DeviceClasses_t::DEVICE_CLASS_GE: halLightDevice.populateCache(device); break;
      case DeviceClasses_t::DEVICE_CLASS_GR: halShadeDevice.populateCache(device); break;
      case DeviceClasses_t::DEVICE_CLASS_BL: break;
      case DeviceClasses_t::DEVICE_CLASS_TK: break;
      case DeviceClasses_t::DEVICE_CLASS_MG: break;
      case DeviceClasses_t::DEVICE_CLASS_RT: break;
      case DeviceClasses_t::DEVICE_CLASS_GN: break;
      case DeviceClasses_t::DEVICE_CLASS_SW: break;
      case DeviceClasses_t::DEVICE_CLASS_WE: break;
      default: break;
    }
  };
  ALLOW_CALL(hal, populateCache(_)).LR_SIDE_EFFECT(populateCache(*apt.getDeviceByDSID(_1)));

  GIVEN("A new device with no scene cache") {
    auto dev1 = createDevice(apt, {.hardwareType = HardwareType::GE_KL200});
    const auto& dsuid1 = dev1->getDSID();

    dev1->setIsPresent(true);
    dev1->setIsConnected(true);

    DeviceCache empty;
    auto& db = deviceCache;
    db.update(dsuid1, empty);

    WHEN("We start the periodic updater") {
      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/brightness)")))
          .TIMES(69)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            std::regex pathRegex(R"(([0-9]+))");

            auto words_begin = std::sregex_iterator(_2.begin(), _2.end(), pathRegex);
            auto words_end = std::sregex_iterator();

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].channels[ChannelType::brightness].value = 2.1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(ANY(DeviceSceneChannelPropertyId)))
          .WITH(_1.channel.scene.dsuid == dsuid1)
          .WITH(_1.channel.type == ChannelType::brightness)
          .WITH(_1.property == ChannelProperty::dontCare)
          .TIMES(69)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[_1.channel.scene.sceneId].channels[ChannelType::brightness].dontCare = false;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
          .WITH(_1.scene.dsuid == dsuid1)
          .WITH(_1.property == SceneProperty::dontCare)
          .TIMES(69)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto id = _1;
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[id.scene.sceneId].flags.dontCare = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.ignoreLocalPrio = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.specialHandling = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.blink = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.ledConfig = LedConfig::ledCon0;
            devCache.scene.sceneTable[id.scene.sceneId].flags.dimTime = DimTime::dimTime2;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readString(dsuid1, trompeloeil::re(R"(/property/outputMode)"))).RETURN("1").LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::outputMode] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/switchedThreshold)")))
          .RETURN(1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.property[DeviceProperty::switchedThreshold] = 1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepUp)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepUp] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepDown)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepDown] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/minDim)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::minDim] = 1;
        db.update(dsuid1, devCache);
      });

      std::shared_ptr<Event> pEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent);
      std::shared_ptr<Event> pResumeEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEvent);
      auto subscription = std::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);
      interface.handleEvent(*pEvent, *subscription);

      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      THEN("The device should have all scences fetched from device") { REQUIRE(!interface.isUpdateRunning()); }
    }
  }

  GIVEN("A multi channel light device with no scene cache") {
    auto dev1 = createDevice(apt, {.hardwareType = HardwareType::GE_KL200});
    const auto& dsuid1 = dev1->getDSID();

    dev1->setIsPresent(true);
    dev1->setIsConnected(true);
    dev1->setOutputChannels({ChannelType::brightness, ChannelType::hue, ChannelType::saturation, ChannelType::colortemp,
                             ChannelType::x, ChannelType::y});

    DeviceCache empty;
    auto& db = deviceCache;
    db.update(dsuid1, empty);

    WHEN("We start the periodic updater") {
      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/[a-z]*)")))
          .TIMES(6 * 69)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channel}", _2);

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[strToInt(route.args["sceneNr"])]
                .channels[ds::toEnum<ChannelType>(route.args["channel"])]
                .value = 2.1;
            db.update(dsuid1, devCache);
          });
      REQUIRE_CALL(hal, readBool(ANY(DeviceSceneChannelPropertyId)))
          .WITH(_1.channel.scene.dsuid == dsuid1)
          .WITH(_1.property == ChannelProperty::dontCare)
          .TIMES(6 * 69)
          .RETURN(true)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[_1.channel.scene.sceneId].channels[_1.channel.type].dontCare = false;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
          .WITH(_1.scene.dsuid == dsuid1)
          .WITH(_1.property == SceneProperty::dontCare)
          .TIMES(69)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto id = _1;
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[id.scene.sceneId].flags.dontCare = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.ignoreLocalPrio = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.specialHandling = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.blink = false;
            devCache.scene.sceneTable[id.scene.sceneId].flags.ledConfig = LedConfig::ledCon0;
            devCache.scene.sceneTable[id.scene.sceneId].flags.dimTime = DimTime::dimTime2;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readString(dsuid1, trompeloeil::re(R"(/property/outputMode)"))).RETURN("1").LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::outputMode] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/switchedThreshold)")))
          .RETURN(1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.property[DeviceProperty::switchedThreshold] = 1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepUp)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepUp] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepDown)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepDown] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/minDim)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::minDim] = 1;
        db.update(dsuid1, devCache);
      });

      std::shared_ptr<Event> pEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent);
      std::shared_ptr<Event> pResumeEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEvent);
      auto subscription = std::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);
      interface.handleEvent(*pEvent, *subscription);

      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      THEN("The device should have all scences fetched from device") { REQUIRE(!interface.isUpdateRunning()); }
    }
  }

  GIVEN("An existing device in cache") {
    auto dev2 = createDevice(apt, {.hardwareType = HardwareType::GR_KL200});
    const auto& dsuid2 = dev2->getDSID();

    dev2->setIsPresent(true);
    dev2->setIsConnected(true);

    hal.populateCache(dsuid2);
    auto& db = deviceCache;
    auto mostlyThere = db.tryLookup(dsuid2);

    mostlyThere.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));
    mostlyThere.property[DeviceProperty::dimStepLen] = 2;

    SceneEntry entry;

    entry.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));

    entry.flags.dontCare = true;
    entry.flags.ignoreLocalPrio = true;
    entry.flags.specialHandling = true;
    entry.flags.blink = true;
    entry.flags.ledConfig = LedConfig::ledCon0;
    entry.flags.dimTime = DimTime::dimTime0;

    entry.action = "";

    entry.channels[ChannelType::shadePositionOutside].value = 50.0;
    entry.channels[ChannelType::shadePositionOutside].dontCare = false;

    for (auto& scene : mostlyThere.scene.sceneTable) {
      scene.second = entry;
    }

    // Now delete some part and check if they get fetched

    mostlyThere.scene.sceneTable[2].flags.specialHandling = {};
    mostlyThere.scene.sceneTable[2].flags.blink = {};
    mostlyThere.scene.sceneTable[56].flags.ignoreLocalPrio = {};
    mostlyThere.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value = {};

    mostlyThere.scene.sceneTable[6].channels.clear();

    mostlyThere.property[DeviceProperty::posTimeDown] = 10;

    db.update(dsuid2, mostlyThere);

    WHEN("Periodic cache update runs") {
      REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
          .WITH(_1.scene.dsuid == dsuid2)
          .WITH(_1.scene.sceneId == 2)
          .WITH(_1.property == SceneProperty::dontCare)
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[2].flags.blink = false;
            devCache.scene.sceneTable[2].flags.specialHandling = false;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/scene/5/shadePositionOutside)")))
          .TIMES(1)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value = 2.1;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
          .WITH(_1.scene.dsuid == dsuid2)
          .WITH(_1.scene.sceneId == 56)
          .WITH(_1.property == SceneProperty::dontCare)
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[56].flags.ignoreLocalPrio = false;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/dimStepNum)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::dimStepNum] = 1;
        db.update(dsuid2, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/dimStepPause)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::dimStepPause] = 1;
        db.update(dsuid2, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/posTimeUp)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::posTimeUp] = 1;
        db.update(dsuid2, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/scene/6/shadePositionOutside)")))
          .TIMES(1)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[6].channels[ChannelType::shadePositionOutside].value = 2.1;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readBool(ANY(DeviceSceneChannelPropertyId)))
          .WITH(_1.channel.scene.dsuid == dsuid2)
          .WITH(_1.channel.scene.sceneId == 6)
          .WITH(_1.channel.type == ChannelType::shadePositionOutside)
          .WITH(_1.property == ChannelProperty::dontCare)
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[6].channels[ChannelType::shadePositionOutside].dontCare = false;
            db.update(dsuid2, devCache);
          });

      std::shared_ptr<Event> pEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventFast);
      std::shared_ptr<Event> pResumeEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEventFast);
      auto subscription = std::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);
      interface.handleEvent(*pEvent, *subscription);
      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      THEN("The device should have all missing scences fetched from device") {}
    }
  }

  GIVEN("3 Devices in apartmentwith 2 existing devices in cache (1 device not supported by HAL)") {
    auto devShade = createDevice(apt, {.hardwareType = HardwareType::GR_KL200});
    const auto& dsuid2 = devShade->getDSID();

    devShade->setIsPresent(true);
    devShade->setIsConnected(true);

    hal.populateCache(dsuid2);
    auto& db = deviceCache;
    auto completeShade = db.tryLookup(dsuid2);

    completeShade.property[DeviceProperty::dimStepNum] = 4;
    completeShade.property[DeviceProperty::dimStepLen] = 3;
    completeShade.property[DeviceProperty::dimStepPause] = 4;
    completeShade.property[DeviceProperty::posTimeUp] = 4;
    completeShade.property[DeviceProperty::posTimeDown] = 4;

    completeShade.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));

    SceneEntry entry;

    entry.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));

    entry.flags.dontCare = true;
    entry.flags.ignoreLocalPrio = true;
    entry.flags.specialHandling = true;
    entry.flags.blink = true;
    entry.flags.ledConfig = LedConfig::ledCon0;
    entry.flags.dimTime = DimTime::dimTime0;

    entry.action = "";

    entry.channels[ChannelType::shadePositionOutside].value = 50.0;
    entry.channels[ChannelType::shadePositionOutside].dontCare = false;

    for (auto& scene : completeShade.scene.sceneTable) {
      scene.second = entry;
    }
    db.update(dsuid2, completeShade);

    auto devLight = createDevice(apt, {.hardwareType = HardwareType::GE_KL200});
    const auto& dsuid1 = devLight->getDSID();

    devLight->setIsPresent(true);
    devLight->setIsConnected(true);

    hal.populateCache(dsuid1);
    auto completeLight = db.tryLookup(dsuid1);

    completeLight.property[DeviceProperty::outputMode] = 1;
    completeLight.property[DeviceProperty::switchedThreshold] = 1;
    completeLight.property[DeviceProperty::dimTime0Up] = 1;
    completeLight.property[DeviceProperty::dimTime0Down] = 1;
    completeLight.property[DeviceProperty::dimTime1Up] = 1;
    completeLight.property[DeviceProperty::dimTime1Down] = 1;
    completeLight.property[DeviceProperty::dimTime2Up] = 1;
    completeLight.property[DeviceProperty::dimTime2Down] = 1;
    completeLight.property[DeviceProperty::dimStepUp] = 1;
    completeLight.property[DeviceProperty::dimStepDown] = 1;
    completeLight.property[DeviceProperty::minDim] = 1;
    completeLight.property[DeviceProperty::dimStepPause] = 1; // superfluous entry

    // Modify entry fore reuse

    entry.channels.clear();
    entry.channels[ChannelType::brightness].value = 100.0;
    entry.channels[ChannelType::brightness].dontCare = false;

    for (auto& scene : completeLight.scene.sceneTable) {
      scene.second = entry;
    }
    completeLight.scene.sceneTable[127] = entry;
    completeLight.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value = 42;
    db.update(dsuid1, completeLight);

    auto devUnsupported = createDevice(apt, {.hardwareType = HardwareType::BL_KM200});

    devUnsupported->setIsPresent(true);
    devUnsupported->setIsConnected(true);

    WHEN("Periodic cache update runs and something changes") {
      // First no calls are allowed
      std::shared_ptr<Event> pEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventFast);
      std::shared_ptr<Event> pResumeEvent = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEventFast);
      auto subscription = std::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);

      interface.handleEvent(*pEvent, *subscription);
      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      // Then things get invalidated and calls are expected
      DeviceCache lightCache = db.tryLookup(dsuid1);
      lightCache.scene.sceneTable[5].channels[ChannelType::brightness].value = {};
      lightCache.scene.sceneTable[2].channels[ChannelType::brightness].dontCare = {};
      db.update(dsuid1, lightCache);

      DeviceCache shadeCache = db.tryLookup(dsuid2);
      shadeCache.scene.sceneTable[2].flags.specialHandling = {};
      shadeCache.property[DeviceProperty::dimStepLen] = {};
      db.update(dsuid2, shadeCache);

      REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
          .WITH(_1.scene.dsuid == dsuid2)
          .WITH(_1.scene.sceneId == 2)
          .WITH(_1.property == SceneProperty::dontCare)
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[2].flags.specialHandling = false;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/scene/5/brightness)")))
          .TIMES(1)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[5].channels[ChannelType::brightness].value = 2.1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(ANY(DeviceSceneChannelPropertyId)))
          .WITH(_1.channel.scene.dsuid == dsuid1)
          .WITH(_1.channel.scene.sceneId == 2)
          .WITH(_1.channel.type == ChannelType::brightness)
          .WITH(_1.property == ChannelProperty::dontCare)
          .TIMES(1)
          .RETURN(true)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[2].channels[ChannelType::brightness].dontCare = true;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/dimStepLen)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::dimStepLen] = 1;
        db.update(dsuid2, devCache);
      });

      interface.handleEvent(*pEvent, *subscription);
      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }
      {
        auto completeLight = db.tryLookup(dsuid1);
        REQUIRE_THROWS(completeLight.property.at(DeviceProperty::dimStepPause));
        REQUIRE_THROWS(completeLight.scene.sceneTable.at(127));
        REQUIRE_THROWS(completeLight.scene.sceneTable[5].channels.at(ChannelType::shadePositionOutside));
      }
    }
  }
}

SCENARIO("Priority device readout", std::string(TAGS) + "[fw387]") {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;
  auto propSystem = std::make_shared<PropertySystemMock>();

  EventInterpreter interpreter(nullptr);
  std::shared_ptr<EventQueue> queue(new EventQueue(&interpreter));
  std::shared_ptr<EventRunner> runner(new EventRunner(&interpreter));

  interpreter.initialize();
  interpreter.setEventQueue(queue.get());
  interpreter.setEventRunner(runner.get());
  queue->setEventRunner(runner.get());
  runner->setEventQueue(queue.get());

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  HardwareAbstractionMock hal{};
  ds::sqlite3::Database db(":memory:");
  auto deviceCache = CacheDb<DeviceCache>(&db);

  ALLOW_CALL(hal, getCache()).LR_RETURN(deviceCache);
  CacheUpdatePublic interface(&apt, &hal, &db, interpreter, propSystem, queue, runner);
  interface.handleModelReady();

  // giving up on restricting them...
  ALLOW_CALL(hal, deviceCacheable(_)).RETURN(true);
  ALLOW_CALL(hal, getCache()).LR_RETURN(deviceCache);

  auto populateCache = [&deviceCache, &db](const auto& device) {
    auto stateCache = CacheDb<DeviceInternalState>(&db);
    auto halLightDevice = hal::HalLightDevice(deviceCache, stateCache);
    auto halShadeDevice = hal::HalShadeDevice(deviceCache, stateCache);
    switch (device.getDeviceClass()) {
      case DeviceClasses_t::DEVICE_CLASS_INVALID: break;
      case DeviceClasses_t::DEVICE_CLASS_GE: halLightDevice.populateCache(device); break;
      case DeviceClasses_t::DEVICE_CLASS_GR: halShadeDevice.populateCache(device); break;
      case DeviceClasses_t::DEVICE_CLASS_BL: break;
      case DeviceClasses_t::DEVICE_CLASS_TK: break;
      case DeviceClasses_t::DEVICE_CLASS_MG: break;
      case DeviceClasses_t::DEVICE_CLASS_RT: break;
      case DeviceClasses_t::DEVICE_CLASS_GN: break;
      case DeviceClasses_t::DEVICE_CLASS_SW: break;
      case DeviceClasses_t::DEVICE_CLASS_WE: break;
      default: break;
    }
  };
  ALLOW_CALL(hal, populateCache(_)).LR_SIDE_EFFECT(populateCache(*apt.getDeviceByDSID(_1)));

  GIVEN("two normal devices") {
    auto light = createDevice(apt, DeviceMockConf{.hardwareType = HardwareType::GE_TKM210});
    markOnline(*light);
    auto blind = createDevice(apt, DeviceMockConf{.hardwareType = HardwareType::GR_KL220, .revisionId = 0x387});
    markOnline(*blind);

    WHEN("When reading out the cache") {
      // ensure the natural/normal scan order is different from priority scanning
      // e.g. if blinds were scanned first anyway, then the follow-up unit tests are futile
      THEN("one device is consistently read-out before the other") {
        REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
            .WITH(_1.scene.dsuid == light->getDSID())
            .WITH(_1.property == SceneProperty::dontCare)
            .TIMES(1)
            .RETURN(false)
            .LR_SIDE_EFFECT({ updateSceneFlags(hal.getCache(), _1.scene, {.dontCare = false}); });
        interface.scanStart(ScanMode::NORMAL);
        REQUIRE(getSceneEntry(hal.getCache(), {.dsuid = light->getDSID(), .sceneId = 0}).flags.dontCare);
      }
    }

    WHEN("prioritizing read-out of the other") {
      THEN("the other device is read-out first") {
        REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
            .WITH(_1.scene.dsuid == blind->getDSID())
            .WITH(_1.property == SceneProperty::dontCare)
            .TIMES(1)
            .RETURN(false)
            .LR_SIDE_EFFECT({ updateSceneFlags(hal.getCache(), _1.scene, {.dontCare = false}); });
        interface.scanDevice(blind->getDSID());
        REQUIRE(getSceneEntry(hal.getCache(), {.dsuid = blind->getDSID(), .sceneId = 0}).flags.dontCare);
      }
    }

    WHEN("running cache-update is interrupted by priority update") {
      REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
          .WITH(_1.scene.dsuid == light->getDSID())
          .WITH(_1.property == SceneProperty::dontCare)
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({ updateSceneFlags(hal.getCache(), _1.scene, {.dontCare = false}); });
      interface.scanStart(ScanMode::NORMAL);
      REQUIRE(interface.isUpdateRunning());

      THEN("Priority update is executed first") {
        REQUIRE_CALL(hal, readBool(ANY(DeviceScenePropertyId)))
            .WITH(_1.scene.dsuid == blind->getDSID())
            .WITH(_1.property == SceneProperty::dontCare)
            .TIMES(1)
            .RETURN(false)
            .LR_SIDE_EFFECT({ updateSceneFlags(hal.getCache(), _1.scene, {.dontCare = false}); });

        interface.scanDevice(blind->getDSID());
        interface.scanStep();
      }

      THEN("cache update continues afterwards and ApartmentProxyChanged is triggered") {
        fillCacheForDevice(hal, blind->getDSID());
        REQUIRE(getSceneEntry(hal.getCache(), {.dsuid = blind->getDSID(), .sceneId = 0}).flags.dontCare);

        DS_NOTICE(light->getDSID());
        DS_NOTICE(blind->getDSID());

        interface.scanDevice(blind->getDSID()); // scan already running, no-action

        // read out of dontCare / channel can flip, ensure its always dontCare
        updateChannelValue(hal.getCache(), DeviceSceneChannelId{{light->getDSID(), SceneOff}, ChannelType::brightness},
                           40.0);

        REQUIRE_CALL(hal, readBool(ANY(DeviceSceneChannelPropertyId)))
            .WITH(_1.channel.scene.dsuid == light->getDSID())
            .WITH(_1.channel.type == ChannelType::brightness)
            .WITH(_1.property == ChannelProperty::dontCare)
            .TIMES(1)
            .RETURN(false)
            .LR_SIDE_EFFECT({ updateChannelDontCare(hal.getCache(), _1.channel, false); });

        REQUIRE_CALL(eventSink, pushEvent(_)); // ApartmentProxyStateChanged
        interface.scanStep();
      }
    }
  }
}

SCENARIO("Migrate fw 3.8.7 cache entries", std::string(TAGS) + "[fw387]") {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  EventInterpreter interpreter{nullptr};
  auto propSystem = std::make_shared<PropertySystemMock>();
  std::shared_ptr<EventQueue> queue{new EventQueue(&interpreter)};
  std::shared_ptr<EventRunner> runner{new EventRunner(&interpreter)};

  interpreter.initialize();
  runner->setEventQueue(queue.get());
  queue->setEventRunner(runner.get());
  interpreter.setEventQueue(queue.get());
  interpreter.setEventRunner(runner.get());

  HardwareAbstractionMock hal{};
  ds::sqlite3::Database db(":memory:");
  auto deviceCache = CacheDb<DeviceCache>(&db);
  ALLOW_CALL(hal, getCache()).LR_RETURN(deviceCache); // migration
  CacheUpdatePublic interface{&apt, &hal, &db, interpreter, propSystem, queue, runner};
  interface.handleModelReady();

  // giving up on restricting them...
  ALLOW_CALL(hal, deviceCacheable(_)).RETURN(true);
  ALLOW_CALL(hal, getCache()).LR_RETURN(deviceCache);

  auto populateCache = [&deviceCache, &db](const auto& device) {
    auto stateCache = CacheDb<DeviceInternalState>(&db);
    auto halLightDevice = hal::HalLightDevice(deviceCache, stateCache);
    auto halShadeDevice = hal::HalShadeDevice(deviceCache, stateCache);
    switch (device.getDeviceClass()) {
      case DeviceClasses_t::DEVICE_CLASS_INVALID: break;
      case DeviceClasses_t::DEVICE_CLASS_GE: halLightDevice.populateCache(device); break;
      case DeviceClasses_t::DEVICE_CLASS_GR: halShadeDevice.populateCache(device); break;
      case DeviceClasses_t::DEVICE_CLASS_BL: break;
      case DeviceClasses_t::DEVICE_CLASS_TK: break;
      case DeviceClasses_t::DEVICE_CLASS_MG: break;
      case DeviceClasses_t::DEVICE_CLASS_RT: break;
      case DeviceClasses_t::DEVICE_CLASS_GN: break;
      case DeviceClasses_t::DEVICE_CLASS_SW: break;
      case DeviceClasses_t::DEVICE_CLASS_WE: break;
      default: break;
    }
  };
  ALLOW_CALL(hal, populateCache(_)).LR_SIDE_EFFECT(populateCache(*apt.getDeviceByDSID(_1)));

  GIVEN("DeviceCache of version 0 and device with fw:3.8.7 in it") {
    auto dev = createDevice(apt, DeviceMockConf{.hardwareType = HardwareType::GR_KL220, .revisionId = 0x387});
    markOnline(*dev);

    fillCacheForDevice(hal, dev->getDSID());
    auto scene1 = getSceneEntry(hal.getCache(), {dev->getDSID(), Scene1});
    REQUIRE(!scene1.flags.dontCare.value()); // -> care
    REQUIRE(!scene1.channels[ChannelType::shadePositionOutside].dontCare.value());
    REQUIRE(scene1.channels[ChannelType::shadePositionOutside].value);
    REQUIRE(scene1.channels[ChannelType::shadeOpeningAngleOutside].value);

    interface.updateMetaInfo({.version = 0});

    WHEN("Cache updater plugin is created") {
      // nop
      THEN("Then cache entry for 3.8.7 device isn't cleared") {
        auto scene1 = getSceneEntry(hal.getCache(), {dev->getDSID(), Scene1});
        REQUIRE(scene1.channels[ChannelType::shadePositionOutside].value);
        REQUIRE(scene1.channels[ChannelType::shadeOpeningAngleOutside].value);
      }

      AND_THEN("Device cache version is unchanged") { REQUIRE(interface.getMetaInfo().version == 0); }
    }

    WHEN("ModelReady is received") {
      interface.handleModelReady();

      THEN("Then cache entry for 3.8.7 device is cleared") {
        auto scene1 = getSceneEntry(hal.getCache(), {dev->getDSID(), Scene1});
        REQUIRE(!scene1.channels[ChannelType::shadePositionOutside].value);
        REQUIRE(!scene1.channels[ChannelType::shadeOpeningAngleOutside].value);
      }
      AND_THEN("Device cache version is updated") {
        REQUIRE(interface.getMetaInfo().version == CacheUpdatePublic::m_deviceCacheVersion);
      }
    }
  }

  GIVEN("Cached non 3.8.7 device") {
    auto dev = createDevice(apt, DeviceMockConf{.hardwareType = HardwareType::GR_KL220});
    markOnline(*dev);

    fillCacheForDevice(hal, dev->getDSID());
    interface.updateMetaInfo({.version = 0});

    WHEN("ModelReady is received") {
      interface.handleModelReady();

      THEN("Then cache entry device is not modified") {
        auto scene1 = getSceneEntry(hal.getCache(), {dev->getDSID(), Scene1});
        REQUIRE(scene1.channels[ChannelType::shadePositionOutside].value);
        REQUIRE(scene1.channels[ChannelType::shadeOpeningAngleOutside].value);
      }
    }
  }

  GIVEN("Device cache of higher version") {
    interface.updateMetaInfo({.version = CacheUpdatePublic::m_deviceCacheVersion + 1});

    WHEN("ModelReady is received") {
      interface.handleModelReady();

      THEN("Device cache version is downgraded") {
        REQUIRE(interface.getMetaInfo().version == CacheUpdatePublic::m_deviceCacheVersion);
      }
    }
  }
}

} // namespace
