#include "route-segments.h"

#include <ds/log.h>

#include <boost/algorithm/string.hpp>
#include <boost/range/algorithm.hpp>

DS_REFLECT_DEFINE_STRUCT((dsApi), RouteSegmentsView);
DS_REFLECT_DEFINE_STRUCT((dsApi), MatchRouteResult);

static ds::LogChannel dsLogChannel("dsApiRouteSegments");

namespace dsApi {

std::ostream& operator<<(std::ostream& ostream, const RouteSegment& x) {
  if (const auto literal = std::get_if<LiteralRouteSegment>(&x)) {
    return ostream << "literal:" << gsl::cstring_span(*literal);
  }
  if (const auto parameter = std::get_if<ArgRouteSegment>(&x)) {
    return ostream << "parameter:" << gsl::cstring_span(*parameter);
  }
  if (const auto wildcardParameter = std::get_if<WildcardArgRouteSegment>(&x)) {
    return ostream << "wildcardParameter:" << gsl::cstring_span(*wildcardParameter);
  }
  return ostream << "?";
}

RouteSegmentsView::RouteSegmentsView(const gsl::cstring_span route) {
  DS_DEBUG_ENTER(route);
  if (route.empty()) {
    DS_DEBUG_LEAVE(*this);
    return;
  }
  auto rest = route;
  while (true) {
    if (rest.empty()) {
      break;
    }
    DS_REQUIRE(boost::starts_with(rest, "/"), "Route segment must start with '/'.", rest, route);
    rest = rest.subspan(1);
    auto restIt = boost::find_first_of(rest, "/");
    auto segment = gsl::cstring_span(rest.begin(), restIt);
    DS_REQUIRE(!segment.empty(), "Route segment must not be empty.", segment, rest, route);
    rest = {restIt, rest.end()};
    if (boost::starts_with(segment, "{") && boost::ends_with(segment, "}")) {
      auto parameter = segment.subspan(1, segment.size() - 2);
      if (boost::starts_with(parameter, "*")) {
        DS_REQUIRE(rest.empty(), "Wildcard route parameter must be the last route segment.", route, rest);
        parameter = parameter.subspan(1);
        segments.emplace_back(WildcardArgRouteSegment{parameter});
      } else {
        segments.emplace_back(ArgRouteSegment{parameter});
      }
    } else {
      segments.emplace_back(LiteralRouteSegment{segment});
    }
    if (rest.empty()) {
      break;
    }
  }
  DS_DEBUG_LEAVE(*this);
}

MatchRouteResult RouteSegmentsView::match(const gsl::cstring_span path) const {
  DS_DEBUG_ENTER(*this, path);
  MatchRouteResult result{};
  auto pathRest = path;
  for (const auto& routeSegment : this->segments) {
    DS_DEBUG(routeSegment, pathRest, result);
    if (const auto wildcardParameter = std::get_if<WildcardArgRouteSegment>(&routeSegment)) {
      // capture path rest as wildcard route parameter
      result.args[gsl::to_string(gsl::cstring_span(*wildcardParameter))] = gsl::to_string(pathRest);
      pathRest = {};
      continue;
    }
    if (!boost::starts_with(pathRest, "/")) {
      DS_DEBUG_LEAVE("Path too short", *this, pathRest);
      return {};
    }
    pathRest = pathRest.subspan(1);
    const auto pathRestIt = boost::find_first_of(pathRest, "/");
    const auto pathSegment = gsl::cstring_span(pathRest.begin(), pathRestIt);
    pathRest = {pathRestIt, pathRest.end()};
    if (const auto literal = std::get_if<LiteralRouteSegment>(&routeSegment)) {
      // check that path segment matches route literal
      if (gsl::cstring_span(*literal) != pathSegment) {
        DS_DEBUG_LEAVE("Literal match failed", *this, *literal, pathSegment, pathRest);
        return {};
      }
      result.fitness += 1;
      continue;
    }
    if (const auto parameter = std::get_if<ArgRouteSegment>(&routeSegment)) {
      // capture path segment as route parameter
      result.args[gsl::to_string(gsl::cstring_span(*parameter))] = gsl::to_string(pathSegment);
      continue;
    }
    DS_FAIL_REQUIRE("Unknown route segment");
  }
  result.matched = pathRest.empty() || pathRest == "/";
  DS_DEBUG_LEAVE(*this, result, pathRest);
  return result;
}

MatchRouteResult matchRoute(const gsl::cstring_span routePattern, const gsl::cstring_span path) {
  return RouteSegmentsView(routePattern).match(path);
}

} // namespace dsApi
