/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "client.h"

#include <avahi-common/simple-watch.h>
#include <ds/asio/timer.h>
#include <ds/log.h>
#include <ds/random.h>

#include <algorithm>
#include <boost/foreach.hpp>
#include <boost/system/error_code.hpp>

static ds::LogChannel dsLogChannel("dsAvahiClient");

typedef boost::system::error_code error_code;

namespace dsAvahi {

/// Avahi event poll integration to asio
class Client::Poll : boost::noncopyable {
public:
  Poll(boost::asio::io_service &ioService) : m_ioService(ioService), m_timer(ioService) {
    DS_DEBUG_ENTER("");
    m_simplePoll.reset(avahi_simple_poll_new());
    if (!m_simplePoll) {
      throw std::runtime_error("Failed to create simple poll object.\n");
    }

    asyncIterate();
  }

  const AvahiPoll *ptr() const { return avahi_simple_poll_get(m_simplePoll.get()); }

  boost::asio::io_service &ioService() { return m_ioService; }

private:
  boost::asio::io_service &m_ioService;
  ds::asio::Timer m_timer;

  struct SimplePollDeleter {
    void operator()(::AvahiSimplePoll *x) { avahi_simple_poll_free(x); }
  };
  std::unique_ptr< ::AvahiSimplePoll, SimplePollDeleter> m_simplePoll;

  // Poor man's avahi loop integration into asio
  // It is possible to eliminate the ticking timer by proper integration
  void asyncIterate() {
    m_timer.expiresFromNow(std::chrono::milliseconds(200), [this] {
      // Function avahi_simple_poll_iterate will not tell us whether it has more
      // events to process or not.
      // Processing only 1 event per timeout leads to slow startup.
      // TODO(someday): write proper avahi loop integration
      for (int i = 0; i < 8; i++) {
        if (avahi_simple_poll_iterate(m_simplePoll.get(), 0) != 0) {
          throw std::runtime_error("Avahi poll iterate error");
        }
      }
      asyncIterate();
    });
  }
};

Client::Hook::~Hook() = default;
void Client::Hook::onRunning() {}

Client::Client(boost::asio::io_service &ioService)
    : m_poll(new Poll(ioService)), m_isRunning(false), m_onReconnectOnError([](const std::exception &) {}) {
  DS_DEBUG_ENTER("");
  reconnect();
}

Client::~Client() = default;

void Client::addHook(Hook &x) {
  m_hooks.push_back(&x);
  try {
    if (m_isRunning) {
      x.onRunning();
    }
  } catch (const std::exception &e) {
    reconnectOnError(e);
  }
}

void Client::removeHook(Hook &x) {
  m_hooks.erase(std::remove(m_hooks.begin(), m_hooks.end(), &x), m_hooks.end());
}

void Client::setRandomTypePrefix() {
  m_typePrefix = ds::str('_', ds::randint(100000000000, 999999999999));
}

boost::asio::io_service &Client::ioService() {
  return m_poll->ioService();
}

void Client::reconnectOnError(const std::exception &e) {
  DS_ERROR("Reconnecting on error.", e);

  // TODO(someday): callback shall be the last call in the function
  // We need to call `reconnect()` asynchronously.
  // But it fine now as we use this callback for only mocking purpuses now.
  m_onReconnectOnError(e);

  // TODO(someday): asynchronous call with exponentional backoff to avoid inifinite cycles

  reconnect();
}

void Client::setRunning() {
  if (m_isRunning) {
    return;
  }
  DS_DEBUG_ENTER("");
  m_isRunning = true;
  // m_hooks is list, so it is correct to delete hooks while iterating.
  BOOST_FOREACH (auto &&x, m_hooks) { x->onRunning(); }
}

void Client::reconnect() {
  DS_DEBUG_ENTER(m_isRunning);
  try {
    m_ptr = nullptr;
    m_uniquePtr.reset();
    m_isRunning = false;
    // Avahi library automatically freed all associated
    // browser, resolve and entry group objects.
    BOOST_FOREACH (auto &&x, m_hooks) { x->freed(); }

    int error;
    // avahi_client_new may call `callback` and set m_ptr
    m_uniquePtr.reset(avahi_client_new(m_poll->ptr(), AVAHI_CLIENT_NO_FAIL, sCallback, this, &error));
    DS_INFO("avahi_client_new finished");

    if (!m_uniquePtr) {
      // According to documentation, this should not happen
      // because we pass AVAHI_CLIENT_NO_FAIL flag.
      // But it happens: http://redmine.digitalstrom.org/issues/20360
      DS_AVAHI_FAIL(error, "Failed to create avahi client.");
    }
    if (!m_ptr) {
      m_ptr = m_uniquePtr.get();
    }
    DS_ASSERT(m_ptr == m_uniquePtr.get(), m_ptr, m_uniquePtr.get());
  } catch (const std::exception &e) {
    // TODO(someday): We may want to reconnect with exponential backoff
    DS_FAIL_ASSERT(e); // Unrecoverable error for now
  }
}

void Client::sCallback(AvahiClient *ptr, AvahiClientState state, void *data) {
  auto &&x = *static_cast<Client *>(data);
  try {
    x.callback(ptr, state);
  } catch (const std::exception &e) {
    x.reconnectOnError(e);
  }
}

void Client::callback(AvahiClient *ptr, AvahiClientState state) {
  DS_INFO_ENTER(state);
  if (!m_ptr) {
    m_ptr = ptr;
    DS_DEBUG("m_ptr set from callback", m_ptr);
  }
  switch (state) {
    case AVAHI_CLIENT_S_RUNNING:
      // The server has startup successfully and registered its host
      // name on the network, so it's time to create our services
      DS_NOTICE("Server running");
      setRunning();
      break;

    case AVAHI_CLIENT_S_COLLISION:
      // Let's drop our registered services. When the server is back
      // in AVAHI_SERVER_RUNNING state we will register them
      // again with the new host name.
      DS_NOTICE("Server collision");
      if (m_isRunning) {
        DS_AVAHI_FAIL_CLIENT_ERROR(*this, "Server is registering");
      }
      break;
    case AVAHI_CLIENT_S_REGISTERING:
      // The server records are now being established. This
      // might be caused by a host name change. We need to wait
      // for our own records to register until the host name is
      // properly established.
      DS_NOTICE("Server registering");
      if (m_isRunning) {
        DS_AVAHI_FAIL_CLIENT_ERROR(*this, "Server is registering");
      }
      break;
    case AVAHI_CLIENT_FAILURE: DS_AVAHI_FAIL_CLIENT_ERROR(*this, "Avahi client failure"); break;
    case AVAHI_CLIENT_CONNECTING:
      DS_NOTICE("Connecting, deamon not available.");
      // Nothing to do. We wait for avahi deamon to be started.
      break;
  }
}

} // namespace dsAvahi
