/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "lookup.h"

#include <ds/log.h>
#include <gsl.h>

#include <boost/algorithm/string/predicate.hpp>

static ds::LogChannel dsLogChannel("dsAvahiLookup");

namespace dsAvahi {

ServicesBrowser::ServicesBrowser(Client &client, std::string type, OnEvent onEvent)
    : m_client(client),
      m_type(type),
      m_flags(AvahiLookupFlags()),
      m_onEvent(std::move(onEvent)),
      m_isPendingReset(false),
      m_pendingResetTimer(client.ioService()) {
  DS_INFO_ENTER(type, m_flags);
  m_client.addHook(*this);
}

ServicesBrowser::~ServicesBrowser() {
  DS_INFO_ENTER("");
  m_client.removeHook(*this);
}

void ServicesBrowser::onRunning() {
  DS_ASSERT(!m_ptr);
  DS_INFO_ENTER("");
  const auto &type = ds::str(typePrefix(), m_type);
  m_ptr.reset(avahi_service_browser_new(m_client.ptr(), AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, type.c_str(), nullptr,
                                        m_flags, sCallback, this));
  DS_NOTICE("Browse services.", type, m_flags);
  if (!m_ptr) {
    DS_AVAHI_FAIL_CLIENT_ERROR(m_client, "Failed to create service browser");
  }
  emitPendingReset();
}

void ServicesBrowser::emitPendingReset() {
  if (!m_isPendingReset) {
    return;
  }
  DS_INFO("Emit Reset");
  m_onEvent(ds::collection::Reset());
}

void ServicesBrowser::freed() {
  if (m_ptr) {
    DS_INFO_ENTER("");
    m_ptr.release();
    m_isPendingReset = true;
    // No callbacks allowed
    m_pendingResetTimer.expiresNow([this]() { emitPendingReset(); });
  }
}

std::string ServicesBrowser::createServiceId(AvahiIfIndex interface, AvahiProtocol protocol, const char *name,
                                             const char *type, const char *domain) {
  std::ostringstream s;
  s << interface << ':' << protocol << ':' << type << ':' << domain << ':' << name;
  return s.str();
}

void ServicesBrowser::sCallback(AvahiServiceBrowser *, AvahiIfIndex interface, AvahiProtocol protocol,
                                AvahiBrowserEvent event, const char *name, const char *type, const char *domain,
                                AvahiLookupResultFlags flags, void *data) {
  auto this_ = static_cast<ServicesBrowser *>(data);
  try {
    this_->callback(interface, protocol, event, name, type, domain, flags);
  } catch (const std::exception &e) {
    this_->m_client.reconnectOnError(e);
  }
}

void ServicesBrowser::callback(AvahiIfIndex interface, AvahiProtocol protocol, AvahiBrowserEvent event,
                               const char *name, const char *type, const char *domain, AvahiLookupResultFlags flags) {
  DS_INFO_ENTER(interface, protocol, event, name, type, domain, flags);
  auto createId = [&]() { return createServiceId(interface, protocol, name, type, domain); };
  switch (event) {
    case AVAHI_BROWSER_NEW: {
      DS_NOTICE("Browsed service.", name, type, domain);
      auto id = createId();
      DS_INFO("Emit  Added", id);
      auto &&typePrefix = this->typePrefix();
      DS_REQUIRE(boost::starts_with(type, typePrefix.c_str()));
      auto typeWithoutPrefix = gsl::cstring_span(type).subspan(typePrefix.size());
      m_onEvent(ds::collection::Added<BrowsedService>(
          std::move(id), {interface, protocol, name, to_string(typeWithoutPrefix), domain}));
      break;
    }
    case AVAHI_BROWSER_REMOVE: {
      DS_NOTICE("Removed service.", name, type);
      auto id = createId();
      DS_INFO("Emit  Removed", id);
      m_onEvent(ds::collection::Removed<BrowsedService>(std::move(id)));
      break;
    }
    case AVAHI_BROWSER_ALL_FOR_NOW:
      DS_NOTICE("AllForNow.", type);
      m_onEvent(ds::collection::AllForNow());
      break;
    case AVAHI_BROWSER_CACHE_EXHAUSTED: break;
    case AVAHI_BROWSER_FAILURE:
      throw std::runtime_error(std::string("Browse failure ") + avahi_strerror(avahi_client_errno(m_client.ptr())));
      break;
  }
}

/// ***
/// *** ServiceResolver
/// ***

ServiceResolver::ServiceResolver(Client &client, const BrowsedService &service, AvahiLookupFlags flags, Replied replied)
    : m_client(client), m_flags(flags), m_replied(std::move(replied)) {
  DS_INFO_ENTER(service.name, service.type, flags);

  DS_REQUIRE(m_client.isRunning(), "ServiceResolver is usable only in client running state");
  m_client.addHook(*this);
  try {
    const auto &interface = service.interface;
    const auto &protocol = service.protocol;
    const auto &name = service.name;
    const auto &type = ds::str(typePrefix(), service.type);
    m_ptr.reset(avahi_service_resolver_new(m_client.ptr(), interface, protocol, name.c_str(), type.c_str(),
                                           service.domain.c_str(), AVAHI_PROTO_UNSPEC, flags, sCallback, this));
    DS_NOTICE("Resolve service.", interface, protocol, name, type);
    if (!m_ptr) {
      DS_AVAHI_FAIL_CLIENT_ERROR(m_client, "Failed to resolve service. ", service.name);
    }
  } catch (const std::exception &e) {
    // reconnectOnError implies that associated browser will be deleted
    m_client.reconnectOnError(e);
  }
}

ServiceResolver::~ServiceResolver() {
  DS_INFO_ENTER(m_ptr.get());
  m_client.removeHook(*this);
}

void ServiceResolver::freed() {
  DS_INFO_ENTER(m_ptr.get());
  m_ptr.release();

  // TODO(someday): emit failure.
  // It must be asynchrously and only when a failure was not already reported.
  //
  // We rely now on Removed or Reset event from ServicesBrowser to cover this use case.
}

void ServiceResolver::sCallback(AvahiServiceResolver *, AvahiIfIndex interface, AvahiProtocol protocol,
                                AvahiResolverEvent event, const char *name, const char *type, const char *domain,
                                const char *hostName, const AvahiAddress *address, uint16_t port, AvahiStringList *txt,
                                AvahiLookupResultFlags flags, void *data) {
  auto this_ = static_cast<ServiceResolver *>(data);
  try {
    this_->callback(interface, protocol, event, name, type, domain, hostName, address, port, txt, flags);
  } catch (const std::exception &e) {
    this_->m_client.reconnectOnError(e);
  }
}

void ServiceResolver::callback(AvahiIfIndex interface, AvahiProtocol protocol, AvahiResolverEvent event,
                               const char *name, const char *type, const char *domain, const char *hostName,
                               const AvahiAddress *address, uint16_t port, AvahiStringList *txtList,
                               AvahiLookupResultFlags flags) {
  DS_INFO_ENTER(interface, protocol, event, name, type, domain, hostName, address, port, flags);
  /* Called whenever a service has been resolved successfully or timed out */
  switch (event) {
    case AVAHI_RESOLVER_FAILURE:
      // TODO(someday): Is this error really failure of the resolve operation?
      // Will avahi resolver send us further events?
      // Or should we signal error to the poll class and remove error from our interface?
      try {
        DS_AVAHI_FAIL_CLIENT_ERROR(m_client, "Service resolve failed.", name, type);
      } catch (const std::exception &e) {
        DS_WARNING("Service resolve failed.", e);
        m_replied(e.what());
      }
      break;
    case AVAHI_RESOLVER_FOUND: {
      char addressZStr[AVAHI_ADDRESS_STR_MAX];
      avahi_address_snprint(addressZStr, sizeof(addressZStr), address);
      std::vector<std::string> txts;
      txts.reserve(avahi_string_list_length(txtList));
      while (txtList) {
        txts.emplace_back(reinterpret_cast<const char *>(avahi_string_list_get_text(txtList)));
        txtList = avahi_string_list_get_next(txtList);
      }
      auto &&typePrefix = this->typePrefix();
      DS_REQUIRE(boost::starts_with(type, typePrefix.c_str()));
      auto typeWithoutPrefix = type + typePrefix.size();
      DS_NOTICE("Service resolved.", name, type, addressZStr, port, txts);
      m_replied(ResolvedService(
          {interface, protocol, name, typeWithoutPrefix, domain, addressZStr, port, std::move(txts), flags}));
    }
  }
}

/// ***
/// *** ServicesResolver
/// ***

ServicesResolver::ServicesResolver(Client &client, std::string type, OnEvent onEvent)
    : m_browser(client, std::move(type), [this](ServicesBrowser::Event event) { onBrowserEvent(std::move(event)); }),
      m_onEvent(std::move(onEvent)),
      m_allForNowPendingCount(1) // waiting for AllForNow from browser
{
  DS_INFO_ENTER(type);
}

ServicesResolver::~ServicesResolver() {
  DS_INFO_ENTER("");
}

struct ServicesResolver::BrowserEventVisitor {
  ServicesResolver &m_parent;

  BrowserEventVisitor(ServicesResolver &parent) : m_parent(parent) {}

  void operator()(const ds::collection::Added<BrowsedService> &x) const {
    auto &parent = m_parent;
    auto &id = x.id;
    auto &value = x.value;
    auto &entries = parent.m_entries;
    DS_DEBUG_ENTER("Add", id, entries.size());

    std::unique_ptr<ServiceResolver> resolver(new ServiceResolver(
        // Do not capture `this` because `this` is temporary closure object
        parent.client(), value, parent.flags(),
        [&parent, id](ServiceResolver::Reply reply) { parent.onResolvedService(id, std::move(reply)); }));
    entries.add(id, std::make_unique<Entry>(Entry{std::move(resolver)}));
    parent.incDecResolvingCount(1);
  }

  void operator()(const ds::collection::Removed<BrowsedService> &x) const {
    auto &parent = m_parent;
    auto &&id = x.id;
    auto &entries = parent.m_entries;
    DS_DEBUG_ENTER("Remove", id, entries.size());
    auto it = entries.find(id);
    auto oldResolved = it->value->resolved;
    entries.erase(it);
    if (oldResolved) {
      parent.m_onEvent(ds::collection::Removed<ResolvedService>(std::move(id)));
    } else {
      parent.incDecResolvingCount(-1);
    }
  }

  void operator()(const ds::collection::Changed<BrowsedService> &) const {
    DS_FAIL_ASSERT("ServicesBrowser does not send Changed event");
  }

  void operator()(const ds::collection::AllForNow &) const {
    DS_DEBUG_ENTER("AllForNow");
    DS_REQUIRE(m_parent.m_allForNowPendingCount > 0, m_parent.m_allForNowPendingCount);
    m_parent.incDecResolvingCount(-1);
  }

  void operator()(const ds::collection::Reset &) const {
    DS_DEBUG_ENTER("Reset");
    m_parent.m_entries.reset();
    m_parent.m_allForNowPendingCount = 1; // waiting for AllForNow from browser
    DS_INFO("Emit  AllForNow");
    m_parent.m_onEvent(ds::collection::Reset());
  }
};

void ServicesResolver::onBrowserEvent(ServicesBrowser::Event event) {
  std::visit(BrowserEventVisitor(*this), event);
}

void ServicesResolver::incDecResolvingCount(int diff) {
  DS_DEBUG_ENTER(m_allForNowPendingCount, diff)
  if (m_allForNowPendingCount == 0) {
    // AllForNow was already sent
    return;
  }
  m_allForNowPendingCount += diff;
  DS_REQUIRE(m_allForNowPendingCount >= 0);
  if (m_allForNowPendingCount == 0) {
    DS_INFO("Emit  AllForNow");
    m_onEvent(ds::collection::AllForNow());
  }
}

void ServicesResolver::onResolvedService(const std::string &id, ServiceResolver::Reply reply) {
  auto &entry = *m_entries.at(id);
  auto oldResolved = entry.resolved;
  bool resolved = std::get_if<ResolvedService>(&reply) != nullptr;
  DS_DEBUG_ENTER(id, oldResolved, resolved)
  if (!resolved) {
    auto &&error = std::get<std::string>(reply);
    DS_WARNING("Resolving failed", id, error);
  }
  entry.resolved = resolved;
  if (oldResolved && resolved) {
    DS_INFO("Emit  Changed", id);
    m_onEvent(ds::collection::Changed<ResolvedService>(id, std::move(std::get<ResolvedService>(reply))));
  } else if (oldResolved && !resolved) {
    DS_INFO("Emit  Removed", id);
    m_onEvent(ds::collection::Removed<ResolvedService>(id));
    incDecResolvingCount(1);
  } else if (!oldResolved && resolved) {
    DS_INFO("Emit  Added", id);
    m_onEvent(ds::collection::Added<ResolvedService>(id, std::move(std::get<ResolvedService>(reply))));
    incDecResolvingCount(-1);
  } else if (!oldResolved && !resolved) {
    // nothing to do
  }
};

} // namespace dsAvahi
