#include "apartment-api.h"

#include <ds-api/service.h>

#include <boost/range/algorithm/copy.hpp>

#include "meterings.h"
#include "open-metering-system.h"
#include "remote-connectivity.h"

DS_JSON_DEFINE_STRUCT((dsModel), ApartmentArgs);
DS_JSON_DEFINE_STRUCT((dsModel), ZoneArgs);
DS_JSON_DEFINE_STRUCT((dsModel), MeteringArgs);
DS_JSON_DEFINE_STRUCT((dsModel), DeviceArgs);
DS_JSON_DEFINE_STRUCT((dsModel), SubmoduleArgs);
DS_JSON_DEFINE_STRUCT((dsModel), FunctionBlockArgs);
DS_JSON_DEFINE_STRUCT((dsModel), OutputStatusArgs);
DS_JSON_DEFINE_STRUCT((dsModel), UserDefinedStateStatusArgs);
DS_JSON_DEFINE_STRUCT((dsModel), UpdateDeviceParameterCacheArgs);
DS_JSON_DEFINE_STRUCT((dsModel), FloorArgs);
DS_JSON_DEFINE_STRUCT((dsModel), ClusterArgs);
DS_JSON_DEFINE_STRUCT((dsModel), TemperatureSetpointArgs);
DS_JSON_DEFINE_STRUCT((dsModel), TemperatureStatusArgs);
DS_JSON_DEFINE_STRUCT((dsModel), InvokeStandardScenarioArgs);
DS_JSON_DEFINE_STRUCT((dsModel), ScenarioArgs);

namespace dsModel {

void ApartmentApi::requireApiRevisionRemovalsLessThan(int x) {
  auto apiRevisionRemovals = apiRevision().attributes.removals.value_or(0);
  DS_REQUIRE(apiRevisionRemovals < x, "Deprecated api endpont not available anymore.", apiRevisionRemovals, x);
};

std::vector<dsApi::Route> apartmentApiRoutes(ScenariosApi& api) {
  return {
      // must be before "/scenarios/{scenarioId}"
      dsApi::makeMethodRoute<InvokeStandardScenarioArgs>(
          "/scenarios/invoke", [&api](auto&& args) { api.invokeStandardScenario(std::forward<decltype(args)>(args)); }),
      dsApi::makeDocumentRoute(
          "/scenarios",
          {
              .methods = {dsApi::makeGet([&api]() { return dsModel::Scenarios{.scenarios = api.scenarios()}; }),
                          dsApi::makeCreate<Scenario>(
                              [&api](auto&& x) { return api.createScenario(std::forward<decltype(x)>(x)); })},
          }),
      dsApi::makeDocumentRoute(
          "/scenarios/{id}",
          {.methods = {dsApi::makeDeleteById<ScenarioId>([&api](auto&& x) { return api.deleteScenario(DS_FWD(x)); })},
           .patchRoutes = {{.pattern = "/name",
                            .operations = {dsApi::makeReplaceByIdPatch<dsModel::ScenarioId, std::string>(
                                [&api](auto&&... args) { api.setScenarioName(DS_FWD(args)...); })}}}}),
      dsApi::makeMethodByIdRoute<ScenarioId>("/scenarios/{id}/invoke",
                                             [&api](auto&&... args) { api.invokeScenario(DS_FWD(args)...); }),
      dsApi::makeMethodByIdRoute<ScenarioId>("/scenarios/{id}/persist",
                                             [&api](auto&&... args) { api.persistScenario(DS_FWD(args)...); }),
  };
}

namespace {
struct InstallationService : public dsApi::Service {
  InstallationService(ApartmentApi& api) {
    routes.push_back(dsApi::makeDocumentRoute("/open-metering-system/meterings",
                                              {.methods = {dsApi::makeGet([]() {
                                                 auto meterings = std::vector<OpenMeteringSystemMetering>{
                                                     {.id = OpenMeteringSystemId::electricityPowerConsumed,
                                                      .type = OpenMeteringSystemType::electricityPowerConsumed,
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {"W", "technical name", {"p1 vendor"}}},
                                                     {.id = OpenMeteringSystemId::electricityPowerProduced,
                                                      .type = OpenMeteringSystemType::electricityPowerProduced,
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {"W", "technical name", {"p1 vendor"}}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyConsumedTariff1,
                                                      .type = OpenMeteringSystemType::electricityEnergyConsumed,
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {"Wh", "technical name", {"p1 vendor"}}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyConsumedTariff2,
                                                      .type = OpenMeteringSystemType::electricityEnergyConsumed,
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {"Wh", "technical name", {"p1 vendor"}}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyProducedTariff1,
                                                      .type = OpenMeteringSystemType::electricityEnergyProduced,
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {"Wh", "technical name", {"p1 vendor"}}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyProducedTariff2,
                                                      .type = OpenMeteringSystemType::electricityEnergyProduced,
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {"Wh", "technical name", {"p1 vendor"}}}};
                                                 return OpenMeteringSystemMeterings{.meterings = meterings};
                                               })}}));

    routes.push_back(dsApi::makeDocumentRoute("/open-metering-system/meterings/values",
                                              {.methods = {dsApi::makeGet([]() {
                                                 auto values = std::vector<OpenMeteringSystemValue>{
                                                     {.id = OpenMeteringSystemId::electricityPowerConsumed,
                                                      .type = "meteringValue",
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {.value = 237.5}},
                                                     {.id = OpenMeteringSystemId::electricityPowerProduced,
                                                      .type = "meteringValue",
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {.value = 87}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyConsumedTariff1,
                                                      .type = "meteringValue",
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {.value = 222200.2}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyConsumedTariff2,
                                                      .type = "meteringValue",
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {.value = 23070.5}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyProducedTariff1,
                                                      .type = "meteringValue",
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {.value = 12212}},
                                                     {.id = OpenMeteringSystemId::electricityEnergyProducedTariff2,
                                                      .type = "meteringValue",
                                                      .lastChanged = ds::toDateTime("2021-05-20T16:42:13.777Z"),
                                                      .attributes = {.value = 124}},
                                                 };
                                                 return OpenMeteringSystemValues{
                                                     .values = values,

                                                 };
                                               })}}));

    routes.push_back(dsApi::makeDocumentRoute("/meterings/values", {.methods = {dsApi::makeGet([&api]() {
                                                                      api.requireApiRevisionRemovalsLessThan(1);
                                                                      return api.meteringValues();
                                                                    })}}));

    routes.push_back(dsApi::makeDocumentRoute("/meterings/values/{meteringId}",
                                              {.methods = {dsApi::makeGet<MeteringArgs>([&api](auto&& args) {
                                                 api.requireApiRevisionRemovalsLessThan(1);
                                                 return api.meteringValue(std::move(args));
                                               })}}));

    routes.push_back(dsApi::makeDocumentRoute("/meterings", {.methods = {dsApi::makeGet([&api]() {
                                                               api.requireApiRevisionRemovalsLessThan(1);
                                                               return api.meterings();
                                                             })}}));

    routes.push_back(dsApi::makeDocumentRoute("/meterings/{meteringId}",
                                              {.methods = {dsApi::makeGet<MeteringArgs>([&api](auto&& args) {
                                                 api.requireApiRevisionRemovalsLessThan(1);
                                                 return api.metering(args);
                                               })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/parameterCache", {.methods = {dsApi::makeGet([&api]() { return api.parameterCacheStatus(); })}}));
    routes.push_back(dsApi::makeMethodRoute<UpdateDeviceParameterCacheArgs>(
        "/parameterCache/update", [&api](auto&& args) { api.updateParameterCache(std::move(args)); }));

    routes.push_back(
        dsApi::makeDocumentRoute("", {.methods = {dsApi::makeGet([&api]() { return api.installation(); })}}));
  }
};

struct ApartmentService : public dsApi::Service {
  InstallationService installation;

  ApartmentService(ApartmentApi& api) : installation{api} {
    boost::copy(apartmentApiRoutes(api.scenarios()), std::back_inserter(routes));

    routes.push_back(dsApi::makeDocumentRoute(
        "/zones/{zoneId}/status",
        {.methods = {dsApi::makeGet<ZoneArgs>([&api](auto&& args) { return api.zoneStatus(std::move(args.zoneId)); })},
         .patchRoutes = {{.pattern = "/applications/temperature/setpoint",
                          .operations = {dsApi::makeReplacePatch<dsModel::TemperatureStatusArgs, double>(
                              [&api](auto&& args, auto&& value) {
                                api.setTemperatureStatusSetpoint(std::move(args), std::move(value));
                              })}},
                         {.pattern = "/applications/temperature/setbackRequested",
                          .operations = {dsApi::makeReplacePatch<dsModel::TemperatureStatusArgs, bool>(
                              [&api](auto&& args, auto&& value) {
                                api.setTemperatureStatusSetbackRequested(std::move(args), std::move(value));
                              })}},
                         {.pattern = "/applications/temperature/ecoPreferred",
                          .operations = {dsApi::makeReplacePatch<dsModel::TemperatureStatusArgs, bool>(
                              [&api](auto&& args, auto&& value) {
                                api.setTemperatureStatusEcoPreferred(std::move(args), std::move(value));
                              })}},
                         {.pattern = "/applications/temperature/passiveCooling",
                          .operations = {dsApi::makeReplacePatch<dsModel::TemperatureStatusArgs, bool>(
                              [&api](auto&& args, auto&& value) {
                                api.setTemperatureStatusPassiveCooling(std::move(args), std::move(value));
                              })}}}}));

    routes.push_back(dsApi::makeDocumentRoute("/zones/status",
                                              {.methods = {dsApi::makeGet([&api]() { return api.zoneStatuses(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/zones/{zoneId}",
        {.methods = {dsApi::makeGet<ZoneArgs>([&api](auto&& args) { return api.zone(std::move(args.zoneId)); })},
         .patchRoutes = {
             {.pattern = "/name",
              .operations = {dsApi::makeReplacePatch<dsModel::ZoneArgs, std::string>(
                  [&api](auto&& args, auto&& value) { api.setZoneName(std::move(args), std::move(value)); })}}}}));

    routes.push_back(
        dsApi::makeDocumentRoute("/zones", {.methods = {dsApi::makeGet([&api]() { return dsModel::Zones{.zones = api.zones()}; })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/dsDevices/{deviceId}/status",
        {.methods = {dsApi::makeGet<DeviceArgs>(
             [&api](auto&& args) { return api.deviceStatus(std::move(args.deviceId)); })},
         .patchRoutes = {
             {.pattern = "/functionBlocks/{functionBlockId}/outputs/{outputId}/value",
              .operations = {dsApi::makeReplacePatch<dsModel::OutputStatusArgs, double>(
                  [&api](auto&& args, auto&& value) { api.setOutputValue(std::move(args), std::move(value)); })}}}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/dsDevices/status", {.methods = {dsApi::makeGet([&api]() { return api.deviceStatuses(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/dsDevices/{deviceId}",
        {.methods = {dsApi::makeGet<DeviceArgs>([&api](auto&& args) { return api.device(std::move(args.deviceId)); })},
         .patchRoutes = {{.pattern = "/shadeAttributes",
                          .operations = {dsApi::makeReplacePatch<dsModel::DeviceArgs, ShadeAttributes>(
                              [&api](auto&& args, auto&& value) {
                                api.setShadeAttributes(std::move(args.deviceId), std::move(value));
                              })}}}}));

    routes.push_back(
        dsApi::makeDocumentRoute("/dsDevices", {.methods = {dsApi::makeGet([&api]() { return api.devices(); })}}));

    routes.push_back(dsApi::makeDocumentRoute("/submodules/{submoduleId}",
                                              {.methods = {dsApi::makeGet<SubmoduleArgs>([&api](auto&& args) {
                                                 return api.submodule(std::move(args.submoduleId));
                                               })}}));

    routes.push_back(
        dsApi::makeDocumentRoute("/submodules", {.methods = {dsApi::makeGet([&api]() { return api.submodules(); })}}));

    routes.push_back(dsApi::makeDocumentRoute("/functionBlocks/{functionBlockId}",
                                              {.methods = {dsApi::makeGet<FunctionBlockArgs>([&api](auto&& args) {
                                                 return api.functionBlock(std::move(args.functionBlockId));
                                               })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/functionBlocks", {.methods = {dsApi::makeGet([&api]() { return api.functionBlocks(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "", {.methods = {dsApi::makeGetWithQuery([&api](const auto& query) { return api.apartment(query); })},
             .patchRoutes = {
                 {.pattern = "/security/alarm1/name",
                  .operations = {dsApi::makeReplacePatch<dsModel::ApartmentArgs, std::string>(
                      [&api](auto&& args, auto&& value) { api.setApartmentAlarm1Name(std::move(value)); })}},
                 {.pattern = "/security/alarm2/name",
                  .operations = {dsApi::makeReplacePatch<dsModel::ApartmentArgs, std::string>(
                      [&api](auto&& args, auto&& value) { api.setApartmentAlarm2Name(std::move(value)); })}},
                 {.pattern = "/security/alarm3/name",
                  .operations = {dsApi::makeReplacePatch<dsModel::ApartmentArgs, std::string>(
                      [&api](auto&& args, auto&& value) { api.setApartmentAlarm3Name(std::move(value)); })}},
                 {.pattern = "/security/alarm4/name",
                  .operations = {dsApi::makeReplacePatch<dsModel::ApartmentArgs, std::string>(
                      [&api](auto&& args, auto&& value) { api.setApartmentAlarm4Name(std::move(value)); })}}}}));
    routes.push_back(dsApi::makeMethodRoute("/poll", [&api]() { api.pollApartment(); }));

    routes.push_back(dsApi::makeDocumentRoute(
        "/status",
        {.methods = {dsApi::makeGetWithQuery([&api](const auto& query) { return api.apartmentStatus(query); })},
         .patchRoutes = {
             {.pattern = "/temperature/ecoPreferred",
              .operations = {dsApi::makeReplacePatch<dsModel::ApartmentArgs, bool>([&api](auto&& args, auto&& value) {
                api.setApartmentTemperatureStatusEcoPreferred(std::move(args), std::move(value));
              })}}}}));
    routes.push_back(dsApi::makeMethodRoute("/status/poll", [&api]() { api.pollApartmentStatus(); }));

    routes.push_back(dsApi::makeDocumentRoute(
        "/userDefinedStates", {.methods = {dsApi::makeGet([&api]() { return api.userDefinedStates(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/userDefinedStates/{userDefinedStateId}/status",
        {.methods = {},
         .patchRoutes = {{.pattern = "/status",
                          .operations = {dsApi::makeReplacePatch<dsModel::UserDefinedStateStatusArgs, std::string>(
                              [&api](auto&& args, auto&& value) {
                                api.setUserDefinedStateStatus(std::move(args), std::move(value));
                              })}}}}));

    routes.push_back(
        dsApi::makeDocumentRoute("/floors", {.methods = {dsApi::makeGet([&api]() { return api.floors(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/floors/{floorId}",
        {.methods = {dsApi::makeGet<FloorArgs>([&api](auto&& args) { return api.floor(std::move(args)); })},
         .patchRoutes = {
             {.pattern = "/name",
              .operations = {dsApi::makeReplacePatch<dsModel::FloorArgs, std::string>(
                  [&api](auto&& args, auto&& value) { api.setFloorName(std::move(args), std::move(value)); })}}}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/applications/{applicationId}",
        {.methods = {},
         .patchRoutes = {{.pattern = "/setpoints/{temperatureLevel}/value",
                          .operations = {dsApi::makeReplacePatch<dsModel::TemperatureSetpointArgs, double>(
                              [&api](auto&& args, auto&& value) {
                                api.setTemperatureSetpoint(std::move(args), std::move(value));
                              })}}}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/clusters/status", {.methods = {dsApi::makeGet([&api]() { return api.clusterStatuses(); })}}));

    routes.push_back(dsApi::makeDocumentRoute("/clusters/{clusterId}/status",
                                              {.methods = {dsApi::makeGet<ClusterArgs>([&api](auto&& args) {
                                                 return api.clusterStatus(std::move(args.clusterId));
                                               })}}));

    routes.push_back(
        dsApi::makeDocumentRoute("/clusters", {.methods = {dsApi::makeGet([&api]() { return api.clusters(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/clusters/{clusterId}",
        {.methods = {dsApi::makeGet<ClusterArgs>([&api](auto&& args) { return api.cluster(std::move(args.clusterId)); })},
         .patchRoutes = {
             {.pattern = "/name",
              .operations = {dsApi::makeReplacePatch<dsModel::ClusterArgs, std::string>(
                  [&api](auto&& args, auto&& value) { api.setClusterName(std::move(args), std::move(value)); })}}}}));

    routes.push_back(dsApi::makeServiceRoute("/installation", installation));

    routes.push_back(
        dsApi::makeDocumentRoute("/dsServer", {.methods = {dsApi::makeGet([&api]() { return api.dsServer(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/applicationTokens", {.methods = {dsApi::makeCreate<dsModel::ApplicationToken>(
                                   [&api](auto&& x) { return api.addApplicationToken(std::move(x)); })}}));
    routes.push_back(dsApi::makeDocumentRoute(
        "/applicationTokens/{id}", {.methods = {dsApi::makeDeleteById<ApplicationTokenId>(
                                       [&api](auto&& x) { return api.revokeApplicationToken(std::move(x)); })}}));

    routes.push_back(dsApi::makeDocumentRoute("/controllers",
                                              {.methods = {dsApi::makeGet([&api]() { return api.controllers(); })}}));
    routes.push_back(dsApi::makeMethodByIdRoute<ControllerId, AuthorizeControllerArgs>(
        "/controllers/{id}/authorize", [&api](auto&&... args) { return api.authorizeController(DS_FWD(args)...); }));

    routes.push_back(dsApi::makeDocumentRoute("/apiRevision",
                                              {.methods = {dsApi::makeGet([&api]() { return api.apiRevision(); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/meterings/values", {.methods = {dsApi::makeGet([&api]() { return api.meteringValues(); })}}));
    routes.push_back(dsApi::makeDocumentRoute(
        "/meterings/values/{meteringId}", {.methods = {dsApi::makeGet<MeteringArgs>(
                                               [&api](auto&& args) { return api.meteringValue(std::move(args)); })}}));
    routes.push_back(
        dsApi::makeDocumentRoute("/meterings", {.methods = {dsApi::makeGet([&api]() { return api.meterings(); })}}));
    routes.push_back(dsApi::makeDocumentRoute(
        "/meterings/{meteringId}",
        {.methods = {dsApi::makeGet<MeteringArgs>([&api](auto&& args) { return api.metering(args); })}}));

    routes.push_back(dsApi::makeDocumentRoute(
        "/remoteConnectivity", {.methods = {dsApi::makeGet([&api]() { return api.remoteConnectivity().get(); })}}));
    routes.push_back(dsApi::makeMethodRoute<AuthorizeRemoteConnectivityArgs>(
        "/remoteConnectivity/authorize",
        [&api](auto&&... args) { return api.remoteConnectivity().authorize(DS_FWD(args)...); }));
    routes.push_back(dsApi::makeMethodRoute<PollAuthorizationRemoteConnectivityArgs>(
        "/remoteConnectivity/pollAuthorization",
        [&api](auto&&... args) { return api.remoteConnectivity().pollAuthorization(DS_FWD(args)...); }));
    routes.push_back(dsApi::makeMethodRoute("/remoteConnectivity/unauthorize", [&api](auto&&... args) {
      return api.remoteConnectivity().unauthorize(DS_FWD(args)...);
    }));
  }
};

struct DssService : public dsApi::Service {
  DssService(ApartmentApi& api) {
    routes.push_back(dsApi::makeMethodByIdRoute<ControllerId, StoreControllerAccessTokenArgs>(
        "/controllers/{id}/storeAccessToken",
        [&api](auto&&... args) { api.storeControllerAccessToken(DS_FWD(args)...); }));
  }
};
} // namespace

std::unique_ptr<dsApi::Service> makeInstallationService(ApartmentApi& api) {
  return std::make_unique<InstallationService>(api);
}
std::unique_ptr<dsApi::Service> makeApartmentService(ApartmentApi& api) {
  return std::make_unique<ApartmentService>(api);
}
std::unique_ptr<dsApi::Service> makeDssService(ApartmentApi& api) {
  return std::make_unique<DssService>(api);
}

} // namespace dsModel
