/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "buffered-message-writer.h"

#include <ds-net/message-stream-mock.h>
#include <ds/asio/test.h>

#include <boost/algorithm/string/predicate.hpp>

using namespace trompeloeil;
using boost::system::error_code;

namespace {

static const char* TAGS = "[dsNetBufferedMessageWriter][dsNet]";

struct Hooks {
public:
  MAKE_MOCK1(failed, void(const std::exception&));
  MAKE_MOCK0(enqueueWritePostponed, void());
};

SCENARIO("dsNetBufferedMessageWriter", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  auto stream = dsNet::MessageStreamMock{loop, "stream"};
  const auto config = dsNet::BufferedMessageWriterConfig{.countCapacity = 3, .sizeCapacity = 10};
  auto hooks = Hooks{};
  auto writer = makeBufferedMessageWriter(loop, stream, dsNet::BufferedMessageWriterConfig{config},
                                          {.failed = [&](const auto& x) { hooks.failed(x); },
                                           .enqueueWritePostponed = [&] { hooks.enqueueWritePostponed(); }});
  CHECK(writer->size() == 0);
  CHECK(writer->count() == 0);
  const auto error = boost::system::errc::make_error_code(boost::system::errc::operation_canceled);

  SECTION("First message is written synchronously") {
    dsNet::MessageStream::WriteHandler writeHandler;
    REQUIRE_CALL(stream, asyncWrite("aaaa", _)).LR_SIDE_EFFECT(writeHandler = std::move(_2));
    writer->enqueue("aaaa");
    CHECK(writer->size() == 0);
    CHECK(writer->count() == 0);

    SECTION("Second message is blocked") {
      {
        REQUIRE_CALL(hooks, enqueueWritePostponed());
        writer->enqueue("123456");
      }
      CHECK(writer->size() == 6);
      CHECK(writer->count() == 1);

      SECTION("Second message is written after write handler finished") {
        {
          REQUIRE_CALL(stream, asyncWrite("123456", _));
          writeHandler({});
        }
        CHECK(writer->size() == 0);
        CHECK(writer->count() == 0);
      }

      SECTION("countLimit") {
        SECTION("Message reaching count limit is queued") {
          {
            REQUIRE_CALL(hooks, enqueueWritePostponed());
            writer->enqueue("1");
          }
          {
            REQUIRE_CALL(hooks, enqueueWritePostponed());
            writer->enqueue("1");
          }
          CHECK(writer->size() == 8);
          CHECK(writer->count() == 3);

          SECTION("Message overflowing count limit is refused") {
            CHECK_THROWS_FIND(writer->enqueue("1"), "Buffer message count overflow");
          }
        }

        SECTION("sizeLimit") {
          SECTION("Message reaching size size limit is queued") {
            {
              REQUIRE_CALL(hooks, enqueueWritePostponed());
              writer->enqueue("1234");
            }
            CHECK(writer->size() == 10);
            CHECK(writer->count() == 2);
          }

          SECTION("Message overflowing size limit is refused") {
            CHECK_THROWS_FIND(writer->enqueue("12345"), "Buffer total size overflow");
          }
        }
      }
    }

    SECTION("Freed before writeHandler arrives") {
      writer.reset();
      writeHandler({});
    }
  }

  SECTION("Failed write is reported to failed hook") {
    REQUIRE_CALL(hooks, failed(_))
        .LR_SIDE_EFFECT(CHECK(boost::starts_with(std::string_view{_1.what()}, "Write failed")));
    REQUIRE_CALL(stream, asyncWrite("a", _)).LR_SIDE_EFFECT(_2(error));
    writer->enqueue("a");
  }
}

} // namespace
