#include "signalr.h"

#include <ds-net/websocket.h>
#include <ds/clock.h>
#include <ds/json.h>

#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/intrusive/list.hpp>

#include "buffered-message-writer.h"
#include "session-base.h"

static ds::LogChannel dsLogChannel("dsNetSignalr");

namespace intrusive = boost::intrusive;
using boost::system::error_code;
using namespace std::literals;

namespace dsNet {

// Signalr messages. See https://github.com/aspnet/SignalR/blob/master/specs/HubProtocol.md
namespace signalr {
struct HandshakeRequest {
  std::string protocol;
  int version;
};

enum class MessageType { invocation = 1, completion = 3, ping = 6 };

struct Message {
  int type;
  std::optional<std::string> invocationId;
  std::string target;
  std::vector<ds::JsonFragment> arguments;
  std::string error;
};

} // namespace signalr
using namespace signalr;

namespace {

struct SignalrSessionImpl final : Session<SignalrSessionImpl>, dsNet::SessionsHook, SignalrSession, SessionStateBase {
  using This = SignalrSessionImpl;
  ClosePtr<dsNet::MessageStream> m_stream;
  dsNet::ClosePtr<dsNet::BufferedMessageWriter> m_bufferedWriter;
  SignalrSessionHooks m_hooks;
  SignalrSessionImpl(Sessions& sessions, ClosePtr<dsNet::MessageStream>&& stream)
      : Session<SignalrSessionImpl>{sessions, std::string{stream->id()}},
        dsNet::SessionsHook{sessions},
        m_stream(std::move(stream)) {}

  void asyncRun(SignalrSessionHooks&& hooks) override {
    DS_INFO_ENTER(*this, state());
    setOpening();
    m_hooks = std::move(hooks);
    m_stream->asyncRun({.opened = weakCallback(&This::onStreamOpened), .closed = weakCallback(&This::onStreamClosed)});
    m_bufferedWriter =
        dsNet::makeBufferedMessageWriter(*this, *m_stream, std::move(m_hooks.bufferedMessageWriterConfig),
                                         {.failed = weakCallback(&This::onBufferedWriterFailed)});
  }

  void onBufferedWriterFailed(const std::exception& e) {
    DS_ERROR("Buffered write failed.", e, *this);
    close(boost::asio::error::operation_aborted);
  }

  void onStreamOpened() {
    DS_INFO_ENTER(*this);
    asyncReadHandshake();
  }

  void onStreamClosed(error_code ec) {
    DS_INFO_ENTER(*this);
    close(ec);
  }

  void close() override { close(boost::asio::error::operation_aborted); }

  void close(error_code ec) {
    DS_INFO_ENTER(*this, ec, state());
    assertSynchronized();
    if (isClosed()) {
      return;
    }
    setClosed();
    m_stream.reset();
    m_hooks.closed(ec);
  }

  void invokeNoReply(std::string&& target, std::vector<ds::JsonFragment>&& arguments) override {
    requireOpen();
    write(ds::toJson(Message{.type = static_cast<int>(MessageType::invocation),
                             .invocationId = {},
                             .target = std::move(target),
                             .arguments = std::move(arguments)}));
  }

  static constexpr char recordSeparator = '\x1e';
  void write(std::string&& s) {
    DS_DEBUG_ENTER(*this, s);
    // All JSON messages must be terminated by the ASCII character 0x1E (record separator).
    s.push_back(recordSeparator);
    m_bufferedWriter->enqueue(std::move(s));
  }

  std::vector<std::string> splitRecords(const std::string& message) {
    auto result = std::vector<std::string>{};
    boost::split(result, message, boost::is_any_of(std::string_view{&recordSeparator, 1}));
    if (result.size() && result.back().empty()) {
      result.pop_back();
    }
    return result;
  }

  void asyncReadHandshake() {
    DS_DEBUG_ENTER(*this, state());
    m_stream->asyncRead([this, sharedThis = shared_from_this()](error_code ec, std::string&& rawMessage) {
      try {
        DS_DEBUG_ENTER(*this, state(), ec, rawMessage);
        DS_REQUIRE(!ec, "Handshake read failed.", ec);
        auto records = splitRecords(rawMessage);
        DS_REQUIRE(records.size() == 1, "Signalr handshake message must have one record.");
        const auto handshake = ds::fromJson<HandshakeRequest>(records.at(0));
        DS_REQUIRE(handshake.protocol == "json" && handshake.version == 1, "Requested protocol is not available.",
                   handshake);
        write(ds::toJson(Message{}));

        setOpen();
        asyncReadLoop();
        m_hooks.opened();
      } catch (const std::exception& e) {
        DS_ERROR("Handshake failed.", e, *this);
        close(boost::asio::error::operation_aborted);
      }
    });
  }

  void asyncReadLoop() {
    DS_DEBUG_ENTER(*this, state());
    m_stream->asyncRead([this, sharedThis = shared_from_this()](error_code ec, std::string&& rawMessage) {
      try {
        DS_INFO_ENTER(*this, state(), ec, rawMessage);
        DS_REQUIRE(!ec, "Read failed.", ec);
        requireOpen();
        for (const auto& record : splitRecords(rawMessage)) {
          DS_DEBUG(record);
          auto message = ds::fromJson<Message>(record);
          DS_DEBUG("Read message", message);
          switch (static_cast<MessageType>(message.type)) {
            case MessageType::ping: {
              write(ds::toJson(Message{.type = static_cast<int>(MessageType::ping)}));
              break;
            }
            case MessageType::completion: {
              break;
            }
            case MessageType::invocation: {
              auto error = std::string{};
              try {
                m_hooks.invokeNoReply(std::move(message.target), std::move(message.arguments));
              } catch (const std::exception& e) {
                DS_WARNING("Invocation failed.", e, *this);
                error = e.what();
              }
              if (message.invocationId) {
                write(ds::toJson(Message{.type = static_cast<int>(MessageType::completion),
                                         .invocationId = message.invocationId,
                                         .error = std::move(error)}));
              }
              break;
            }
          }
        }
        asyncReadLoop();
      } catch (const std::exception& e) {
        DS_ERROR("Message processing failed.", e, *this);
        close(boost::asio::error::operation_aborted);
      }
    });
  }

  const ds::asio::ResourceInside& inside() override {
    using This = SignalrSessionImpl;
    static constexpr ds::asio::ResourceInside instance{
        .description = "SignalR protocol session",
        .properties = {
            ds::asio::ResourceProperty{
                .id = "state",
                .description = "Session state.",
                .value = [](const ResourceBase& x) { return ds::toJsonFragment(static_cast<const This&>(x).state()); }},
        }};
    return instance;
  }
};

struct SignalrAcceptorImpl final : ds::asio::Resource<>, SignalrAcceptor {
  std::unique_ptr<MessageStreamAcceptor> m_acceptor;
  dsNet::Sessions m_sessions{*this};
  SignalrAcceptorHooks m_hooks;

  SignalrAcceptorImpl(ds::asio::ResourceBase& parent, std::unique_ptr<MessageStreamAcceptor> acceptor)
      : ds::asio::Resource<>{parent, "signalrAcceptor"}, m_acceptor{std::move(acceptor)} {}

  void asyncRun(SignalrAcceptorHooks&& hooks) override {
    DS_INFO_ENTER(*this);
    m_hooks = std::move(hooks);
    m_acceptor->asyncRun({.accept = [this](auto&& stream) {
      DS_INFO("accept", *this, *stream);
      m_hooks.accept(makeSignalrSession(m_sessions, DS_FWD(stream)));
    }});
  }
};

} // namespace

ClosePtr<SignalrSession> makeSignalrSession(Sessions& parent, ClosePtr<dsNet::MessageStream>&& stream) {
  return std::make_shared<SignalrSessionImpl>(parent, std::move(stream));
}

std::unique_ptr<SignalrAcceptor> makeSignalrAcceptor(ds::asio::ResourceBase& parent,
                                                     std::unique_ptr<MessageStreamAcceptor> acceptor) {
  return std::make_unique<SignalrAcceptorImpl>(parent, std::move(acceptor));
}

} // namespace dsNet

DS_JSON_DECLARE_STRUCT((dsNet)(signalr), HandshakeRequest, (protocol)(version));
DS_JSON_DECLARE_STRUCT((dsNet)(signalr), Message, (type)(invocationId)(target)(arguments)(error));

DS_JSON_DEFINE_STRUCT((dsNet)(signalr), HandshakeRequest);
DS_JSON_DEFINE_STRUCT((dsNet)(signalr), Message);
