#include "websocket.h"

#include <ds/log.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/asio/connect.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/ssl.hpp>
#include <boost/beast/core.hpp>
#include <boost/beast/http.hpp>
#include <boost/beast/websocket.hpp>
#include <boost/beast/websocket/ssl.hpp>
#include <boost/range/algorithm/find_first_of.hpp>

#include "session-base.h"

static ds::LogChannel dsLogChannel("dsNetWebsocket");
// log channel to support investigation of dss-cloud communication
static ds::LogChannel messages("dsConnectorMessages");

namespace asio = boost::asio;
namespace beast = boost::beast;
namespace http = boost::beast::http;
namespace websocket = boost::beast::websocket;
using boost::system::error_code;
using tcp = asio::ip::tcp;

namespace dsNet {
namespace {

// Static polyformism using Curiously Recurring Template Pattern
// so that the same code works with both SSL streams and regular sockets.
//
// Inspired by beast/example/advanced/server-flex
template <class Derived>
struct BaseSession : Session<BaseSession<Derived>>, dsNet::SessionsHook, MessageStream, SessionStateBase {
  beast::multi_buffer m_readBuffer;
  std::string m_writeBuffer;
  MessageStreamHooks m_hooks;

  Derived &derived() { return static_cast<Derived &>(*this); }
  using Session<BaseSession<Derived>>::shared_from_this;

  BaseSession(Sessions &sessions, std::string &&id)
      : Session<BaseSession<Derived>>{sessions, std::move(id)}, dsNet::SessionsHook{sessions} {
    DS_INFO_ENTER(*this);
  }

  void asyncRun(MessageStreamHooks &&hooks) override {
    assertSynchronized();
    setOpening();
    m_hooks = std::move(hooks);
    derived().asyncRunImpl();
  }

  void asyncWrite(std::string &&buffer, std::function<void(error_code)> &&handler) override {
    assertSynchronized();
    DS_INFO_ENTER(buffer, state());
    DS_CHANNEL_INFO(messages, "Write", id(), ds::hideArg(buffer));
    // We relly on next layer to report error in non-open state
    m_writeBuffer = std::move(buffer);
    derived().ws().async_write(
        asio::buffer(m_writeBuffer.data(), m_writeBuffer.size()),
        [this, handler = std::move(handler), shared = shared_from_this()](error_code ec, std::size_t) {
          DS_INFO_ENTER(ec);
          m_writeBuffer.clear();
          if (ec) {
            close(ec);
          }
          return handler(ec);
        });
  }

  void asyncRead(std::function<void(error_code, std::string &&)> &&handler) override {
    assertSynchronized();
    DS_INFO_ENTER(state());
    // We relly on next layer to report error in non-open state
    derived().ws().async_read(
        m_readBuffer, [this, handler = std::move(handler), shared = shared_from_this()](error_code ec, std::size_t) {
          DS_INFO_ENTER(ec);
          if (ec) {
            close(ec);
            return handler(ec, {});
          }
          auto message = beast::buffers_to_string(m_readBuffer.data());
          m_readBuffer.consume(message.size());
          DS_CHANNEL_INFO(messages, "Read", id(), ds::hideArg(message));
          DS_INFO_LEAVE(message);
          return handler({}, std::move(message));
        });
  }

  void close() override { close(asio::error::operation_aborted); }

  void close(error_code ec) {
    DS_INFO_ENTER(*this, ec, state());
    assertSynchronized();
    if (isClosed()) {
      return;
    }
    setClosed();
    {
      // TODO(someday): nice websocket async_close
      error_code x;
      derived().tcpSocket().close(x);
    }
    m_hooks.closed(ec);
  }

  void open() {
    setOpen();
    m_hooks.opened();
  }
};

/// Server accept mixin
template <class Derived>
struct ServerAccept {
  Derived &derived() { return static_cast<Derived &>(*this); }
  WebsocketAcceptorHooks &m_acceptorHooks;
  ServerAccept(WebsocketAcceptorHooks &acceptorHooks) : m_acceptorHooks{acceptorHooks} {}

  void asyncWebsocketAccept() {
    DS_INFO_ENTER(derived());
    auto buffer = std::make_shared<boost::beast::flat_buffer>(4096);
    auto request = std::make_shared<http::request<http::string_body>>();
    http::async_read(derived().ws().next_layer(), *buffer, *request,
                     [this, shared = derived().shared_from_this(), buffer = buffer, request = request](
                         error_code ec, std::size_t) mutable {
                       DS_INFO_ENTER(derived(), ec);
                       if (derived().isClosed()) {
                         return;
                       }
                       if (ec) {
                         DS_ERROR("Http read failed.", derived(), ec, ec.message());
                         derived().close(ec);
                         return;
                       }
                       try {
                         m_acceptorHooks.authenticate(*request);

                         asyncWebsocketAccept2(std::move(buffer), std::move(request));
                       } catch (const std::exception &e) {
                         // TODO(someday): proper reply.
                         // TODO(someday): Merge with http_server code.
                         DS_ERROR("Websocket accept failed.", e);
                         derived().close();
                       }
                     });
  }

  void asyncWebsocketAccept2(std::shared_ptr<boost::beast::flat_buffer> &&buffer,
                             std::shared_ptr<http::request<http::string_body>> &&request) {
    DS_INFO_ENTER(derived());
    derived().ws().async_accept(*request, [this, shared = derived().shared_from_this()](error_code ec) {
      DS_INFO_ENTER(derived(), ec);
      if (derived().isClosed()) {
        return;
      }
      if (ec) {
        DS_ERROR("Websocket accept failed.", derived(), ec, ec.message());
        derived().close(ec);
        return;
      }
      derived().open();
    });
  }
};

struct WSServerSession final : BaseSession<WSServerSession>, ServerAccept<WSServerSession> {
  websocket::stream<tcp::socket> m_ws;

  WSServerSession(Sessions &sessions, tcp::socket &&tcpSocket, WebsocketAcceptorHooks &acceptorHooks)
      : BaseSession<WSServerSession>{sessions, normalizeId(ds::str(tcpSocket.remote_endpoint()))},
        ServerAccept<WSServerSession>{acceptorHooks},
        m_ws{std::move(tcpSocket)} {
    DS_INFO_ENTER(*this);
  }

  auto &ws() { return m_ws; }
  const auto &ws() const { return m_ws; }
  tcp::socket &tcpSocket() { return m_ws.next_layer(); }
  const tcp::socket &tcpSocket() const { return m_ws.next_layer(); }

  void asyncRunImpl() {
    DS_INFO_ENTER(*this);
    asyncWebsocketAccept();
  }
};

struct WSAcceptorImpl final : ds::asio::Resource<>, MessageStreamAcceptor {
  std::unique_ptr<TcpAcceptor> m_tcpAcceptor;
  WebsocketAcceptorHooks m_acceptorHooks;
  Sessions m_sessions{*this};
  MessageStreamAcceptorHooks m_hooks;

  WSAcceptorImpl(ds::asio::ResourceBase &parent, std::unique_ptr<TcpAcceptor> tcpAcceptor,
                 WebsocketAcceptorHooks &&acceptorHooks)
      : ds::asio::Resource<>{parent, "wsAcceptor"},
        m_tcpAcceptor{std::move(tcpAcceptor)},
        m_acceptorHooks{std::move(acceptorHooks)} {}

  void asyncRun(MessageStreamAcceptorHooks &&hooks) override {
    DS_INFO_ENTER(*this);
    m_hooks = std::move(hooks);
    m_tcpAcceptor->asyncRun(dsNet::TcpAcceptorHooks{.accept = [this](tcp::socket &&socket) {
      DS_INFO("accept", *this);
      m_hooks.accept(std::make_shared<WSServerSession>(m_sessions, std::move(socket), m_acceptorHooks));
    }});
  }
};

struct WSSServerSession final : BaseSession<WSSServerSession>, ServerAccept<WSSServerSession> {
  websocket::stream<asio::ssl::stream<tcp::socket>> m_ws;

  WSSServerSession(Sessions &sessions, tcp::socket &&tcpSocket, asio::ssl::context &sslContext,
                   WebsocketAcceptorHooks &acceptorHooks)
      : BaseSession<WSSServerSession>{sessions, normalizeId(ds::str(tcpSocket.remote_endpoint()))},
        ServerAccept<WSSServerSession>{acceptorHooks},
        m_ws{std::move(tcpSocket), sslContext} {}

  auto &ws() { return m_ws; }
  const auto &ws() const { return m_ws; }
  auto &sslStream() { return m_ws.next_layer(); }
  const auto &sslStream() const { return m_ws.next_layer(); }
  tcp::socket &tcpSocket() { return m_ws.next_layer().next_layer(); }
  const tcp::socket &tcpSocket() const { return m_ws.next_layer().next_layer(); }

  void asyncRunImpl() {
    DS_INFO_ENTER(*this);
    asyncSSLHandshake();
  }

  void asyncSSLHandshake() {
    DS_INFO_ENTER(*this);
    sslStream().async_handshake(asio::ssl::stream_base::server, [this, shared = shared_from_this()](error_code ec) {
      DS_INFO_ENTER(ec);
      if (isClosed()) {
        return;
      }
      if (ec) {
        DS_ERROR("SSL handshake failed.", ec, *this);
        return close(ec);
      }
      asyncWebsocketAccept();
    });
  }
};

struct WSSAcceptorImpl final : ds::asio::Resource<>, MessageStreamAcceptor {
  std::unique_ptr<TcpAcceptor> m_tcpAcceptor;
  asio::ssl::context &m_sslContext;
  WebsocketAcceptorHooks m_acceptorHooks;
  Sessions m_sessions{*this};
  MessageStreamAcceptorHooks m_hooks;

  WSSAcceptorImpl(ds::asio::ResourceBase &parent, std::unique_ptr<TcpAcceptor> tcpAcceptor,
                  asio::ssl::context &sslContext, WebsocketAcceptorHooks &&acceptorHooks)
      : ds::asio::Resource<>{parent, "wssAcceptor"},
        m_tcpAcceptor{std::move(tcpAcceptor)},
        m_sslContext{sslContext},
        m_acceptorHooks{std::move(acceptorHooks)} {}

  void asyncRun(MessageStreamAcceptorHooks &&hooks) override {
    DS_INFO_ENTER(*this);
    m_hooks = std::move(hooks);
    m_tcpAcceptor->asyncRun(dsNet::TcpAcceptorHooks{.accept = [this](tcp::socket &&socket) {
      DS_INFO("accept", *this);
      m_hooks.accept(std::make_shared<WSSServerSession>(m_sessions, std::move(socket), m_sslContext, m_acceptorHooks));
    }});
  }
};

struct WSSClientSession final : BaseSession<WSSClientSession> {
  asio::ssl::context &m_sslContext;
  const WSSUrl m_url;
  tcp::resolver m_resolver{io()};
  websocket::stream<asio::ssl::stream<tcp::socket>> m_ws{io(), m_sslContext};

  WSSClientSession(Sessions &sessions, asio::ssl::context &sslContext, const std::string &url)
      : BaseSession<WSSClientSession>{sessions, normalizeId(url)}, m_sslContext(sslContext), m_url(parseWSSUrl(url)) {}

  auto &ws() { return m_ws; }
  const auto &ws() const { return m_ws; }
  auto &sslStream() { return m_ws.next_layer(); }
  const auto &sslStream() const { return m_ws.next_layer(); }
  tcp::socket &tcpSocket() { return m_ws.next_layer().next_layer(); }
  const tcp::socket &tcpSocket() const { return m_ws.next_layer().next_layer(); }
  const auto &url() { return m_url; }

  void asyncRunImpl() {
    DS_INFO_ENTER(*this);
    asyncResolve();
  }

  void asyncResolve() {
    DS_INFO_ENTER(m_url);
    m_resolver.async_resolve({m_url.host, m_url.port},
                             [this, shared = shared_from_this()](error_code ec, tcp::resolver::iterator result) {
                               DS_INFO_ENTER(ec);
                               if (isClosed()) {
                                 return;
                               }
                               if (ec) {
                                 DS_ERROR("Resolve failed.", *this, ec, m_url);
                                 return close(ec);
                               }
                               return asyncConnect(std::move(result));
                             });
  }

  void asyncConnect(tcp::resolver::iterator &&result) {
    DS_INFO_ENTER();
    asio::async_connect(tcpSocket(), result,
                        [this, result, shared = shared_from_this()](error_code ec, tcp::resolver::iterator) {
                          DS_INFO_ENTER(ec);
                          if (isClosed()) {
                            return;
                          }
                          if (ec) {
                            DS_ERROR("Connect failed.", ec, *this, m_url);
                            return close(ec);
                          }
                          asyncSSLHandshake();
                        });
  }

  void asyncSSLHandshake() {
    DS_INFO_ENTER();
    sslStream().async_handshake(asio::ssl::stream_base::client, [this, shared = shared_from_this()](error_code ec) {
      DS_INFO_ENTER(ec);
      if (isClosed()) {
        return;
      }
      if (ec) {
        DS_ERROR("SSL handshake failed.", *this, ec);
        return close(ec);
      }
      asyncWebsocketHandshake();
    });
  }

  // client websocket handshake, opem
  void asyncWebsocketHandshake() {
    const auto &url = m_url;
    DS_INFO_ENTER(*this, url);
    ws().async_handshake(url.host, url.target, [this, shared = shared_from_this()](error_code ec) {
      DS_INFO_ENTER(ec);
      if (isClosed()) {
        return;
      }
      if (ec) {
        DS_ERROR("Ws handshake failed.", *this, ec.message());
        return close(ec);
      }
      open();
    });
  }
};

class WSSStreamClient final : public ds::asio::Resource<WSSClientConfig>, public MessageStreamClient {
  std::string m_url;
  asio::ssl::context &m_sslContext;
  Sessions m_sessions{*this};

public:
  WSSStreamClient(ds::asio::ResourceBase &parent, WSSClientConfig &&defaultConfig, asio::ssl::context &sslContext)
      : ds::asio::Resource<WSSClientConfig>{parent, "wssClient", std::move(defaultConfig)},
        m_url{std::string{config().url}},
        m_sslContext(sslContext) {}

  ClosePtr<MessageStream> asyncConnect() override {
    assertSynchronized();
    DS_DEBUG_ENTER();
    return std::make_shared<WSSClientSession>(m_sessions, m_sslContext, m_url);
  }

  std::string url() const override {
    assertSynchronized();
    return m_url;
  }

  std::string defaultUrl() const override { return config().url; }

  void setUrl(std::string &&url) override {
    DS_NOTICE("Set url", url, *this);
    assertSynchronized();
    m_url = std::move(url);
  }
};

} // namespace

WSSUrl parseWSSUrl(gsl::cstring_span url) {
  WSSUrl result;
  auto s = url;

  const auto protocol = gsl::cstring_span{"wss://"};
  DS_REQUIRE(boost::starts_with(s, protocol), "Failed to parse wss url (missing wss://)", url);
  s = {s.begin() + protocol.size(), s.end()};
  const auto hostEnd = boost::find_first_of(s, ":/");
  DS_REQUIRE(hostEnd != s.end(), "Failed to parse wss url (missing ':'  or '/')", url);
  result.host = {s.begin(), hostEnd};

  if (*hostEnd == ':') {
    auto portEnd = boost::find_first_of(s, "/");
    DS_REQUIRE(portEnd != s.end(), "Failed to parse wss url (missing `/` after `:`)", url);
    result.port = {hostEnd + 1, portEnd};
    s = {portEnd, s.end()};
  } else {
    result.port = "443";
    s = {hostEnd, s.end()};
  }

  result.target = gsl::to_string(s);
  return result;
}

asio::ssl::context systemClientSslContext() {
  auto result = asio::ssl::context{asio::ssl::context::sslv23_client};
  result.set_default_verify_paths();
  result.set_verify_mode(asio::ssl::verify_peer); // default is asio::ssl::verify_none !!!
  return result;
}

std::unique_ptr<MessageStreamClient> makeWSSClient(ds::asio::ResourceBase &parent, WSSClientConfig &&config,
                                                   asio::ssl::context &sslContext) {
  return std::make_unique<WSSStreamClient>(parent, std::move(config), sslContext);
}

std::unique_ptr<MessageStreamAcceptor> makeWSAcceptor(ds::asio::ResourceBase &parent,
                                                      std::unique_ptr<TcpAcceptor> tcpAcceptor,
                                                      WebsocketAcceptorHooks &&hooks) {
  return std::make_unique<WSAcceptorImpl>(parent, std::move(tcpAcceptor), std::move(hooks));
}

std::unique_ptr<MessageStreamAcceptor> makeWSSAcceptor(ds::asio::ResourceBase &parent,
                                                       std::unique_ptr<TcpAcceptor> tcpAcceptor,
                                                       asio::ssl::context &sslContext, WebsocketAcceptorHooks &&hooks) {
  return std::make_unique<WSSAcceptorImpl>(parent, std::move(tcpAcceptor), sslContext, std::move(hooks));
}

} // namespace dsNet

DS_JSON_DEFINE_STRUCT((dsNet), WSSUrl);
DS_JSON_DEFINE_STRUCT((dsNet), WSSClientConfig);
