/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/asio/ds-signal.h>
#include <ds/asio/resource.h>
#include <ds/box.h>

namespace ds {

/// Property publishes value and notifies changes.
///
/// Owner (with mutable `&`) publishes changes with `set`.
///
/// Clients (with `const &`) query and watch value with thead safe `get`, `changed`.
///
/// T is wrapped in ds::Box to provide immutability, fast copy.
template <class T>
class Property : boost::noncopyable {
public:
  Property(ds::asio::ResourceBase& parent, T value) : m_parent{parent}, m_value{std::move(value)} {}

  /// Query value.
  ///
  /// Thread safe.
  ds::Box<T> get() const {
    // TODO(someday): make `m_value` atomic? Protect m_value with mutex?
    return m_parent.invoke([&] { return m_value; });
  }

  /// Set value.
  ///
  /// Caller must be synchronized to resource thread.
  void set(ds::Box<T> value) {
    m_parent.assertSynchronized();
    if (m_value == value) {
      return;
    }
    m_value = std::move(value);
    m_changed.emit(m_value);
  }

  /// Value was changed
  ds::Slots<ds::Box<T>>& changed() { return m_changed; }

private:
  ds::asio::ResourceBase& m_parent;
  ds::Box<T> m_value;
  ds::Signal<ds::Box<T>> m_changed;
};

} // namespace ds
