/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <chrono>

namespace ds {

/// Steady mockable timer.
///
/// Nearly all code wants to use steady clock, so `steady` is skipped from the name.
///
/// Supports mocking by `ClockMocker` (see timer-mock.h).
///
/// Design rationale: The class fulfils chrono timer concept
/// and is usable e.g. with asio timers.
/// But concepts leave no place for mocking by virtual functions.
/// So mocking is achieved by presence of ClockMocker class instance.
class Clock {
public:
  typedef std::chrono::steady_clock::duration duration;
  typedef duration::rep rep;
  typedef duration::period period;
  typedef std::chrono::time_point<Clock> time_point;

  // Mocked clock is not steady.
  // Without mocking (in production), the clock is steady.
  // Would we gain much if we made the clock steady in production?
  static constexpr bool is_steady = false;

  // Static is required by chrono::clock concept.
  static time_point now() noexcept;

private:
  friend class ClockMocker;
  friend struct ClockWaitTraits;
};

struct ClockWaitTraits {
  static Clock::duration to_wait_duration(const Clock::duration& d);
};

} // namespace ds
