#pragma once

#include <gsl.h>
#include <boost/noncopyable.hpp>
#include <iosfwd>
#include <type_traits>
#include <utility>
#include <vector>

namespace ds {

#ifdef __GNUC__
/// Branch prediction macros.  Evaluates to the condition given, but also tells the compiler that we
/// expect the condition to be true/false enough of the time that it's worth hard-coding branch
/// prediction.
#define DS_LIKELY(condition) __builtin_expect(bool(condition), true)
#define DS_UNLIKELY(condition) __builtin_expect(bool(condition), false)

#else

#define DS_LIKELY(condition) (bool(condition))
#define DS_UNLIKELY(condition) (bool(condition))

#endif

/// Macro expanding to `,`. Useful to pass comma to macro without starting next argument
#define DS_COMMA ,

#define DS_CONCAT_(x, y) x##y

/// Concatenate two identifiers into one
#define DS_CONCAT(x, y) DS_CONCAT_(x, y)

/// Create a unique identifier name.
/// We use concatenate __LINE__ rather than __COUNTER__ so that
/// the name can be used multiple times in the same macro
/// and evaluate the same.
/// If the code relies on DS_UNIQUE_NAME generating the same
/// name multiple times, wrap the code into macro to force
/// compilation into one line.
#define DS_UNIQUE_NAME(prefix) DS_CONCAT(prefix, __LINE__)

// Run the given code when current scope exits, whether by leave, return or exception.
#define DS_DEFER(code) auto DS_UNIQUE_NAME(_dsDefer) = ::ds::defer([&]() { code; })

namespace detail {
template <typename Func>
class Deferred : boost::noncopyable {
public:
  Deferred(Func&& func) : m_func(std::forward<Func>(func)), m_canceled(false) {}
  ~Deferred() {
    if (!m_canceled)
      m_func();
  }
  Deferred(Deferred&& x) : m_func(std::move(x.m_func)), m_canceled(false) { x.m_canceled = true; }

private:
  Func m_func;
  bool m_canceled;
};
} // namespace detail

// Returns an object which will invoke the given functor in its destructor.  The object is not
// copyable but is movable with the semantics you'd expect.  Since the return type is private,
// you need to assign to an `auto` variable.
//
// The DS_DEFER macro provides slightly more convenient syntax for the common case where you
// want some code to run at current scope exit.
template <typename Func>
detail::Deferred<Func> defer(Func&& func) {
  return detail::Deferred<Func>(std::forward<Func>(func));
}

// Version of KJ_REQUIRE() to be used inside inline methods.
// Does not include the whole logging framework.
#define DS_IREQUIRE(condition, message)              \
  [](bool condition_, const std::string& message_) { \
    if (DS_LIKELY(condition_)) {                     \
    } else {                                         \
      throw std::runtime_error(message_);            \
    }                                                \
  }((condition), (message))

#define DS_FAIL_IREQUIRE(message) throw std::runtime_error(message)

/// Size of one dimensional array.
///
/// `ds::arraySize(x)` is easier to read than `std::extend<decltype(x)>::value`
template <unsigned N = 0, class T>
constexpr size_t arraySize(T&) {
  return std::extent<T, N>::value;
}

template <class T, class = void>
struct IsRange : std::false_type {};

template <class T>
struct IsRange<T, typename std::void_t<decltype(std::declval<T>().begin()), decltype(std::declval<T>().end())>>
    : std::true_type {};

/// False in dependent context. Useful for if constexpr together with static_assert
/// From https://en.cppreference.com/w/cpp/utility/variant/visit
template <class T>
struct DependentFalse : std::false_type {};

/// Builds visitor from inheriting lambas.
/// From https://en.cppreference.com/w/cpp/utility/variant/visit
///
///    std::visit(::ds::Visitor {
///      [](auto arg) { std::cout << arg << ' '; },
///      [](double arg) { std::cout << std::fixed << arg << ' '; },
///      [](const std::string& arg) { std::cout << std::quoted(arg) << ' '; },
///    }, v);
template <class... Ts>
struct Visitor : Ts... {
  using Ts::operator()...;
};
template <class... Ts>
Visitor(Ts...)->Visitor<Ts...>;

/// Shortcut for `std::forward<decltype(a)>(a)` where a needs to be repeated twice
/// and is hard to read compared to `DS_FWD(a)`
///
/// See https://github.com/SuperV1234/vrm_core/issues/1 for discussion
/// regarding __VA_ARGS__ in this context.
///
/// Origin: https://vittorioromeo.info/index/blog/capturing_perfectly_forwarded_objects_in_lambdas.html
#define DS_FWD(...) ::std::forward<decltype(__VA_ARGS__)>(__VA_ARGS__)

} // namespace ds
