/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

// Requirements:
// - readonly key - value storage.
//   - key is string id, `camelCase` for consistency with api guideline.
//   - value is json value
// - possible to override in device. This override is preserved during update.
// - default is in source code
//
// Config file `/etc/ds-conf/{key}.json` holds device override.
// All config files (`/etc/ds-conf/*json`) files are preserved during update.

#include <optional>
#include <boost/filesystem/operations.hpp>
#include "json.h"
#include "log.h"
#include "system.h"

namespace ds {

class Conf {
public:
  virtual ~Conf() = default;

  /// Read conf entry
  ///
  /// @param key conf key id
  /// @param defaultValue default value if no device specific conf entry is found
  template <typename T>
  T readOr(std::string_view key, T&& defaultValue) const;

  // implementation details

  virtual std::optional<JsonFragment> readJsonIfExists(std::string_view key) const = 0;
  virtual void writeJson(std::string_view key, const std::string &json) const = 0;

protected:
  static LogChannel& logChannel();
};

std::unique_ptr<Conf> makeConf(System& system);

// Implementation details

template <typename T>
T Conf::readOr(std::string_view key, T&& defaultValue) const {
  try {
    if (const auto& json = readJsonIfExists(key)) {
      return fromJson<T>(*json);
    }
  } catch (const std::exception& e) {
    DS_CHANNEL_ERROR(logChannel(), "Failed to parse sysconf value.", key, e);
  }
  return std::move(defaultValue);
}

} // namespace ds
