/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

// Introduces reflection for enums and handy template conversion functions
// using this reflection.
//
// Example use:
//
// 1. define the reflected enum in global namespace with macro
//
//   DS_REFLECTED_ENUM((ds), Transport, (car)(plane));
//
//   or with enum option initializers
//
//   DS_REFLECTED_ENUM_WITH_INITIALIZERS((ds), Transport, ((car)(0))((plane)(2)));
//
// DS_REFLECTED_ENUM(namespaces, enumType, enumOptions):
// - namespaces: sequence of namespaces for enum type definition
// - enumType: enum type inside namespace
// - options: sequence of enum options
//
// 2. use
//
//   ds::tryFromEnum<const char*>(Transport:CAR) -> std::optional<const char*> as "car", "plane", none
//   ds::tryFromEnum<std::string>(Transport:CAR) -> std::optional<std::string> as "car", "plane", none
//   ds::tryFromEnum<int>(Transport:CAR) -> std::optional<int> as 0, 1, none
//
//   // Use when the code is determined to use the value whether the option is known or not.
//   ds::fromEnum<std::string>(Transport:CAR) -> std::string as "car", "plane", "2"
//
//   ds::tryToEnum<Transport>("car") -> std::optional<Transport>
//   ds::tryToEnum<Transport>(1) -> std::optional<Transport>
//
//   // Use when the code is determined to use the value whether the option is known or not.
//   ds::toEnum<Transport>("car") -> Transport::CAR
//   ds::toEnum<Transport>("5") -> static_cast<Transport>(5)
//   ds::toEnum<Transport>("unknown") -> Transport()
//
//   // Use when the code wants to fail on unknown option
//   ds::requireEnum<Transport>(1) -> Transport as Transport::CAR, Transport::PLANE or throws
//
//   ds::isValidEnum(Transport::CAR) -> bool
//
//   std::cout << Transport:CAR -> serializes as "car", "plane", "2"
//
// 3. Enum to enum conversion
//
// DS_REFLECTED_ENUM((ds)(enumTest), AB, (a)(b));
// DS_REFLECTED_ENUM((ds)(enumTest), BC, (b)(c));
//
// template <>
// struct DSToEnum<ds::enumTest::BC, ds::enumTest::AB, void> {
//   static std::optional<ds::enumTest::BC> tryTo(ds::enumTest::AB x) {
//     switch (x) {
//       case ds::enumTest::AB::a: return {};
//       case ds::enumTest::AB::b: return ds::enumTest::BC::b;
//     }
//     return {};
//   }
// };
//
// ds::tryToEnum<ds::enumTest::BC>(ds::enumTest::AB::b) -> ds::enumTest::BC::b
// ds::toEnum<ds::enumTest::BC>(ds::enumTest::AB::b)  -> ds::enumTest::BC::b
// ds::requireEnum<ds::enumTest::BC>(ds::enumTest::AB::b) -> ds::enumTest::BC::b
//
//
// 1b. DEPRECATED: define the enum, reflection and operator ostream<<
//
//   enum class Transport { CAR, PLANE };
//   ds::Enum<Transport> dsEnum(Transport) {
//     static constexpr std::pair<Transport, const char*> x[] = {
//             {Transport::CAR, "car"}, {Transport::PLANE, "plane"},
//     };
//     return {gsl::make_span(x), "ds::Transport"};
//   }
//   inline std::ostream& operator<<(std::ostream& stream, Transport x) { return stream << ds::fromEnum<std::string>(x);
//   }
//
// NOTE: `dsEnum` and `operator ostream <<` must be declared
// in the same namespace as enum to let it just work everywhere.
//
// The `dsEnum` argument value is ignored,
// the argument only triggers C++ argument dependent lookup.
//

#include "enum-core.h"
#include "log.h"
#include "reflect.h"

namespace ds {
/// Log channel for enum conversion errors.
/// TODO(someday): attach error to client provided context
LogChannel& enumLogChannel();

template <typename R, typename T>
auto tryToEnum(const T& x) {
  return ::DSToEnum<R, std::decay_t<T>>::tryTo(x);
}

template <typename R, typename T>
R toEnum(const T& value) {
  if (auto out = ds::tryToEnum<R>(value)) {
    return std::move(*out);
  }
  DS_CHANNEL_DEBUG(enumLogChannel(), "Unknown", ds::hideArg(dsEnum(R()).name), "enum option.", value);
  return {};
}

template <typename R, typename T>
R requireEnum(const T& value) {
  if (auto out = ds::tryToEnum<R>(value)) {
    return *out;
  }
  DS_FAIL_REQUIRE("Unknown", ds::hideArg(dsEnum(R()).name), "enum option.", value);
}

template <typename R, typename T>
std::optional<R> tryFromEnum(T x) {
  return ::DSFromEnum<R, std::decay_t<T>>::tryFrom(x);
}

template <typename R, typename T>
R fromEnum(T value) {
  if (auto out = ds::tryFromEnum<R>(value)) {
    return std::move(*out);
  }
  DS_CHANNEL_DEBUG(enumLogChannel(), "Unknown", ds::hideArg(dsEnum(T()).name), "enum value",
                   ds::hideArg(static_cast<std::underlying_type_t<T>>(value)));
  return {};
}

template <typename T>
bool isValidEnum(T x) {
  return bool(tryFromEnum<const char*>(x));
}

} // namespace ds

#define DS_REFLECTED_ENUM_OPTION(r, data, elem) elem,
#define DS_REFLECTED_ENUM_OPTION_DATA(r, TYPE, elem) {TYPE::elem, BOOST_PP_STRINGIZE(elem)},

#define DS_REFLECTED_ENUM(NAME_SPACES, TYPE, OPTIONS)                                                                \
  BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_OPEN_NAMESPACE, _, NAME_SPACES)                                                   \
  enum class TYPE { BOOST_PP_SEQ_FOR_EACH(DS_REFLECTED_ENUM_OPTION, _, OPTIONS) };                                   \
  inline ds::Enum<TYPE> dsEnum(TYPE) {                                                                               \
    static constexpr std::pair<TYPE, const char*> x[] = {                                                            \
        BOOST_PP_SEQ_FOR_EACH(DS_REFLECTED_ENUM_OPTION_DATA, TYPE, OPTIONS)};                                        \
    return {gsl::make_span(x),                                                                                       \
            BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_NAMESPACE_STRINGIZE, _, NAME_SPACES) "::" BOOST_PP_STRINGIZE(TYPE)};    \
  }                                                                                                                  \
  inline std::ostream& operator<<(std::ostream& ostream, TYPE x) { return ostream << ds::fromEnum<std::string>(x); } \
  BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_CLOSE_NAMESPACE, _, NAME_SPACES)

#define DS_REFLECTED_ANONYMOUS_ENUM(TYPE, OPTIONS)                                                                   \
  namespace {                                                                                                        \
  enum class TYPE { BOOST_PP_SEQ_FOR_EACH(DS_REFLECTED_ENUM_OPTION, _, OPTIONS) };                                   \
  inline ds::Enum<TYPE> dsEnum(TYPE) {                                                                               \
    static constexpr std::pair<TYPE, const char*> x[] = {                                                            \
        BOOST_PP_SEQ_FOR_EACH(DS_REFLECTED_ENUM_OPTION_DATA, TYPE, OPTIONS)};                                        \
    return {gsl::make_span(x), BOOST_PP_STRINGIZE(TYPE)};                                                            \
  }                                                                                                                  \
  inline std::ostream& operator<<(std::ostream& ostream, TYPE x) { return ostream << ds::fromEnum<std::string>(x); } \
  }

#define DS_REFLECTED_ENUM_WITH_INITIALIZERS_OPTION(r, data, elem) \
  BOOST_PP_SEQ_ELEM(0, elem) = BOOST_PP_SEQ_ELEM(1, elem),
#define DS_REFLECTED_ENUM_WITH_INITIALIZERS_OPTION_DATA(r, TYPE, elem) \
  {TYPE::BOOST_PP_SEQ_ELEM(0, elem), BOOST_PP_STRINGIZE(BOOST_PP_SEQ_ELEM(0, elem))},

#define DS_REFLECTED_ENUM_WITH_INITIALIZERS(NAME_SPACES, TYPE, OPTIONS)                                              \
  static_assert(BOOST_PP_SEQ_ELEM(1, BOOST_PP_SEQ_ELEM(0, OPTIONS)) == 0,                                            \
                "First enum option must by initiliazed to 0");                                                       \
  BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_OPEN_NAMESPACE, _, NAME_SPACES)                                                   \
  enum class TYPE { BOOST_PP_SEQ_FOR_EACH(DS_REFLECTED_ENUM_WITH_INITIALIZERS_OPTION, _, OPTIONS) };                 \
  inline ds::Enum<TYPE> dsEnum(TYPE) {                                                                               \
    static constexpr std::pair<TYPE, const char*> x[] = {                                                            \
        BOOST_PP_SEQ_FOR_EACH(DS_REFLECTED_ENUM_WITH_INITIALIZERS_OPTION_DATA, TYPE, OPTIONS)};                      \
    return {gsl::make_span(x),                                                                                       \
            BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_NAMESPACE_STRINGIZE, _, NAME_SPACES) "::" BOOST_PP_STRINGIZE(TYPE)};    \
  }                                                                                                                  \
  inline std::ostream& operator<<(std::ostream& ostream, TYPE x) { return ostream << ds::fromEnum<std::string>(x); } \
  BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_CLOSE_NAMESPACE, _, NAME_SPACES)
