/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "tracing.h"
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/log.h>
#include <thread>
#include <trompeloeil.hpp>

static const char* TAGS = "[dsTracing][ds]";

namespace {

using namespace trompeloeil;
using namespace std::literals;

struct TracerFinished {
public:
  MAKE_MOCK1(call, void(ds::Trace&&));
};

SCENARIO("dsTracer", TAGS) {
  auto clock = ds::ClockMocker{};

  SECTION("onFinished") {
    auto finishedMock = TracerFinished{};
    auto finished = [&](auto&& x) { finishedMock.call(std::move(x)); };

    SECTION("0s duration, 0s offset") {
      REQUIRE_CALL(finishedMock, call(ds::Trace{.name = "test"}));
      auto tracer = ds::Tracer("test", {}, finished);
    }

    SECTION("2s duration, 1s offset") {
      REQUIRE_CALL(finishedMock, call(ds::Trace{.name = "test", .parentOffset = 1s, .duration = 2s}));

      auto tracer = ds::Tracer("test", 1s, finished);
      clock.advance(2s);
    }

    SECTION("children since 500ms and 800ms") {
      REQUIRE_CALL(finishedMock,
                   call(ds::Trace{.name = "test",
                                  .parentOffset = {},
                                  .duration = 1s,
                                  .children = {{.name = "child2", .parentOffset = 800ms, .duration = 200ms},
                                               {.name = "child1", .parentOffset = 500ms, .duration = 500ms}}}));

      auto tracer = ds::Tracer("test", {}, finished);
      clock.advance(500ms);
      auto childTracer1 = ds::Tracer("child1", tracer);
      clock.advance(300ms);
      auto childTracer2 = ds::Tracer("child2", &tracer);
      auto childTracer3 = ds::Tracer("childWithoutParent", nullptr);
      clock.advance(200ms);
    }

    SECTION("duration()") {
      auto tracer = ds::Tracer("test");
      clock.advance(2s);
      CHECK(tracer.duration() == 2s);
    }
  }

  SECTION("Trace can be consumed by consume ") {
    auto tracer = ds::Tracer("test", 1s);
    clock.advance(2s);
    CHECK(tracer.consume() == ds::Trace{.name = "test", .parentOffset = 1s, .duration = 2s});
  }
}

SCENARIO("dsThreadLocalTracer", TAGS) {
  auto clock = ds::ClockMocker{};
  auto tracer = ds::ThreadLocalTracer{"test"};

  SECTION("accumulates ThreadLocalTracer children") {
    {
      clock.advance(400ms);
      auto childTracer1 = ds::ThreadLocalTracer{"child1"};
      clock.advance(600ms);
    }
    CHECK(tracer.consume() == ds::Trace{.name = "test",
                                        .parentOffset = {},
                                        .duration = 1s,
                                        .children = {{.name = "child1", .parentOffset = 400ms, .duration = 600ms}}});
  }

  SECTION("works thread locally") {
    auto thread = std::thread{[&] {
      clock.advance(400ms);
      auto childTracer1 = ds::ThreadLocalTracer{"child1"};
      clock.advance(600ms);
    }};
    thread.join();
    CHECK(tracer.consume() == ds::Trace{.name = "test", .parentOffset = {}, .duration = 1s});
  }
}

} // namespace
