/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "tracing.h"
#include "log.h"

namespace ds {

Tracer::Tracer(std::string &&name, Tracer &parent) : Tracer{std::move(name), &parent} {}

Tracer::Tracer(std::string &&name, Tracer *maybeParent)
    : Tracer{std::move(name), maybeParent ? maybeParent->duration() : ds::Clock::duration{},
             maybeParent ? std::function<void(Trace &&)>{[maybeParent](Trace &&trace) {
               // TODO(someday): synchronization? Detect dead reference to `this`?
               maybeParent->addChild(std::move(trace));
             }}
                         : std::function<void(Trace &&)>{}} {}

Tracer::Tracer(std::string &&name, ds::Clock::duration parentOffset, std::function<void(Trace &&)> onFinished)
    : m_trace{.name = std::move(name), .parentOffset = parentOffset}, m_onFinished{std::move(onFinished)} {}

Tracer::~Tracer() {
  if (m_onFinished) {
    auto onFinished = std::move(m_onFinished);
    m_onFinished = {};
    onFinished(consume());
  }
}

ds::Clock::duration Tracer::duration() const {
  return ds::Clock::now() - m_startedAt;
}

Trace Tracer::consume() {
  DS_REQUIRE(!m_onFinished, "Cannot consume when onFinished is waiting for result.");
  m_trace.duration = duration();
  return std::move(m_trace);
}

void Tracer::addChild(Trace &&childTrace) {
  m_trace.children.push_back(std::move(childTrace));
}

static thread_local ThreadLocalTracer *s_topThreadLocalTracer;

ThreadLocalTracer::ThreadLocalTracer(std::string &&name) : Tracer{std::move(name), s_topThreadLocalTracer} {
  push();
}

ThreadLocalTracer::ThreadLocalTracer(std::string &&name, Tracer &parent) : Tracer{std::move(name), parent} {
  push();
}

ThreadLocalTracer::ThreadLocalTracer(std::string &&name, Tracer *maybeParent) : Tracer{std::move(name), maybeParent} {
  push();
}

ThreadLocalTracer::ThreadLocalTracer(std::string &&name, ds::Clock::duration parentOffset,
                                     std::function<void(Trace &&)> onFinished)
    : Tracer{std::move(name), parentOffset, onFinished} {
  push();
}

ThreadLocalTracer::~ThreadLocalTracer() {
  // pop
  DS_ASSERT(s_topThreadLocalTracer == this, "");
  s_topThreadLocalTracer = m_next;
}

void ThreadLocalTracer::push() {
  m_next = s_topThreadLocalTracer;
  s_topThreadLocalTracer = this;
}

} // namespace ds
