/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#include "ds485-device.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ds/locale/ds-locale.h>
#include <ds/log.h>
#include <bitset>

#include "cache/apartment-proxy.h"
#include "ds485/dsbusinterface.h"
#include "dss.h"
#include "logger.h"
#include "vdc-messages.pb.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/modelconst.h"
#include "model/modelevent.h"
#include "model/modelmaintenance.h"
#include "model/dsmeter.h"
#include "model/set.h"

#define TASK_REQUE_ON_FAILURE_SLEEP_SECONDS 4

static ds::LogChannel dsLogChannel("deviceBusInterface");

namespace dss {

//================================================== DSDeviceBusInterface

// default timeout for synchronous calls to dsm api
static const int kDSM_API_TIMEOUT = 8;

uint8_t DSDeviceBusInterface::getDeviceConfig(const Device& device, uint8_t class_, uint8_t index) {
  DS_CONTEXT(device, class_, index);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint8_t retVal;

  int ret = DeviceConfig_get_sync_8(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, kDSM_API_TIMEOUT, &retVal);
  DSBusInterface::checkResultCode(ret);

  DS_NOTICE("Config.", retVal);

  /* The set config call that originates from us triggers a ds485 callback,
   * but the get call  does not, so we intercept it here directly */
  if (DSS::hasInstance()) {
    DSS::getInstance()->getApartment().proxy().updateDeviceConfig(device.getDSID(), class_, index, retVal);
  }
  return retVal;
} // getDeviceConfig

uint16_t DSDeviceBusInterface::getDeviceConfigWord(const Device& device, uint8_t class_, uint8_t index) {
  DS_CONTEXT(device, class_, index);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t retVal;

  int ret = DeviceConfig_get_sync_16(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, kDSM_API_TIMEOUT, &retVal);
  DSBusInterface::checkResultCode(ret);

  DS_NOTICE("Config.", retVal);
  return retVal;
} // getDeviceConfigWord

void DSDeviceBusInterface::setDeviceConfig(const Device& device, uint8_t class_, uint8_t index, uint8_t value) {
  DS_CONTEXT(device, class_, index, value);
  std::shared_ptr<DSMeter> pMeter = DSS::getInstance()->getApartment().getDSMeterByDSID(device.getDSMeterDSID());

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set config.");
  int ret;
  if (pMeter->getApiVersion() >= 0x301) {
    ret = DeviceConfig_set_sync(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, value, 30);
  } else {
    ret = DeviceConfig_set(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), class_, index, value);
  }
  DSBusInterface::checkResultCode(ret);
} // setDeviceConfig

void DSDeviceBusInterface::setDeviceButtonActiveGroup(const Device& device, uint8_t group) {
  DS_CONTEXT(device, group);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set button active group.");
  int ret = DeviceProperties_set_button_active_group(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group);
  DSBusInterface::checkResultCode(ret);
} // setDeviceButtonActiveGroup

void DSDeviceBusInterface::setDeviceProgMode(const Device& device, uint8_t mode) {
  DS_CONTEXT(device, mode);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set prog mode.");
  int ret;
  if (mode) {
    ret = DeviceActionRequest_action_programming_mode_on(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress());
  } else {
    ret = DeviceActionRequest_action_programming_mode_off(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress());
  }
  DSBusInterface::checkResultCode(ret);
} // setDeviceProgMode

void DSDeviceBusInterface::setValue(const Device& device, uint8_t value) {
  DS_CONTEXT(device, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set value.");
  int ret =
      DeviceActionRequest_action_set_outval(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), value);
  DSBusInterface::checkResultCode(ret);

  if (m_pModelMaintenance) {
    ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceSetOutputValue, device.getDSMeterDSID());
    pEvent->addParameter(device.getShortAddress());
    pEvent->addParameter(value);
    m_pModelMaintenance->addModelEvent(pEvent);
  }
} // setValue

uint16_t DSDeviceBusInterface::getSensorValue(const Device& device, const int sensorIndex) {
  DS_CONTEXT(device, sensorIndex);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  uint16_t retVal;

  int ret = DeviceSensor_get_value_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), sensorIndex, kDSM_API_TIMEOUT, &retVal);
  DSBusInterface::checkResultCode(ret);

  ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceSensorValue, device.getDSMeterDSID());
  pEvent->addParameter(device.getShortAddress());
  pEvent->addParameter(sensorIndex);
  pEvent->addParameter(retVal);
  if (DSS::hasInstance()) {
    DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
  } else {
    delete pEvent;
  }

  DS_NOTICE("Sensor value.", retVal);
  return retVal;
} // getSensorValue

DeviceSensorValue_t DSDeviceBusInterface::getSensorValueEx(const Device& device, const int sensorIndex) {
  DS_CONTEXT(device, sensorIndex);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  uint8_t contextMsg[64];
  DeviceSensorValue_t retVal;

  int ret = DeviceSensor_get_value_extended_sync(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(),
      sensorIndex, kDSM_API_TIMEOUT, &retVal.value, &retVal.valueAge, &retVal.contextId, contextMsg);
  DSBusInterface::checkResultCode(ret);

  retVal.contextMsg = std::string((const char*)(contextMsg));
  retVal.timestamp = DateTime().addMilliSeconds((int)retVal.valueAge * (-1));

  ModelEventWithSensorEx* pEvent = new ModelEventWithSensorEx();
  pEvent->m_deviceDSID = device.getDSMeterDSID();
  pEvent->addParameter(device.getShortAddress());
  pEvent->addParameter(sensorIndex);
  pEvent->addParameter(retVal.valueAge);
  pEvent->addParameter(retVal.contextId);
  pEvent->setSingleStringParameter(retVal.contextMsg);
  pEvent->m_sensorValue = retVal.value;

  if (DSS::hasInstance()) {
    DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
  } else {
    delete pEvent;
  }

  DS_NOTICE("Sensor value.", retVal);
  return retVal;
} // getSensorValueEx

void DSDeviceBusInterface::addGroup(const Device& device, const int group) {
  DS_CONTEXT(device, group);
  std::shared_ptr<DSMeter> pMeter = DSS::getInstance()->getApartment().getDSMeterByDSID(device.getDSMeterDSID());

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }
  DS_NOTICE("Add group.");
  int ret = 0;
  if (pMeter->getApiVersion() >= 0x301) {
    ret = DeviceGroupMembershipModify_add_sync(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group, 30);
  } else {
    ret = DeviceGroupMembershipModify_add(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group);
  }
  if (ret == ERROR_WRONG_PARAMETER) {
    Logger::getInstance()->log("addGroup: dsm-api returned wrong parameter", lsWarning);
  } else {
    DSBusInterface::checkResultCode(ret);
  }
} // addGroup

void DSDeviceBusInterface::removeGroup(const Device& device, const int group) {
  DS_CONTEXT(device, group);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Remove group.");
  int ret =
      DeviceGroupMembershipModify_remove(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), group);
  if (ret == ERROR_WRONG_PARAMETER) {
    Logger::getInstance()->log("addGroup: dsm-api returned wrong parameter", lsWarning);
  } else {
    DSBusInterface::checkResultCode(ret);
  }
} // removeGroup

void DSDeviceBusInterface::lockOrUnlockDevice(const Device& device, const bool locked) {
  DS_CONTEXT(device, locked);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set locked.");
  int ret = DeviceProperties_set_locked_flag(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), locked);
  DSBusInterface::checkResultCode(ret);
} // lockOrUnlockDevice

std::pair<uint8_t, uint16_t> DSDeviceBusInterface::getTransmissionQuality(const Device& device) {
  DS_CONTEXT(device);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint8_t downstream;
  uint16_t upstream;

  int ret = TestTransmissionQuality_get_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), kDSM_API_TIMEOUT, &downstream, &upstream);
  DSBusInterface::checkResultCode(ret);

  const auto result = std::make_pair(downstream, upstream);
  DS_NOTICE("Transmission quality.", result);
  return result;
}

void DSDeviceBusInterface::increaseDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Increase output channel value.");
  int ret =
      DeviceActionRequest_action_opc_inc(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel);
  DSBusInterface::checkResultCode(ret);
} // increaseDeviceOutputChannelValue

void DSDeviceBusInterface::decreaseDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Decrease output channel value.");
  int ret =
      DeviceActionRequest_action_opc_dec(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel);
  DSBusInterface::checkResultCode(ret);
} // decreaseDeviceOutputChannelValue

void DSDeviceBusInterface::stopDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Stop output channel value.");
  int ret =
      DeviceActionRequest_action_opc_stop(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel);
  DSBusInterface::checkResultCode(ret);
} // stopDeviceOutputChannelValue

uint16_t DSDeviceBusInterface::getDeviceOutputChannelValue(const Device& device, uint8_t channel) {
  DS_CONTEXT(device, channel);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
    ;
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_current_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, 2 * kDSM_API_TIMEOUT, &out);
  DSBusInterface::checkResultCode(ret);
  DS_NOTICE("Output channel value.", out);
  return out;
} // getDeviceOutputChannelValue

void DSDeviceBusInterface::setDeviceOutputChannelValue(
    const Device& device, uint8_t channel, uint8_t size, uint16_t value, bool applyNow) {
  DS_CONTEXT(device, channel, size, value, applyNow);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  int ret;

  DS_NOTICE("Set output channel value.");
  if (applyNow) {
    ret = DeviceOPCConfig_set_current_and_apply(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, size, value);
  } else {
    ret = DeviceOPCConfig_set_current(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, size, value);
  }
  DSBusInterface::checkResultCode(ret);
} // setDeviceOutputChannelValue

uint16_t DSDeviceBusInterface::getDeviceOutputChannelSceneValue(const Device& device, uint8_t channel, uint8_t scene) {
  DS_CONTEXT(device, channel, scene);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_scene_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, scene, 2 * kDSM_API_TIMEOUT, &out);
  DSBusInterface::checkResultCode(ret);
  DS_NOTICE("Output channel scene value.", out);
  return out;
} // getDeviceOutputChannelSceneValue

void DSDeviceBusInterface::setDeviceOutputChannelSceneValue(
    const Device& device, uint8_t channel, uint8_t size, uint8_t scene, uint16_t value) {
  DS_CONTEXT(device, channel, size, scene, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set output channel scene value.");
  int ret = DeviceOPCConfig_set_scene(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), channel, size, scene, value);
  DSBusInterface::checkResultCode(ret);
} // setDeviceOutputChannelSceneValue

uint16_t DSDeviceBusInterface::getDeviceOutputChannelSceneConfig(const Device& device, uint8_t scene) {
  DS_CONTEXT(device, scene);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_scene_config_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, 2 * kDSM_API_TIMEOUT, &out);
  DSBusInterface::checkResultCode(ret);
  DS_NOTICE("Output channel scene config.", out);
  return out;
} // getDeviceOutputChannelSceneConfig

void DSDeviceBusInterface::setDeviceOutputChannelSceneConfig(const Device& device, uint8_t scene, uint16_t value) {
  DS_CONTEXT(device, scene, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set output channel scene config.");
  int ret =
      DeviceOPCConfig_set_scene_config(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, value);
  DSBusInterface::checkResultCode(ret);
} // setDeviceOutputChannelSceneConfig

void DSDeviceBusInterface::setDeviceOutputChannelDontCareFlags(const Device& device, uint8_t scene, uint16_t value) {
  DS_CONTEXT(device, scene, value);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set output channel don't care.");
  int ret =
      DeviceOPCConfig_set_dc_flags(m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, value);
  DSBusInterface::checkResultCode(ret);
}

uint16_t DSDeviceBusInterface::getDeviceOutputChannelDontCareFlags(const Device& device, uint8_t scene) {
  DS_CONTEXT(device, scene);
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t out = 0;

  int ret = DeviceOPCConfig_get_dc_flags_sync(
      m_DSMApiHandle, device.getDSMeterDSID(), device.getShortAddress(), scene, 2 * kDSM_API_TIMEOUT, &out);
  DS_NOTICE("Output channel don't care.", ret);
  DSBusInterface::checkResultCode(ret);
  return out;
}

void DSDeviceBusInterface::genericRequest(const Device& device, const std::string& methodName,
    const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& params, const std::locale& locale) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  vdcapi::Message message;
  message.set_type(vdcapi::VDSM_REQUEST_GENERIC_REQUEST);
  auto genericRequest = message.mutable_vdsm_request_generic_request();
  genericRequest->set_dsuid(dsuid2str(device.getDSID()));
  genericRequest->set_methodname(methodName);
  *genericRequest->mutable_params() = params;
  uint8_t arrayOut[REQUEST_LEN];
  if (!message.SerializeToArray(arrayOut, sizeof(arrayOut))) {
    throw std::runtime_error("SerializeToArray failed");
  }
  uint8_t arrayIn[RESPONSE_LEN];
  uint16_t arrayInSize;
  int ret = UserProtobufMessageRequest(
      m_DSMApiHandle, device.getDSMeterDSID(), message.ByteSizeLong(), arrayOut, &arrayInSize, arrayIn);
  DSBusInterface::checkResultCode(ret);
  if (!message.ParseFromArray(arrayIn, arrayInSize)) {
    throw std::runtime_error("ParseFromArray failed");
  }
  if (message.type() != vdcapi::GENERIC_RESPONSE) {
    throw std::runtime_error("Invalid vdc response");
  }
  const vdcapi::GenericResponse& response = message.generic_response();
  if (response.code() != vdcapi::ERR_OK) {
    auto type = [&] {
      switch (response.errortype()) {
        case vdcapi::ERROR_TYPE_FAILED:
          return ds::Exception::Type::FAILED;
        case vdcapi::ERROR_TYPE_OVERLOADED:
          return ds::Exception::Type::OVERLOADED;
        case vdcapi::ERROR_TYPE_DISCONNECTED:
          return ds::Exception::Type::DISCONNECTED;
        case vdcapi::ERROR_TYPE_UNIMPLEMENTED:
          return ds::Exception::Type::UNIMPLEMENTED;
      }
      DS_ERROR("Unrecognized errorType", static_cast<int>(response.errortype()));
      return ds::Exception::Type::FAILED;
    }();
    throw ds::Exception(type, response.description(),
        dsDtr(locale, device.vdcTrDomain(), "error", response.usermessagetobetranslated()));
  }
}

void DSDeviceBusInterface::setProperty(
    const Device& device, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& properties) {
  return setVdcProperty(device.getDSID(), device.getDSMeterDSID(), properties);
}

vdcapi::Message DSDeviceBusInterface::getProperty(
    const Device& device, const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& query) {
  return getVdcProperty(device.getDSID(), device.getDSMeterDSID(), query);
}

DSDeviceBusInterface::ConfigReaderHelper::ConfigReaderHelper(const std::string& busConnection)
    : m_busConnection(busConnection), m_dsmApiHandle(NULL), m_deviceAdress(0), m_dsmId(), m_revisionID(0) {}

DSDeviceBusInterface::ConfigReaderHelper::~ConfigReaderHelper() {
  if (m_dsmApiHandle != NULL) {
    DsmApiClose(m_dsmApiHandle);
    DsmApiCleanup(m_dsmApiHandle);
    m_dsmApiHandle = NULL;
  }
}

void DSDeviceBusInterface::ConfigReaderHelper::setup(std::shared_ptr<Device> device) {
  m_deviceAdress = device->getShortAddress();
  m_dsmId = device->getDSMeterDSID();
  m_revisionID = device->getRevisionID();
  if (DSS::hasInstance()) {
    m_dsm = DSS::getInstance()->getApartment().getDSMeterByDSID(m_dsmId);
  }
}

uint16_t DSDeviceBusInterface::ConfigReaderHelper::getDeviceConfigWord(uint8_t class_, uint8_t index) const {
  if (m_dsmApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  // wait until dsm has finished device registrations
  while (m_dsm && m_dsm->isPresent() && (m_dsm->getState() == DSM_STATE_REGISTRATION)) {
    sleep(5);
  }
  if ((m_dsm == NULL) || !m_dsm->isPresent()) {
    throw ItemNotFoundException("OEMDataReader: dsm is not present");
  }

  uint16_t retVal;
  int ret = DeviceConfig_get_sync_16(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  if (ret == ERROR_SYNC_RESPONSE_TIMEOUT || ret == ERROR_TIMEOUT) {
    ret = DeviceConfig_get_sync_16(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  }
  DSBusInterface::checkResultCode(ret);

  return retVal;
}

uint8_t DSDeviceBusInterface::ConfigReaderHelper::getDeviceConfig(uint8_t class_, uint8_t index) const {
  if (m_dsmApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  // wait until dsm has finished device registrations
  while (m_dsm && m_dsm->isPresent() && (m_dsm->getState() == DSM_STATE_REGISTRATION)) {
    sleep(5);
  }
  if ((m_dsm == NULL) || !m_dsm->isPresent()) {
    throw ItemNotFoundException("OEMDataReader: dsm is not present");
  }

  uint8_t retVal;
  int ret = DeviceConfig_get_sync_8(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  if (ret == ERROR_SYNC_RESPONSE_TIMEOUT || ret == ERROR_TIMEOUT) {
    ret = DeviceConfig_get_sync_8(m_dsmApiHandle, m_dsmId, m_deviceAdress, class_, index, 15, &retVal);
  }
  DSBusInterface::checkResultCode(ret);

  return retVal;
}

DSDeviceBusInterface::OEMDataReader::OEMDataReader(const std::string& busConnection)
    : ConfigReaderHelper(busConnection) {}

DSDeviceBusInterface::OEMDataReader::~OEMDataReader() {}

void DSDeviceBusInterface::OEMDataReader::run() {
  uint64_t ean = 0;
  uint16_t serialNumber = 0;
  uint8_t partNumber = 0;
  bool isVisible = false;
  bool isIndependent = false;
  bool isConfigLocked = false;
  uint8_t pairedDevices = 0;
  DeviceOEMState_t state = DEVICE_OEM_UNKNOWN;
  DeviceOEMInetState_t deviceInetState = DEVICE_OEM_EAN_NO_EAN_CONFIGURED;

  m_dsmApiHandle = DsmApiInitialize();
  if (!m_dsmApiHandle) {
    throw std::runtime_error("OEMDataReader: Unable to get dsmapi handle");
  }

  int result = DsmApiOpen(m_dsmApiHandle, m_busConnection.c_str(), 0);
  if (result < 0) {
    throw std::runtime_error(std::string("OEMDataReader: Unable to open connection to: ") + m_busConnection);
  }

  try {
    // read device-active data and EAN part number
    uint16_t result0;
    result0 = getDeviceConfigWord(1, 0x1e);
    partNumber = result0 & 0x7F;
    isIndependent = (result0 & 0x80);
    uint8_t upper = result0 >> 8;
    isVisible = !(upper & (1 << 4));
    pairedDevices = (upper & 0x0f);

    // optimize readout time: avoid switching banks
    serialNumber = getDeviceConfigWord(1, 0x1c);

    // check if EAN is programmed: Bank 3: 0x2e-0x2f 0x0000 < x < 0xffff
    uint16_t result;
    result = getDeviceConfigWord(3, 0x2e);
    deviceInetState = (DeviceOEMInetState_t)(result >> 12);

    if (deviceInetState == DEVICE_OEM_EAN_NO_EAN_CONFIGURED) {
      // no EAN programmed
      state = DEVICE_OEM_NONE;
    } else {
      // EAN programmed
      ean |= ((long long unsigned int)(result & 0xFFF) << 32);

      result = getDeviceConfigWord(3, 0x2a);
      ean |= result;

      result = getDeviceConfigWord(3, 0x2c);
      ean |= ((long long unsigned int)result << 16);

      state = DEVICE_OEM_VALID;
    }

    if (m_revisionID >= TBVersion_OemConfigLock) {
      uint8_t deviceActive;
      deviceActive = getDeviceConfig(3, 0x1f);
      if (std::bitset<8>(deviceActive).test(0)) {
        isConfigLocked = true;
      }
    }

    ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceEANReady, m_dsmId);
    pEvent->addParameter(m_deviceAdress);
    pEvent->addParameter(state);
    pEvent->addParameter(deviceInetState);
    pEvent->addParameter(ean >> 32);
    pEvent->addParameter(ean & 0xFFFFFFFF);
    pEvent->addParameter(serialNumber);
    pEvent->addParameter(partNumber);
    pEvent->addParameter(isIndependent);
    pEvent->addParameter(isConfigLocked);
    pEvent->addParameter(pairedDevices);
    pEvent->addParameter(isVisible);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
    } else {
      delete pEvent;
    }
  } catch (ItemNotFoundException& infe) {
    Logger::getInstance()->log("OEMDataReader::run: dSM not found or left, aborting readout\n");
    if (DSS::hasInstance()) {
      if (m_dsm) {
        std::shared_ptr<Device> dev = m_dsm->getDevices().getByBusID(m_deviceAdress, m_dsm).getDevice();
        dev->setOemInfoState(DEVICE_OEM_UNKNOWN);
        DSS::getInstance()->getModelMaintenance().addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
      }
    }
  } catch (const std::exception& er) {
    Logger::getInstance()->log(std::string("OEMDataReader::run: bus error: ") + er.what() + " reading from dSM " +
                                   dsuid2str(m_dsmId) + " DeviceId " + intToString(m_deviceAdress) +
                                   ", requeueing task", lsWarning);

    if (!DSS::hasInstance() || !m_dsm) {
      return;
    }

    DsmApiClose(m_dsmApiHandle);
    DsmApiCleanup(m_dsmApiHandle);
    m_dsmApiHandle = NULL;

    try {
      std::shared_ptr<Device> dev = m_dsm->getDevices().getByBusID(m_deviceAdress, m_dsm).getDevice();
      DS_REQUIRE(dev->isPresent(), "Device was physically removed");
      sleep(TASK_REQUE_ON_FAILURE_SLEEP_SECONDS);
      DSS::getInstance()->getModelMaintenance().scheduleDeviceReadout(
          m_dsmId, std::shared_ptr<DSDeviceBusInterface::OEMDataReader>(shared_from_this()));
    } catch (const std::exception& e) {
      DS_ERROR("OEMDataReader::run: bus error: ", e, " reading from dSM ", m_dsmId,
                " DeviceId ", m_deviceAdress);
    }
  }
}

DSDeviceBusInterface::TNYConfigReader::TNYConfigReader(const std::string& busConnection)
    : ConfigReaderHelper(busConnection) {}

DSDeviceBusInterface::TNYConfigReader::~TNYConfigReader() {}

void DSDeviceBusInterface::TNYConfigReader::run() {
  bool isVisible = false;
  uint8_t pairedDevices = 0;

  m_dsmApiHandle = DsmApiInitialize();
  if (!m_dsmApiHandle) {
    throw std::runtime_error("TNYConfigReader: Unable to get dsmapi handle");
  }

  int result = DsmApiOpen(m_dsmApiHandle, m_busConnection.c_str(), 0);
  if (result < 0) {
    throw std::runtime_error(std::string("TNYConfigReader: Unable to open connection to: ") + m_busConnection);
  }

  try {
    // read device-active data and pairing data
    uint8_t cfg;
    cfg = getDeviceConfig(1, CfgFunction_DeviceActive);
    isVisible = !(cfg & (1 << 4));
    pairedDevices = (cfg & 0x0f);

    ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceDataReady, m_dsmId);
    pEvent->addParameter(m_deviceAdress);
    pEvent->addParameter(pairedDevices);
    pEvent->addParameter(isVisible);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
    } else {
      delete pEvent;
    }
  } catch (ItemNotFoundException& infe) {
    Logger::getInstance()->log("TNYConfigReader::run: dSM not found or left, aborting readout\n");
  } catch (std::runtime_error& er) {
    // Bus error
    Logger::getInstance()->log(std::string("TNYConfigReader::run: bus error: ") + er.what() + " reading from dSM " +
                                   dsuid2str(m_dsmId) + " DeviceId " + intToString(m_deviceAdress) +
                                   ", requeueing task",
        lsWarning);

    if (!DSS::hasInstance() || !m_dsm) {
      return;
    }

    DsmApiClose(m_dsmApiHandle);
    DsmApiCleanup(m_dsmApiHandle);
    m_dsmApiHandle = NULL;

    try {
      std::shared_ptr<Device> dev = m_dsm->getDevices().getByBusID(999, m_dsm).getDevice();
      DS_REQUIRE(dev->isPresent(), "Device was physically removed");
      sleep(TASK_REQUE_ON_FAILURE_SLEEP_SECONDS);
      DSS::getInstance()->getModelMaintenance().scheduleDeviceReadout(
          m_dsmId, std::shared_ptr<DSDeviceBusInterface::TNYConfigReader>(shared_from_this()));
    } catch (const std::exception& e) {
      DS_ERROR("OEMDataReader::run: bus error: ", e, " reading from dSM ", m_dsmId,
                " DeviceId ", m_deviceAdress, ", requeueing task");
    }
  }
}

void DSDeviceBusInterface::reregister(const Device& device) {
  DS_CONTEXT(device);
  std::shared_ptr<DSMeter> pMeter = DSS::getInstance()->getApartment().getDSMeterByDSID(device.getDSMeterDSID());

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Reregister.");
  int ret;
  ret = DeviceConfig_trigger_readout(
        m_DSMApiHandle, device.getDSMeterDSID(), device.getDSID());
  DSBusInterface::checkResultCode(ret);
} // reregister


} // namespace dss
