/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dsactionrequest.h"

#include "dsbusinterface.h"
#include "dss.h"
#include "logger.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/group.h"
#include "model/set.h"

static ds::LogChannel dsLogChannel("ds485Action");

using namespace std::literals;

namespace dss {

//================================================== DSActionRequest

void DSActionRequest::callScene(AddressableModelItem* pTarget, const uint16_t _scene, Caller&& caller,
                                const bool _force) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Call scene.", _scene, *pGroup, _force);
    DSS::getInstance()->assertBlockingAllowed();
    if (_force) {
      ret = ZoneGroupActionRequest_action_force_call_scene(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(),
                                                           pGroup->getID(), 0, _scene);
    } else {
      ret = ZoneGroupActionRequest_action_call_scene(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(),
                                                     pGroup->getID(), 0, _scene);
    }
    DSBusInterface::checkBroadcastResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onGroupCallScene(NULL, DSUID_NULL, pGroup->getZoneID(), pGroup->getID(), 0, _scene,
                                        std::move(caller), _force);
    }
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Call scene.", _scene, *pDevice, _force);
    if (_force) {
      ret = DeviceActionRequest_action_force_call_scene(m_DSMApiHandle, pDevice->getDSMeterDSID(),
                                                        pDevice->getShortAddress(), _scene);
    } else {
      ret = DeviceActionRequest_action_call_scene(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress(),
                                                  _scene);
    }
    DSBusInterface::checkResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onDeviceCallScene(NULL, pDevice->getDSMeterDSID(), pDevice->getShortAddress(), 0, _scene,
                                         std::move(caller), _force);
    }
  }
}

void DSActionRequest::callSceneMin(AddressableModelItem* pTarget, const uint16_t _scene, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Call scene.", _scene, *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_call_scene_min(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(),
                                                       pGroup->getID(), 0, _scene);
    DSBusInterface::checkBroadcastResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onGroupCallSceneMin(NULL, DSUID_NULL, pGroup->getZoneID(), pGroup->getID(), 0, _scene,
                                           std::move(caller));
    }
    // TODO extend dsm api, dsm and vdsm to support command
  }
}

void DSActionRequest::saveScene(AddressableModelItem* pTarget, const uint16_t _scene, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Save scene.", _scene, *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_save_scene(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(),
                                                   pGroup->getID(), 0, _scene);
    DSBusInterface::checkBroadcastResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onGroupSaveScene(NULL, DSUID_NULL, pGroup->getZoneID(), pGroup->getID(), 0, _scene,
                                        std::move(caller));
    }
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Save scene.", _scene, *pDevice);
    ret = DeviceActionRequest_action_save_scene(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress(),
                                                _scene);
    DSBusInterface::checkResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onDeviceSaveScene(NULL, pDevice->getDSMeterDSID(), pDevice->getShortAddress(), 0, _scene,
                                         std::move(caller));
    }
  }
}

void DSActionRequest::increaseOutputChannelValue(AddressableModelItem* pTarget, ChannelType _channel, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Increase opc.", _channel, *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_opc_inc(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(), pGroup->getID(),
                                                0, static_cast<int>(_channel));
    DSBusInterface::checkBroadcastResultCode(ret);
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Increase opc.", _channel, *pDevice);
    ret = DeviceActionRequest_action_opc_inc(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress(),
                                             static_cast<int>(_channel));
    DSBusInterface::checkResultCode(ret);
  }
}
void DSActionRequest::decreaseOutputChannelValue(AddressableModelItem* pTarget, ChannelType _channel, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Decrease opc.", _channel, *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_opc_dec(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(), pGroup->getID(),
                                                0, static_cast<int>(_channel));
    DSBusInterface::checkBroadcastResultCode(ret);
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Decrease opc.", _channel, *pDevice);
    ret = DeviceActionRequest_action_opc_dec(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress(),
                                             static_cast<int>(_channel));
    DSBusInterface::checkResultCode(ret);
  }
}
void DSActionRequest::stopOutputChannelValue(AddressableModelItem* pTarget, ChannelType _channel, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Stop opc.", _channel, *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_opc_stop(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(), pGroup->getID(),
                                                 0, static_cast<int>(_channel));
    DSBusInterface::checkBroadcastResultCode(ret);
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Stop opc.", _channel, *pDevice);
    ret = DeviceActionRequest_action_opc_stop(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress(),
                                              static_cast<int>(_channel));
    DSBusInterface::checkResultCode(ret);
  }
}

void DSActionRequest::undoScene(AddressableModelItem* pTarget, const uint16_t _scene, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Undo scene.", _scene, *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_undo_scene_number(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(),
                                                          pGroup->getID(), 0, _scene);
    DSBusInterface::checkBroadcastResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onGroupUndoScene(NULL, DSUID_NULL, pGroup->getZoneID(), pGroup->getID(), 0, _scene, true,
                                        std::move(caller));
    }
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Undo scene.", _scene, *pDevice);
    ret = DeviceActionRequest_action_undo_scene_number(m_DSMApiHandle, pDevice->getDSMeterDSID(),
                                                       pDevice->getShortAddress(), _scene);
    DSBusInterface::checkResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onDeviceUndoScene(NULL, pDevice->getDSMeterDSID(), pDevice->getShortAddress(), 0,
                                         _scene, true, std::move(caller));
    }
  }
}

void DSActionRequest::undoSceneLast(AddressableModelItem* pTarget, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Undo scene.", *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_undo_scene(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(),
                                                   pGroup->getID(), 0);
    DSBusInterface::checkBroadcastResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onGroupUndoScene(NULL, DSUID_NULL, pGroup->getZoneID(), pGroup->getID(), 0, -1, false,
                                        std::move(caller));
    }
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Undo scene.", *pDevice);
    ret = DeviceActionRequest_action_undo_scene(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress());
    DSBusInterface::checkResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onDeviceUndoScene(NULL, pDevice->getDSMeterDSID(), pDevice->getShortAddress(), 0, -1, false,
                                         std::move(caller));
    }
  }
}

void DSActionRequest::blink(AddressableModelItem* pTarget, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Blink.", *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_blink(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(), pGroup->getID(),
                                              gsl::narrow<uint16_t>(caller.origin));
    DSBusInterface::checkBroadcastResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onGroupBlink(NULL, DSUID_NULL, pGroup->getZoneID(), pGroup->getID(), 0, std::move(caller));
    }
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Blink.", *pDevice);
    ret = DeviceActionRequest_action_blink(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress());
    DSBusInterface::checkResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onDeviceBlink(NULL, pDevice->getDSMeterDSID(), pDevice->getShortAddress(), 0, std::move(caller));
    }
  }
}

void DSActionRequest::setValue(AddressableModelItem* pTarget, const uint8_t _value, Caller&& caller) {
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    DS_NOTICE("Set value.", _value, *pGroup);
    DSS::getInstance()->assertBlockingAllowed();
    ret = ZoneGroupActionRequest_action_set_outval(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(),
                                                   pGroup->getID(), 0, _value);
    DSBusInterface::checkBroadcastResultCode(ret);
    m_pBusEventSink->onZoneGroupSetValue(NULL, pGroup->getZoneID(), pGroup->getID(), _value, std::move(caller));
  }
  if (const auto& pDevice = dynamic_cast<Device*>(pTarget)) {
    DS_NOTICE("Set value.", _value, *pDevice);
    ret = DeviceActionRequest_action_set_outval(m_DSMApiHandle, pDevice->getDSMeterDSID(), pDevice->getShortAddress(),
                                                _value);
    DSBusInterface::checkResultCode(ret);
    if (m_pBusEventSink) {
      m_pBusEventSink->onDeviceSetValue(NULL, pDevice->getDSMeterDSID(), pDevice->getShortAddress(), _value,
                                        std::move(caller));
    }
  }
}

void DSActionRequest::pushSensor(AddressableModelItem* pTarget, dsuid_t _sourceID, SensorType _sensorType,
                                 double _sensorValueFloat, Caller&& caller) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  if (const auto& pGroup = dynamic_cast<Group*>(pTarget)) {
    uint16_t convertedSensorValue = doubleToSensorValue(_sensorType, _sensorValueFloat);
    uint8_t precisionvalue = sensorTypeToPrecision(_sensorType);

    DS_NOTICE("Push sensor.", convertedSensorValue, precisionvalue, *pGroup);
    if (static_cast<int>(_sensorType) <= SensorMaxIdForBus) {
      DSS::getInstance()->assertBlockingAllowed();
      int ret = ZoneGroupSensorPush(m_DSMApiHandle, DSUID_BROADCAST, pGroup->getZoneID(), pGroup->getID(), _sourceID,
                                    static_cast<uint8_t>(_sensorType), convertedSensorValue, precisionvalue);
      DSBusInterface::checkBroadcastResultCode(ret);
    }

    if (m_pBusEventSink) {
      m_pBusEventSink->onZoneSensorValue(NULL, DSUID_NULL, _sourceID, pGroup->getZoneID(), pGroup->getID(), _sensorType,
                                         convertedSensorValue, precisionvalue, std::move(caller));
    }
  }
} // pushSensor

void DSActionRequest::muteFire() {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Mute fire");
  DSS::getInstance()->assertBlockingAllowed();
  auto ret = ZoneGroupActionRequest_action_fire_mute(m_DSMApiHandle, DSUID_BROADCAST, 0, 0, 0);
  DSBusInterface::checkBroadcastResultCode(ret);
}

void DSActionRequest::setBusEventSink(BusEventSink* _eventSink) {
  m_pBusEventSink = _eventSink;
}

bool DSActionRequest::isOperationLock(const dsuid_t& _dSM, int _clusterId) {
  uint8_t lockState;
  int ret;

  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return false;
  }

  ret = ClusterProperties_get_operation_lock(m_DSMApiHandle, _dSM, _clusterId, &lockState);
  DSBusInterface::checkResultCode(ret);
  return (lockState == 1);
}

void DSActionRequest::setState(Group* group, const uint8_t zoneState, const uint8_t areaState1,
                               const uint8_t areaState2, const uint8_t areaState3, const uint8_t areaState4) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiPrioHandleMutex);
  if (m_DSMApiPrioHandle == NULL) {
    return;
  }

  DS_NOTICE("Set Zone State", group->getZoneID(), group->getID(), zoneState, areaState1, areaState2, areaState3,
            areaState4);
  DSS::getInstance()->assertBlockingAllowed();
  auto ret = ZoneGroupProperties_set_state(m_DSMApiPrioHandle, DSUID_BROADCAST, group->getZoneID(), group->getID(),
                                           zoneState, areaState1, areaState2, areaState3, areaState4);
  DSBusInterface::checkBroadcastResultCode(ret);
}

void DSActionRequest::setState(Cluster* cluster, const uint8_t state) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("Set Cluster State", cluster->getID(), state);
  DSS::getInstance()->assertBlockingAllowed();
  auto ret = ClusterProperties_set_state(m_DSMApiHandle, DSUID_BROADCAST, cluster->getID(), state);
  DSBusInterface::checkBroadcastResultCode(ret);
}

void DSActionRequest::stopStopTimer(Group* group, const uint8_t zoneOrArea) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("StopStopTimer", group->getZoneID(), group->getID(), zoneOrArea);
  DSS::getInstance()->assertBlockingAllowed();
  auto ret = ZoneGroupProperties_stop_stop_timer(m_DSMApiHandle, DSUID_BROADCAST, group->getZoneID(), group->getID(),
                                                 zoneOrArea);
  DSBusInterface::checkBroadcastResultCode(ret);
}

void DSActionRequest::stopStopTimer(Cluster* cluster) {
  boost::recursive_mutex::scoped_lock lock(m_DSMApiHandleMutex);
  if (m_DSMApiHandle == NULL) {
    return;
  }

  DS_NOTICE("StopStopTimer", cluster->getID());
  DSS::getInstance()->assertBlockingAllowed();
  auto ret = ClusterProperties_stop_stop_timer(m_DSMApiHandle, DSUID_BROADCAST, cluster->getID());
  DSBusInterface::checkBroadcastResultCode(ret);
}
} // namespace dss
