/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hal-device.h"

#include <chrono>
#include <cmath>

#include "base.h"
#include "ds-api/route-segments.h"
#include "ds-model.h"
#include "ds/log.h"
#include "dss.h"
#include "model/device.h"
#include "util.h"

static ds::LogChannel dsLogChannel("halHalDevice");
static constexpr int DefaultSwitchThreshold = 127;

namespace {

using dss::DeviceSceneSpec_t;
using dss::DimTime;
using dss::LedConfig;
using dss::SceneFlags;

int sceneFlagsFromCache(const SceneFlags& flags) {
  int deviceFlags = 0;
  if (*flags.dontCare) {
    deviceFlags |= 0x01;
  }
  if (*flags.ignoreLocalPrio) {
    deviceFlags |= 0x02;
  }
  if (*flags.specialHandling) {
    deviceFlags |= 0x04;
  }
  if (*flags.blink) {
    deviceFlags |= 0x08;
  }
  deviceFlags |= (static_cast<int>(*flags.ledConfig) << 4);
  deviceFlags |= (static_cast<int>(*flags.dimTime) << 6);
  return deviceFlags;
}

DeviceSceneSpec_t sceneSpecFromCache(const SceneFlags& flags) {
  DeviceSceneSpec_t spec;
  if (flags.dontCare) {
    spec.dontcare = *flags.dontCare;
  }
  if (flags.ignoreLocalPrio) {
    spec.localprio = *flags.ignoreLocalPrio;
  }
  if (flags.specialHandling) {
    spec.specialmode = *flags.specialHandling;
  }
  if (flags.blink) {
    spec.flashmode = *flags.blink;
  }
  if (flags.ledConfig) {
    spec.ledconIndex = static_cast<int>(*flags.ledConfig);
  }
  spec.dimtimeIndex = 0;
  return spec;
}

void sceneFlagsCacheFromDevice(int deviceFlags, SceneFlags& flags) {
  flags.dontCare = (deviceFlags & 0x01);
  flags.ignoreLocalPrio = (deviceFlags & 0x02);
  flags.specialHandling = (deviceFlags & 0x04);
  flags.blink = (deviceFlags & 0x08);
  flags.ledConfig = static_cast<LedConfig>(((deviceFlags)&0x30) >> 4);
  flags.dimTime = static_cast<DimTime>(((deviceFlags)&0xC0) >> 6);
}

void sceneFlagsCacheSceneSpec(const DeviceSceneSpec_t& spec, SceneFlags& flags) {
  flags.dontCare = spec.dontcare;
  flags.ignoreLocalPrio = spec.localprio;
  flags.specialHandling = spec.specialmode;
  flags.blink = spec.flashmode;
  flags.ledConfig = static_cast<LedConfig>(spec.ledconIndex);
  flags.dimTime = DimTime::notSupported;
}

} // namespace

namespace dss::hal {

void HalDevice::write(const Device& device, const std::string& parameter, double value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontCare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    DS_FAIL_REQUIRE("Unknown channel requested", param);
    m_sceneCacheDb.update(device.getDSID(), devCache);
  }
}

void HalDevice::write(const Device& device, const std::string& parameter, const std::string& value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontCare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channelFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      if ((param == "ledConfig") || (param == "dimTime")) {
        SceneFlags& flags = sceneCache.flags;
        if (!(flags.blink && flags.dimTime && flags.dontCare && flags.ignoreLocalPrio && flags.ledConfig &&
              flags.specialHandling)) {
          // cache is not valid
          int deviceFlags = device.getDeviceConfig(128, 128 + sceneNr);
          sceneFlagsCacheFromDevice(deviceFlags, flags);
        }
        if (param == "ledConfig") {
          if (const auto ledValue = ds::tryToEnum<LedConfig>(value)) {
            DS_REQUIRE(static_cast<int>(*ledValue) < 4, "ledConfig is out of range");
            flags.ledConfig = ledValue;
          } else {
            DS_FAIL_REQUIRE("Unknown value for LedConfig", value);
          }
        } else if (param == "dimTime") {
          if (const auto dimValue = ds::tryToEnum<DimTime>(value)) {
            flags.dimTime = dimValue;
          } else {
            DS_FAIL_REQUIRE("Unknown value for DimTime", value);
          }
        }
        int deviceFlags = sceneFlagsFromCache(flags);
        device.setDeviceConfig(128, 128 + sceneNr, deviceFlags, false);
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown channel requested", param);
      }
    } else {
      if ((param == "ledConfig")) {
        SceneFlags& flags = sceneCache.flags;
        if (!(flags.dontCare && flags.ignoreLocalPrio && flags.ledConfig && flags.specialHandling)) {
          // cache is not valid
          DeviceSceneSpec_t sceneSpec;
          device.getDeviceOutputChannelSceneConfig(sceneNr, sceneSpec);
          sceneFlagsCacheSceneSpec(sceneSpec, flags);
        }
        if (param == "ledConfig") {
          if (const auto ledValue = ds::tryToEnum<LedConfig>(value)) {
            flags.ledConfig = ledValue;
          } else {
            DS_FAIL_REQUIRE("Unknown value for LedConfig", value);
          }
        }
        auto sceneSpec = sceneSpecFromCache(flags);
        device.setDeviceOutputChannelSceneConfig(sceneNr, sceneSpec);
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown channel requested", param);
      }
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
}

void HalDevice::write(const Device& device, const std::string& parameter, bool value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channelFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      if ((param == "dontCare") || (param == "ignoreLocalPrio") || (param == "specialHandling") || (param == "blink")) {
        SceneFlags& flags = sceneCache.flags;
        if (!(flags.blink && flags.dimTime && flags.dontCare && flags.ignoreLocalPrio && flags.ledConfig &&
              flags.specialHandling)) {
          // cache is not valid
          int deviceFlags = device.getDeviceConfig(128, 128 + sceneNr);
          sceneFlagsCacheFromDevice(deviceFlags, flags);
        }
        if (param == "dontCare") {
          flags.dontCare = value;
        } else if (param == "ignoreLocalPrio") {
          flags.ignoreLocalPrio = value;
        } else if (param == "specialHandling") {
          flags.specialHandling = value;
        } else if (param == "blink") {
          flags.blink = value;
        }
        int deviceFlags = sceneFlagsFromCache(flags);
        device.setDeviceConfig(128, 128 + sceneNr, deviceFlags, false);
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown boolean channel", param);
      }
    } else {
      if ((param == "dontCare") || (param == "ignoreLocalPrio") || (param == "specialHandling") || (param == "blink")) {
        SceneFlags& flags = sceneCache.flags;
        if (!(flags.blink && flags.dimTime && flags.dontCare && flags.ignoreLocalPrio && flags.ledConfig &&
              flags.specialHandling)) {
          // cache is not valid
          DeviceSceneSpec_t sceneSpec;
          device.getDeviceOutputChannelSceneConfig(sceneNr, sceneSpec);
          sceneFlagsCacheSceneSpec(sceneSpec, flags);
        }
        if (param == "dontCare") {
          flags.dontCare = value;
        } else if (param == "ignoreLocalPrio") {
          flags.ignoreLocalPrio = value;
        } else if (param == "specialHandling") {
          flags.specialHandling = value;
        } else if (param == "blink") {
          flags.blink = value;
        }
        auto sceneSpec = sceneSpecFromCache(flags);
        device.setDeviceOutputChannelSceneConfig(sceneNr, sceneSpec);
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown boolean channel", param);
      }
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channel}/dontCare", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& channelName = route.args["channel"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    auto channelSpec = getOutputChannelIdAndSize(channelName);
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      DS_WARNING("This device does not support channel dontCare flags.", device);
      sceneCache.channels[static_cast<ChannelType>(channelSpec.first)].dontCare = false;
    } else {
      auto flags = device.getDeviceOutputChannelDontCareFlags(sceneNr);
      for (auto channel : getSceneChannels(device)) {
        sceneCache.channels[channel].dontCare = (flags & (1 << device.getOutputChannelIndex(channel))) > 0;
      }
      sceneCache.channels[static_cast<ChannelType>(channelSpec.first)].dontCare = value;
      uint16_t value = 0;
      for (auto channel : getSceneChannels(device)) {
        value |= *sceneCache.channels[channel].dontCare ? (1 << device.getOutputChannelIndex(channel)) : 0;
      }
      device.setDeviceOutputChannelDontCareFlags(sceneNr, value);
    }
    sceneCache.lastChanged = std::chrono::system_clock::now();
    devCache.lastChanged = sceneCache.lastChanged;
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
}

double HalDevice::readDouble(const Device& device, const std::string& parameter) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontcare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
  double returnValue = 0;
  DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  return returnValue;
}

bool HalDevice::readBool(const Device& device, const std::string& parameter) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
  bool returnValue = false;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channelFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      if ((param == "dontCare") || (param == "ignoreLocalPrio") || (param == "specialHandling") || (param == "blink")) {
        SceneFlags& flags = sceneCache.flags;
        int deviceFlags = device.getDeviceConfig(128, 128 + sceneNr);
        sceneFlagsCacheFromDevice(deviceFlags, flags);
        if (param == "dontCare") {
          returnValue = *flags.dontCare;
        } else if (param == "ignoreLocalPrio") {
          returnValue = *flags.ignoreLocalPrio;
        } else if (param == "specialHandling") {
          returnValue = *flags.specialHandling;
        } else if (param == "blink") {
          returnValue = *flags.blink;
        }
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown boolean channel", param);
      }
    } else {
      if ((param == "dontCare") || (param == "ignoreLocalPrio") || (param == "specialHandling") || (param == "blink")) {
        SceneFlags& flags = sceneCache.flags;
        DeviceSceneSpec_t sceneSpec;
        device.getDeviceOutputChannelSceneConfig(sceneNr, sceneSpec);
        sceneFlagsCacheSceneSpec(sceneSpec, flags);
        if (param == "dontCare") {
          returnValue = *flags.dontCare;
        } else if (param == "ignoreLocalPrio") {
          returnValue = *flags.ignoreLocalPrio;
        } else if (param == "specialHandling") {
          returnValue = *flags.specialHandling;
        } else if (param == "blink") {
          returnValue = *flags.blink;
        }
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown boolean channel", param);
      }
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channel}/dontCare", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& channelName = route.args["channel"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    auto channelSpec = getOutputChannelIdAndSize(channelName);
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      DS_WARNING("This device does not support channel dontCare flags.", device);
      sceneCache.channels[static_cast<ChannelType>(channelSpec.first)].dontCare = false;
    } else {
      auto flags = device.getDeviceOutputChannelDontCareFlags(sceneNr);
      for (auto channel : getSceneChannels(device)) {
        sceneCache.channels[channel].dontCare = (flags & (1 << device.getOutputChannelIndex(channel))) > 0;
      }
    }
    returnValue = *sceneCache.channels[static_cast<ChannelType>(channelSpec.first)].dontCare;
    sceneCache.lastChanged = std::chrono::system_clock::now();
    devCache.lastChanged = sceneCache.lastChanged;
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
  return returnValue;
}

std::string HalDevice::readString(const Device& device, const std::string& parameter) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontcare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
  std::string returnValue;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channelFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      if ((param == "ledConfig") || (param == "dimTime")) {
        SceneFlags& flags = sceneCache.flags;
        int deviceFlags = device.getDeviceConfig(128, 128 + sceneNr);
        sceneFlagsCacheFromDevice(deviceFlags, flags);
        if (param == "ledConfig") {
          returnValue = ds::fromEnum<std::string>(*flags.ledConfig);
        } else if (param == "dimTime") {
          returnValue = ds::fromEnum<std::string>(*flags.dimTime);
        }
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown channel requested", param);
      }
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      if (param == "ledConfig") {
        SceneFlags& flags = sceneCache.flags;
        DeviceSceneSpec_t sceneSpec;
        device.getDeviceOutputChannelSceneConfig(sceneNr, sceneSpec);
        sceneFlagsCacheSceneSpec(sceneSpec, flags);
        if (param == "ledConfig") {
          returnValue = ds::fromEnum<std::string>(*flags.ledConfig);
        }
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown channel requested", param);
      }
      m_sceneCacheDb.update(device.getDSID(), devCache);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
  return returnValue;
}

void HalDevice::updateDeviceConfig(const Device& device, int bank, int offset, int value) const {
  /* This is a generic implementation for device flags (bank 128,
   * offset >=128) */
  if (bank != CfgClassScene) {
    return;
  }

  if (offset < 128) {
    return;
  }

  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  SceneEntry& sceneCache = devCache.scene.sceneTable[offset - 128];
  SceneFlags& flags = sceneCache.flags;
  sceneFlagsCacheFromDevice(value, flags);
  sceneCache.lastChanged = std::chrono::system_clock::now();
  devCache.lastChanged = sceneCache.lastChanged;
  m_sceneCacheDb.update(device.getDSID(), devCache);
}

void HalDevice::publishToPropertyTree(const Device& device) const {
  auto state = getState(device);
  auto& propNode = device.getPropertyNode();
  if (propNode) {
    auto statusNode = propNode->createProperty("status");
    statusNode->createProperty("lastChanged")->setStringValue(ds::toJson(state.lastChanged));
    statusNode->createProperty("terminatesAt")->setStringValue(ds::toJson(state.terminatesAt));
    statusNode->createProperty("status")->setStringValue(ds::toJson(state.status));
    auto outputNode = statusNode->createProperty("outputs");
    for (auto& output : state.outputs) {
      auto channelName = ds::fromEnum<std::string>(output.first);
      auto channel = outputNode->createProperty(channelName);
      channel->createProperty("value")->setFloatingValue(state.outputs[output.first].value);
      channel->createProperty("targetValue")->setFloatingValue(state.outputs[output.first].targetValue);
    }
  }
}

void HalDevice::initializeStatus(const Device& device) const {
  auto state = m_StateCache.tryLookup(device.getDSID());
  for (auto& ch : getSceneChannels(device)) {
    if (state.channels.find(ch) == state.channels.end()) {
      state.channels[ch] = {};
    }
  }
  m_StateCache.update(device.getDSID(), state);
}

bool HalDevice::updateCurrentValues(const Device& device, const ds::DateTime& now, DeviceInternalState& state) const {
  bool needsSaving = false;
  if (state.dimmingTimeout &&
      (std::chrono::duration_cast<std::chrono::milliseconds>(*state.dimmingTimeout - now).count() <= 0)) {
    state.dimmStatus = DeviceDimmingStatus::none;
    state.lastDimmCommand.reset();
    state.dimmingTimeout.reset();
    needsSaving = true;
  }
  if (!state.terminatesAt) {
    return needsSaving;
  }
  if (std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - now).count() <= 0) {
    for (auto& channel : state.channels) {
      if (channel.second.target) {
        channel.second.current = channel.second.target;
        channel.second.target.reset();
      }
    }
    state.terminatesAt.reset();
    needsSaving = true;
  } else {
    for (auto& channel : state.channels) {
      if (!channel.second.current || !channel.second.target) {
        continue;
      }
      auto diff = *channel.second.target - *channel.second.current;
      auto prop =
          (double)std::chrono::duration_cast<std::chrono::milliseconds>(now - state.lastChanged).count() /
          std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count();
      *channel.second.current += diff * prop;
    }
  }
  updateSwitchStatus(device, state);
  return needsSaving;
}

bool HalDevice::isDimmingUp(int sceneId, bool forceLocalPrio, DeviceCache& cache) const {
  return ((sceneId == SceneInc && forceLocalPrio) ||
          (sceneId == SceneIncA1 && !cache.scene.sceneTable[SceneA11].flags.dontCare.value_or(true)) ||
          (sceneId == SceneIncA2 && !cache.scene.sceneTable[SceneA21].flags.dontCare.value_or(true)) ||
          (sceneId == SceneIncA3 && !cache.scene.sceneTable[SceneA31].flags.dontCare.value_or(true)) ||
          (sceneId == SceneIncA4 && !cache.scene.sceneTable[SceneA41].flags.dontCare.value_or(true)));
}

bool HalDevice::isDimmingDown(int sceneId, bool forceLocalPrio, DeviceCache& cache) const {
  return ((sceneId == SceneDec && forceLocalPrio) ||
          (sceneId == SceneDecA1 && !cache.scene.sceneTable[SceneA11].flags.dontCare.value_or(true)) ||
          (sceneId == SceneDecA2 && !cache.scene.sceneTable[SceneA21].flags.dontCare.value_or(true)) ||
          (sceneId == SceneDecA3 && !cache.scene.sceneTable[SceneA31].flags.dontCare.value_or(true)) ||
          (sceneId == SceneDecA4 && !cache.scene.sceneTable[SceneA41].flags.dontCare.value_or(true)));
}

bool HalDevice::isStop(int sceneId, bool forceLocalPrio, DeviceCache& cache) const {
  return ((sceneId == SceneStop && forceLocalPrio) ||
          (sceneId == SceneStopA1 && !cache.scene.sceneTable[SceneA11].flags.dontCare.value_or(true)) ||
          (sceneId == SceneStopA2 && !cache.scene.sceneTable[SceneA21].flags.dontCare.value_or(true)) ||
          (sceneId == SceneStopA3 && !cache.scene.sceneTable[SceneA31].flags.dontCare.value_or(true)) ||
          (sceneId == SceneStopA4 && !cache.scene.sceneTable[SceneA41].flags.dontCare.value_or(true)));
}

DeviceState HalDevice::getState(const Device& device) const {
  auto state = m_StateCache.tryLookup(device.getDSID());
  DeviceState ret{};
  ret.lastChanged = state.lastChanged;
  ret.terminatesAt = state.terminatesAt;
  for (auto& ch : state.channels) {
    if (ch.second.current) {
      ret.outputs[ch.first].value = *ch.second.current;
    }
    ret.outputs[ch.first].targetValue = ch.second.target.value_or(ret.outputs[ch.first].value);
    ret.outputs[ch.first].level = ch.second.level;
  }
  ret.status = dsModel::OutputStatusStatus::ok;
  if (state.operationLocked) {
    ret.status = dsModel::OutputStatusStatus::blocked;
  } else if (state.lastDimmCommand) {
    ret.status = dsModel::OutputStatusStatus::dimming;
  } else if (state.terminatesAt) {
    ret.status = dsModel::OutputStatusStatus::moving;
  }
  ret.localPriority = state.localPriority;
  return ret;
}

bool HalDevice::callSceneMin(const Device& device, int sceneId, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  auto sceneCache = cache.scene.sceneTable[sceneId];

  if ((sceneCache.channels[getDefaultChannel()].value && *sceneCache.channels[getDefaultChannel()].value > 0) &&
      (sceneCache.flags.dontCare && !*sceneCache.flags.dontCare) &&
      (state.channels[getDefaultChannel()].current && *state.channels[getDefaultChannel()].current == 0)) {
    // make current value > 0 and do one dim step up
    state.channels[getDefaultChannel()].current = 0.0001;
    state.channels[getDefaultChannel()].target.reset();
    state.terminatesAt.reset();
    m_StateCache.update(device.getDSID(), state);
    return callScene(device, SceneInc, false, now);
  }
  return false;
}

bool HalDevice::undoScene(const Device& device, int sceneId, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());

  if (sceneId == -1 || (state.lastSceneCalled && sceneId == *state.lastSceneCalled)) {
    if (state.beforeLastSceneCalled && *state.beforeLastSceneCalled >= 0) {
      return callScene(device, *state.beforeLastSceneCalled, false,
                       now); // TODO(someday): how does undo handle local prio?
    }
  }
  return false;
}

bool HalDevice::setSceneLocalPrio(const Device& device, int sceneId) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());

  if (sceneId == -1 || (state.lastSceneCalled && sceneId == *state.lastSceneCalled)) {
    state.localPriority = true;
    m_StateCache.update(device.getDSID(), state);
    publishToPropertyTree(device);
    return true;
  }
  return false;
}

void HalDevice::setOperationLock(const Device& device, bool lock) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());

  state.operationLocked = lock;
  m_StateCache.update(device.getDSID(), state);
}

bool HalDevice::stopDimming(const Device& device, int sceneId, ds::DateTime now) const {
  return callScene(device, SceneStop, true, now);
}

bool HalDevice::completeMoving(const Device& device, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  if (updateCurrentValues(device, now, state)) {
    m_StateCache.update(device.getDSID(), state);
    return true;
  }
  return false;
}

bool HalDevice::setOutputValue(const Device& device, ChannelType channel, double value, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  auto channels = getSceneChannels(device);
  if (std::find(channels.begin(), channels.end(), channel) == channels.end()) {
    DS_FAIL_REQUIRE("No such channel", channel);
  }

  state.channels[channel].current = value;
  state.channels[channel].target.reset();
  state.lastDimmCommand.reset();
  state.localPriority = false;
  state.lastSceneCalled.reset();
  state.sceneInRAMBank.reset();
  state.terminatesAt.reset();
  state.lastChanged = now;
  updateSwitchStatus(device, state);
  m_StateCache.update(device.getDSID(), state);
  publishToPropertyTree(device);
  return true;
}

void HalDevice::updateSwitchStatus(const Device& device, DeviceInternalState& state) const {
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (outputMode(cache.property[DeviceProperty::outputMode].value_or(0)) == dsModel::OutputMode::switched) {
    auto threshold = convertFromOutputChannelValue(
        getDefaultChannel(), cache.property[DeviceProperty::switchedThreshold].value_or(DefaultSwitchThreshold));
    if (state.channels[getDefaultChannel()].current.value_or(0) >= threshold) {
      state.channels[getDefaultChannel()].level = 1;
    } else {
      state.channels[getDefaultChannel()].level = 0;
    }
  } else {
    state.channels[getDefaultChannel()].level.reset();
  }
}

void HalDevice::populateCache(const Device& device) const {
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  populateCache(device, cache);
  m_sceneCacheDb.update(device.getDSID(), cache);
}

void HalDevice::populateCache(const Device& device, DeviceCache& cache) const {
  // reading from the maps will create missing entries
  auto scenes = getApplicableScenes();
  auto channels = getSceneChannels(device);
  for (auto scene : scenes) {
    auto& sceneEntry = cache.scene.sceneTable[scene];
    for (auto it = sceneEntry.channels.begin(); it != sceneEntry.channels.end();) {
      if (std::find(channels.begin(), channels.end(), it->first) == channels.end()) {
        it = sceneEntry.channels.erase(it);
      } else {
        ++it;
      }
    }
    for (auto channel : channels) {
      (void)sceneEntry.channels[channel];
    }
  }
  std::vector<int> scenesInCache{};
  scenesInCache.reserve(cache.scene.sceneTable.size());
  for (auto const& scene : cache.scene.sceneTable) {
    scenesInCache.push_back(scene.first);
  }
  std::vector<int> diff(scenesInCache.size());
  auto it = std::set_difference(scenesInCache.begin(), scenesInCache.end(), scenes.begin(), scenes.end(), diff.begin());
  diff.resize(it - diff.begin());
  for (auto s : diff) {
    cache.scene.sceneTable.erase(s);
  }

  const auto& deviceProperties = getDeviceProperties(device);

  DS_DEBUG("properties before", cache.property);
  for (auto it = cache.property.begin(); it != cache.property.end();) {
    if (std::find(deviceProperties.begin(), deviceProperties.end(), it->first) == deviceProperties.end()) {
      it = cache.property.erase(it);
    } else {
      ++it;
    }
  }
  for (auto& property : deviceProperties) {
    (void)cache.property[property];
  }
  DS_DEBUG("properties after", cache.property);
}

DeviceCacheState HalDevice::cacheStatus(const Device& device) const {
  DeviceCacheState result{};

  result.id = ds::str(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  populateCache(device, cache);

  for (auto& property : cache.property) {
    result.total++;
    if (property.second) {
      result.completed++;
    }
  }

  for (auto& scene : cache.scene.sceneTable) {
    result.total++;
    if (scene.second.flags.dontCare) {
      result.completed++;
    }
    for (auto& channel : scene.second.channels) {
      result.total += getSceneSize();
      if (channel.second.value) {
        result.completed += getSceneSize();
      }
    }
  }

  result.completion = (result.total > 0) ? (static_cast<double>(result.completed) / result.total) * 100 : 100;

  return result;
}

std::map<DeviceProperty, std::tuple<int, int, int>> HalDevice::getDeviceParameters(const Device& device) const {
  return m_deviceParameters;
}

std::tuple<int, int, int> HalDevice::getDeviceParameter(const Device& device, DeviceProperty property) const {
  return m_deviceParameters.at(property);
}

double HalDevice::convertThresholdToPercentage(const double& value) const {
  return value / 255.0 * 100;
}

std::vector<double> HalDevice::getDeviceThresholds(const Device& device) const {
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (outputMode(cache.property[DeviceProperty::outputMode].value_or(0)) == dsModel::OutputMode::switched) {
    std::vector<double> thresholds;
    if (cache.property[DeviceProperty::switchedThreshold]) {
      thresholds.push_back(convertThresholdToPercentage(*cache.property[DeviceProperty::switchedThreshold]));
    }
    return thresholds;
  } else {
    return {};
  }
}

void HalDevice::saveScene(const Device& device, int sceneId, ds::DateTime now) const {
  auto cacheEntry = m_sceneCacheDb.tryLookup(device.getDSID());
  for (auto& channel : cacheEntry.scene.sceneTable[sceneId].channels) {
    channel.second.registerValues.clear();
    channel.second.value.reset();
  }
  if (SceneHelper::isAreaScene(sceneId) && sceneId >= SceneA11) {
    // saving an areaOn scene implicitly changes the related areaOff scene
    auto areaOffScene = sceneId - SceneA11 + SceneOffA1;
    for (auto& channel : cacheEntry.scene.sceneTable[areaOffScene].channels) {
      channel.second.value.reset();
      channel.second.registerValues.clear();
    }
    // TB has builtin logic when area scenes are stored
    // new value == 0 -> dontCare:true
    // new value  > 0 -> dontCare:false
    cacheEntry.scene.sceneTable[sceneId].flags = {};
    cacheEntry.scene.sceneTable[areaOffScene].flags = {};
  }
  m_sceneCacheDb.update(device.getDSID(), cacheEntry);

  // see Device::evaluateGRKL387Workaround, DSD-1352 why it's a bad idea to
  // trust CfgClassRuntime upon saveScene, wait for callScene
  // evtl. we could limit ourselves to if !apartmentScene
  auto state = m_StateCache.tryLookup(device.getDSID());
  state.sceneInRAMBank.reset();
  m_StateCache.update(device.getDSID(), state);
}

} // namespace dss::hal
