/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hal-device.h"

#include "ds-api/route-segments.h"
#include "model/apartment.h"
#include "model/device.h"
#include "ds-model.h"
#include "util.h"
#include "event.h"


namespace dss::hal {

static ds::LogChannel dsLogChannel("halJokerDevice");

std::map<DeviceProperty, std::tuple<int, int, int>> UMRParameters = {
    {DeviceProperty::switchedThreshold, {3, 0x30, 1}}
};
std::map<DeviceProperty, std::tuple<int, int, int>> UMRPrimaryOnly = {
      {DeviceProperty::switchedThresholdTwoStep0, {3, 0x32, 1}},
      {DeviceProperty::switchedThresholdTwoStep1, {3, 0x33, 1}},
      {DeviceProperty::switchedThreshold3Step0, {3, 0x34, 1}},
      {DeviceProperty::switchedThreshold3Step1, {3, 0x35, 1}},
      {DeviceProperty::switchedThreshold3Step2, {3, 0x36, 1}}
};

[[maybe_unused]] static constexpr int DefaultDimStep = 11;
[[maybe_unused]] static constexpr int DefaultMinDim = 16;
[[maybe_unused]] static constexpr int DefaultDimTime0 = 47;
[[maybe_unused]] static constexpr int DefaultDimTime1 = 162;
[[maybe_unused]] static constexpr int DefaultDimTime2 = 0;
[[maybe_unused]] static constexpr int DefaultSwitchThreshold2Step0 = 85;
[[maybe_unused]] static constexpr int DefaultSwitchThreshold2Step1 = 170;
[[maybe_unused]] static constexpr int DefaultSwitchThreshold3Step0 = 64;
[[maybe_unused]] static constexpr int DefaultSwitchThreshold3Step1 = 128;
[[maybe_unused]] static constexpr int DefaultSwitchThreshold3Step2 = 192;

void HalJokerDevice::write(const Device& device, const std::string& parameter, double value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontCare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "powerState") {
      auto channelSpec = getOutputChannelIdAndSize("powerState");
      device.setDeviceConfig(128, sceneNr, convertToOutputChannelValue(channelSpec.first, value), false);
      SceneChannel& brightness = sceneCache.channels[ChannelType::powerState];
      brightness.value = value;
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else if (param == "brightness") {
      DS_FAIL_REQUIRE("Unknown channel for this device", param);
    } else {
      return HalLightDevice::write(device, parameter,value);
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property) {
      unsigned int intValue = 0;
      if (*property == DeviceProperty::switchedThresholdTwoStep0 || \
          *property == DeviceProperty::switchedThresholdTwoStep1 || \
          *property == DeviceProperty::switchedThreshold3Step0 || \
          *property == DeviceProperty::switchedThreshold3Step1 || \
          *property == DeviceProperty::switchedThreshold3Step2) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 100, "Value too high");
        intValue = static_cast<unsigned int>(round((255 * value) / 100.0));
      } else {
        return HalLightDevice::write(device, parameter,value);
      }
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      device.setDeviceConfig(bank, offset, intValue, false);
      devCache.property[*property] = intValue;
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      return HalLightDevice::write(device, parameter,value);
    }
  } else {
    return HalLightDevice::write(device, parameter,value);
  }
}

double HalJokerDevice::readDouble(const Device& device, const std::string& parameter) const {
  // "/scene/5/powerState" -> 128, 5
  // "/scene/6/dontcare"   -> 128, 134 (=128+6)
  // "/scene/7/powerState/dontcare" -> 128, 135 (=128+7)
  double returnValue = 0;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "powerState") {
      auto value = device.getDeviceConfig(128, sceneNr);
      SceneChannel& powerState = sceneCache.channels[ChannelType::powerState];
      auto channelSpec = getOutputChannelIdAndSize("powerState");
      returnValue = convertFromOutputChannelValue(channelSpec.first, value);
      powerState.value = returnValue;
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
      DS_DEBUG("Dev cache to save: ", value, returnValue, devCache);
    } else if (param == "brightness") {
      DS_FAIL_REQUIRE("Unknown channel for this device", param);
    } else {
      return HalLightDevice::readDouble(device, parameter);
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property && (*property == DeviceProperty::switchedThresholdTwoStep0 || \
                     *property == DeviceProperty::switchedThresholdTwoStep1 || \
                     *property == DeviceProperty::switchedThreshold3Step0 || \
                     *property == DeviceProperty::switchedThreshold3Step1 || \
                     *property == DeviceProperty::switchedThreshold3Step2)) {
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      auto value = device.getDeviceConfig(bank, offset);
      auto channelSpec = getOutputChannelIdAndSize("powerState");
      returnValue = convertFromOutputChannelValue(channelSpec.first, value);
      devCache.property[*property] = value;
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      return HalLightDevice::readDouble(device, parameter);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
  return returnValue;
}

bool HalJokerDevice::callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const {
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());

  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_POWERSAVE && isIgnoredInPowersave(sceneId)) {
    return false;
  } else if (cache.property[DeviceProperty::outputMode] && *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_WIPE) {
    if (isWipeStandbyScene(sceneId)) {
      bool needsSaving = false;
      auto sceneCache = cache.scene.sceneTable[sceneId];
      auto state = m_StateCache.tryLookup(device.getDSID());

      if (state.operationLocked) {
        return false;
      }
      if (forced || *sceneCache.flags.ignoreLocalPrio || !state.localPriority) {
        for (auto& channel : state.channels) {
          if (sceneCache.channels[channel.first].dontCare && *sceneCache.channels[channel.first].dontCare) {
            continue;
          }
          state.dimmStatus = DeviceDimmingStatus::none;
          state.lastDimmCommand.reset();

          if (channel.second.level > 0) {
            state.standby = true;
            needsSaving = true;
          }
        }
      }
      if (needsSaving) {
        state.lastChanged = now;
        m_StateCache.update(device.getDSID(), state);
        publishToPropertyTree(device);
        return true;
      }
      return false;
    }
  }

  return HalLightDevice::callScene(device, sceneId, forced, now);
}

bool HalJokerDevice::isIgnoredInPowersave(int sceneId) const {
  return (sceneId == SceneOff || sceneId == SceneOffA1 || \
          sceneId == SceneOffA2 || sceneId == SceneOffA3 || \
          sceneId == SceneOffA4 || sceneId == SceneOffE1 || \
          sceneId == SceneOffE2 || sceneId == SceneOffE3 || \
          sceneId == SceneOffE4);
}

bool HalJokerDevice::isWipeStandbyScene(int sceneId) const {
  return (isIgnoredInPowersave(sceneId) || sceneId == SceneStandBy);
}

const std::vector<ChannelType> HalJokerDevice::getSceneChannels(const Device& device) const {
  return std::vector<ChannelType>({ChannelType::powerState});
}

DeviceState HalJokerDevice::getState(const Device& device) const {
  auto state = m_StateCache.tryLookup(device.getDSID());
  DeviceState ret{};
  ret.lastChanged = state.lastChanged;
  ret.terminatesAt = state.terminatesAt;

  bool isVentilation = device.isInGroup(GroupIDVentilation) || device.isInGroup(GroupIDRecirculation) || device.isInGroup(GroupIDGlobalAppDsVentilation) || device.isInGroup(GroupIDGlobalAppDsRecirculation);
  bool isLight = device.isInGroup(GroupIDYellow);
  ChannelType type = ChannelType::powerState;
  if (isLight) {
    type = ChannelType::brightness;
  } else if (isVentilation) {
    type = ChannelType::airFlowIntensity;
  }

  for (auto& ch : state.channels) {
    if (ch.second.current) {
      ret.outputs[type].value = convertValueToOtherChannelType(ChannelType::powerState, *ch.second.current, type);
    }
    ret.outputs[type].targetValue = ch.second.target ? convertValueToOtherChannelType(ChannelType::powerState, *ch.second.target, type) : ret.outputs[type].value;
    ret.outputs[type].level = ch.second.level;
  }
  ret.status = dsModel::OutputStatusStatus::ok;
  if (state.operationLocked) {
    ret.status = dsModel::OutputStatusStatus::blocked;
  } else if (state.lastDimmCommand) {
    ret.status = dsModel::OutputStatusStatus::dimming;
  } else if (state.terminatesAt) {
    ret.status = dsModel::OutputStatusStatus::moving;
  } else if (state.standby) {
    ret.status = dsModel::OutputStatusStatus::standby;
  }
  ret.localPriority = state.localPriority;
  return ret;
}

bool HalJokerDevice::updateOutputValue(const Device& device, uint8_t value, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  state.channels[ChannelType::powerState].current = convertFromOutputChannelValue(ChannelType::powerState, value);
  updateSwitchStatus(device, state);
  m_StateCache.update(device.getDSID(), state);
  publishToPropertyTree(device);
  return true;
}

void HalJokerDevice::updateDeviceConfig(const Device& device, int bank, int offset, int value) const {
  HalDevice::updateDeviceConfig(device, bank, offset, value);

  if (bank == CfgClassRuntime) {
    if (offset == 0) {
      updateOutputValue(device, value, std::chrono::system_clock::now());
    }
  } else if (bank == CfgClassScene) {
    if ((offset < 0) || (offset >= 128)) {
      return;
    }

    DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
    SceneEntry& sceneCache = devCache.scene.sceneTable[offset];
    SceneChannel& powerState = sceneCache.channels[ChannelType::powerState];
    powerState.value = convertFromOutputChannelValue(ChannelType::powerState, value);
    sceneCache.lastChanged = std::chrono::system_clock::now();
    devCache.lastChanged = sceneCache.lastChanged;
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else {
    for (const auto& property : getDeviceParameters(device)) {
      int propBank;
      int propOffset;
      int propLength;
      std::tie(propBank, propOffset, propLength) = property.second;
      if ((propBank == bank) && (propOffset == offset)) {
        DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
        devCache.property[property.first] = value;
        devCache.lastChanged = std::chrono::system_clock::now();
        m_sceneCacheDb.update(device.getDSID(), devCache);
        if (property.first == DeviceProperty::outputMode || \
            property.first == DeviceProperty::switchedThreshold || \
            property.first == DeviceProperty::switchedThresholdTwoStep0 || \
            property.first == DeviceProperty::switchedThresholdTwoStep1 || \
            property.first == DeviceProperty::switchedThreshold3Step0 || \
            property.first == DeviceProperty::switchedThreshold3Step1 || \
            property.first == DeviceProperty::switchedThreshold3Step2) {
          auto state = m_StateCache.tryLookup(device.getDSID());
          updateSwitchStatus(device, state);
          m_StateCache.update(device.getDSID(), state);
        }
      }
    }
  }
}

std::map<DeviceProperty, std::tuple<int, int, int>> HalJokerDevice::getDeviceParameters(const Device& device) const {
  DS_DEBUG("getDeviceParameters", device.hardwareType());
  if (device.hardwareType() == HardwareType::SW_UMR200) {
    auto parameters = HalDevice::getDeviceParameters(device);
    for (const auto& parameter : UMRParameters) {
      parameters[parameter.first] = parameter.second;
    }
    if (device.multiDeviceIndex() == 2) {
      for (const auto& parameter : UMRPrimaryOnly) {
        parameters[parameter.first] = parameter.second;
      }
    }
    return parameters;
  } else {
    return HalDevice::getDeviceParameters(device);
  }
}

std::tuple<int, int, int> HalJokerDevice::getDeviceParameter(const Device& device, DeviceProperty property) const {
  DS_DEBUG("getDeviceParameter", device.hardwareType());
  if (device.hardwareType() == HardwareType::SW_UMR200) {
    if (property == DeviceProperty::switchedThreshold) {
      return UMRParameters[DeviceProperty::switchedThreshold];
    } else if (device.multiDeviceIndex() == 2) {
      auto it = UMRPrimaryOnly.find(property);
      if (it != UMRPrimaryOnly.end()) {
        return (*it).second;
      } else {
        return HalDevice::getDeviceParameter(device, property);
      }
    } else {
      return HalDevice::getDeviceParameter(device, property);
    }
  } else {
    return HalDevice::getDeviceParameter(device, property);
  }
}

void HalJokerDevice::updateSwitchStatus(const Device& device, DeviceInternalState& state) const {
  DS_DEBUG("updateSwitchStatus");
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (device.hardwareType() == HardwareType::SW_UMR200) {
    DS_DEBUG("updateSwitchStatus", cache.property[DeviceProperty::outputMode].value_or(0), state.channels[getDefaultChannel()].current.value_or(0));
    if (OutputMode(cache.property[DeviceProperty::outputMode].value_or(0)) == OutputMode::TwoStageSwitch) {
      auto thresholdOne = convertFromOutputChannelValue(
          getDefaultChannel(), cache.property[DeviceProperty::switchedThresholdTwoStep0].value_or(DefaultSwitchThreshold2Step0));
      auto thresholdTwo = convertFromOutputChannelValue(
          getDefaultChannel(), cache.property[DeviceProperty::switchedThresholdTwoStep1].value_or(DefaultSwitchThreshold2Step1));
      auto currValue = state.channels[getDefaultChannel()].current.value_or(0);
      if (currValue >= thresholdOne && currValue < thresholdTwo) {
        state.channels[getDefaultChannel()].level = 1;
      } else if (currValue >= thresholdTwo) {
        state.channels[getDefaultChannel()].level = 2;
      } else {
        state.channels[getDefaultChannel()].level = 0;
      }
    } else if (OutputMode(cache.property[DeviceProperty::outputMode].value_or(0)) == OutputMode::ThreeStageSwitch) {
      auto thresholdOne = convertFromOutputChannelValue(
          getDefaultChannel(), cache.property[DeviceProperty::switchedThreshold3Step0].value_or(DefaultSwitchThreshold3Step0));
      auto thresholdTwo = convertFromOutputChannelValue(
          getDefaultChannel(), cache.property[DeviceProperty::switchedThreshold3Step1].value_or(DefaultSwitchThreshold3Step1));
      auto thresholdThree = convertFromOutputChannelValue(
          getDefaultChannel(), cache.property[DeviceProperty::switchedThreshold3Step2].value_or(DefaultSwitchThreshold3Step2));
      auto currValue = state.channels[getDefaultChannel()].current.value_or(0);
      if (currValue >= thresholdOne && currValue < thresholdTwo) {
        state.channels[getDefaultChannel()].level = 1;
      } else if (currValue >= thresholdTwo && currValue < thresholdThree) {
        state.channels[getDefaultChannel()].level = 2;
      } else if (currValue >= thresholdThree) {
        state.channels[getDefaultChannel()].level = 3;
      } else {
        state.channels[getDefaultChannel()].level = 0;
      }
    } else {
      HalDevice::updateSwitchStatus(device, state);
    }
  } else {
    HalDevice::updateSwitchStatus(device, state);
  }
}

const std::vector<DeviceProperty> HalJokerDevice::getDeviceProperties(const Device& device) const {
  if (device.hardwareType() == HardwareType::SW_UMR200 && device.multiDeviceIndex() == 2) {
    std::vector<DeviceProperty> properties = HalLightDevice::getDeviceProperties(device);
    for (auto const& propertyTuple : UMRPrimaryOnly) {
      properties.push_back(propertyTuple.first);
    }
    return properties;
  } else {
    return HalLightDevice::getDeviceProperties(device);
  }
}

std::vector<double> HalJokerDevice::getDeviceThresholds(const Device& device) const {
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (OutputMode(cache.property[DeviceProperty::outputMode].value_or(0)) == OutputMode::TwoStageSwitch) {
    std::vector<double> thresholds;
    for (auto property : {DeviceProperty::switchedThresholdTwoStep0, DeviceProperty::switchedThresholdTwoStep1}) {
      if (cache.property[property]) {
        thresholds.push_back(convertThresholdToPercentage(*cache.property[property]));
      }
    }
  } else if (OutputMode(cache.property[DeviceProperty::outputMode].value_or(0)) == OutputMode::ThreeStageSwitch) {
    std::vector<double> thresholds;
    for (auto property : {DeviceProperty::switchedThreshold3Step0, DeviceProperty::switchedThreshold3Step1, 
                          DeviceProperty::switchedThreshold3Step2}) {
      if (cache.property[property]) {
        thresholds.push_back(convertThresholdToPercentage(*cache.property[property]));
      }
    }
  } else {
    return HalLightDevice::getDeviceThresholds(device);
  }
  return {};
}

} // namespace dss::hal
