/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hal-device.h"

#include <chrono>
#include <cmath>

#include "base.h"
#include "ds-api/route-segments.h"
#include "dss.h"
#include "event.h"
#include "model/apartment.h"
#include "model/device.h"
#include "util.h"

namespace dss::hal {

static ds::LogChannel dsLogChannel("halLightDevice");

static constexpr int DefaultDimStep = 11;
static constexpr int DefaultMinDim = 16;
static constexpr int DefaultDimTime0 = 47;
static constexpr int DefaultDimTime1 = 162;
static constexpr int DefaultDimTime2 = 0;
static constexpr int DefaultIncInterval = 250;

void HalLightDevice::write(const Device& device, const std::string& parameter, double value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontCare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      if (param == "brightness") {
        auto channelSpec = getOutputChannelIdAndSize("brightness");
        device.setDeviceConfig(128, sceneNr, convertToOutputChannelValue(channelSpec.first, value), false);
        SceneChannel& brightness = sceneCache.channels[ChannelType::brightness];
        brightness.value = value;
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown channel for this device", param);
      }
    } else {
      auto channelSpec = getOutputChannelIdAndSize(param);
      DS_REQUIRE(device.getOutputChannelIndex(channelSpec.first) >= 0, "Device does not support this channel type");
      auto channelIndex = device.getOutputChannelIndex(channelSpec.first);
      device.setDeviceOutputChannelSceneValue(channelIndex, channelSpec.second, sceneNr,
                                              convertToOutputChannelValue(channelSpec.first, value));
      SceneChannel& hue = sceneCache.channels[static_cast<ChannelType>(channelSpec.first)];
      hue.value = value;
      invalidateReleatedChannels(sceneCache, static_cast<ChannelType>(channelSpec.first));
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property) {
      unsigned int intValue = 0;
      if (*property == DeviceProperty::switchedThreshold || *property == DeviceProperty::dimStepUp ||
          *property == DeviceProperty::dimStepDown || *property == DeviceProperty::minDim) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 100, "Value too high");
        intValue = static_cast<unsigned int>(round((255 * value) / 100.0));
      } else if (*property == DeviceProperty::dimTime0Up || *property == DeviceProperty::dimTime0Down ||
                 *property == DeviceProperty::dimTime1Up || *property == DeviceProperty::dimTime1Down ||
                 *property == DeviceProperty::dimTime2Up || *property == DeviceProperty::dimTime2Down) {
        DS_REQUIRE(value >= 0.053, "Value too low");
        DS_REQUIRE(value <= 3276.800, "Value too high");
        intValue = convertDimTimeToDevice(value);
      } else {
        DS_FAIL_REQUIRE("No value conversion found for ", propertyName);
      }
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      device.setDeviceConfig(bank, offset, intValue, false);
      devCache.property[*property] = intValue;
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
}

void HalLightDevice::write(const Device& device, const std::string& parameter, const std::string& value) const {
  try {
    return HalDevice::write(device, parameter, value);
  } catch (const ds::Exception& e) {
  }

  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property) {
      unsigned int intValue = 0;
      if (*property == DeviceProperty::outputMode) {
        DS_DEBUG(value);
        auto enumValue = ds::tryToEnum<OutputMode>(value);
        DS_REQUIRE(enumValue, "Unknown value");
        intValue = static_cast<int>(*enumValue);
      } else {
        DS_FAIL_REQUIRE("No value conversion found for ", propertyName);
      }
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      device.setDeviceConfig(bank, offset, intValue, false);
      devCache.property[*property] = intValue;
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
};

void HalLightDevice::invalidateReleatedChannels(SceneEntry& sceneCache, const ChannelType channel) const {
  switch (channel) {
    case ChannelType::none: break;
    case ChannelType::hue:
    case ChannelType::saturation:
      sceneCache.channels[ChannelType::colortemp].value.reset();
      sceneCache.channels[ChannelType::x].value.reset();
      sceneCache.channels[ChannelType::y].value.reset();
      break;
    case ChannelType::colortemp:
      sceneCache.channels[ChannelType::hue].value.reset();
      sceneCache.channels[ChannelType::saturation].value.reset();
      sceneCache.channels[ChannelType::x].value.reset();
      sceneCache.channels[ChannelType::y].value.reset();
      break;
    case ChannelType::x:
    case ChannelType::y:
      sceneCache.channels[ChannelType::hue].value.reset();
      sceneCache.channels[ChannelType::saturation].value.reset();
      sceneCache.channels[ChannelType::colortemp].value.reset();
      break;
    case ChannelType::brightness:
    case ChannelType::shadePositionOutside:
    case ChannelType::shadePositionIndoor:
    case ChannelType::shadeOpeningAngleOutside:
    case ChannelType::shadeOpeningAngleIndoor:
    case ChannelType::transparency:
    case ChannelType::airFlowIntensity:
    case ChannelType::airFlowDirection:
    case ChannelType::airFlapPosition:
    case ChannelType::airLouverPosition:
    case ChannelType::heatingPower:
    case ChannelType::coolingCapacity:
    case ChannelType::audioVolume:
    case ChannelType::powerState:
    case ChannelType::airLouverAuto:
    case ChannelType::airFlowAuto:
    case ChannelType::waterTemperature:
    case ChannelType::waterFlowRate:
    case ChannelType::powerLevel:
    case ChannelType::videoStation:
    case ChannelType::videoInputSource: break;
  }
}

double HalLightDevice::readDouble(const Device& device, const std::string& parameter) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontcare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
  double returnValue = 0;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
      if (param == "brightness") {
        auto value = device.getDeviceConfig(128, sceneNr);
        SceneChannel& brightness = sceneCache.channels[ChannelType::brightness];
        auto channelSpec = getOutputChannelIdAndSize("brightness");
        returnValue = convertFromOutputChannelValue(channelSpec.first, value);
        brightness.value = returnValue;
        sceneCache.lastChanged = std::chrono::system_clock::now();
        devCache.lastChanged = sceneCache.lastChanged;
      } else {
        DS_FAIL_REQUIRE("Unknown channel for this device", param);
      }
    } else {
      auto channelSpec = getOutputChannelIdAndSize(param);
      DS_REQUIRE(device.getOutputChannelIndex(channelSpec.first) >= 0, "Device does not support this channel type");
      auto channelIndex = device.getOutputChannelIndex(channelSpec.first);
      auto value = device.getDeviceOutputChannelSceneValue(channelIndex, sceneNr);
      SceneChannel& hue = sceneCache.channels[static_cast<ChannelType>(channelSpec.first)];
      returnValue = convertFromOutputChannelValue(channelSpec.first, value);
      hue.value = returnValue;
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property && (*property == DeviceProperty::switchedThreshold || *property == DeviceProperty::dimTime0Up ||
                     *property == DeviceProperty::dimTime0Down || *property == DeviceProperty::dimTime1Up ||
                     *property == DeviceProperty::dimTime1Down || *property == DeviceProperty::dimTime2Up ||
                     *property == DeviceProperty::dimTime2Down || *property == DeviceProperty::dimStepUp ||
                     *property == DeviceProperty::dimStepDown || *property == DeviceProperty::minDim)) {
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      auto value = device.getDeviceConfig(bank, offset);
      if (*property == DeviceProperty::switchedThreshold || *property == DeviceProperty::dimStepUp ||
          *property == DeviceProperty::dimStepDown || *property == DeviceProperty::minDim) {
        returnValue = value / 255.0 * 100.0;
      } else if (*property == DeviceProperty::dimTime0Up || *property == DeviceProperty::dimTime0Down ||
                 *property == DeviceProperty::dimTime1Up || *property == DeviceProperty::dimTime1Down ||
                 *property == DeviceProperty::dimTime2Up || *property == DeviceProperty::dimTime2Down) {
        returnValue = convertDimTimeFromDevice(value);
      } else {
        DS_FAIL_REQUIRE("No value conversion found for ", propertyName);
      }
      devCache.property[*property] = value;
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
  return returnValue;
}

std::string HalLightDevice::readString(const Device& device, const std::string& parameter) const {
  try {
    return HalDevice::readString(device, parameter);
  } catch (const ds::Exception& e) {
  }

  std::string returnValue;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property && *property == DeviceProperty::outputMode) {
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      auto value = device.getDeviceConfig(bank, offset);
      auto enumValue = static_cast<OutputMode>(value);
      returnValue = ds::fromEnum<std::string>(enumValue);
      DS_REQUIRE(!returnValue.empty(), "Unknown value");
      devCache.property[*property] = value;
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
  return returnValue;
}

void HalLightDevice::updateDeviceConfig(const Device& device, int bank, int offset, int value) const {
  HalDevice::updateDeviceConfig(device, bank, offset, value);

  if (bank == CfgClassRuntime) {
    if (offset == 0) {
      updateOutputValue(device, value, std::chrono::system_clock::now());
    }
  } else if (bank == CfgClassScene) {
    if ((offset < 0) || (offset >= 128)) {
      return;
    }

    DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
    SceneEntry& sceneCache = devCache.scene.sceneTable[offset];
    SceneChannel& brightness = sceneCache.channels[ChannelType::brightness];
    brightness.value = convertFromOutputChannelValue(ChannelType::brightness, value);
    sceneCache.lastChanged = std::chrono::system_clock::now();
    devCache.lastChanged = sceneCache.lastChanged;
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else {
    for (const auto& property : getDeviceParameters(device)) {
      int propBank;
      int propOffset;
      int propLength;
      std::tie(propBank, propOffset, propLength) = property.second;
      if ((propBank == bank) && (propOffset == offset)) {
        DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
        devCache.property[property.first] = value;
        devCache.lastChanged = std::chrono::system_clock::now();
        m_sceneCacheDb.update(device.getDSID(), devCache);
        if (property.first == DeviceProperty::outputMode || property.first == DeviceProperty::switchedThreshold) {
          auto state = m_StateCache.tryLookup(device.getDSID());
          updateSwitchStatus(device, state);
          m_StateCache.update(device.getDSID(), state);
        }
      }
    }
  }
}

bool HalLightDevice::handleDimming(DeviceInternalState& state, DeviceCache& cache, const ds::DateTime& now,
                                   const int dimmScene, bool up) const {
  auto& channel = state.channels[getDefaultChannel()];
  if (channel.current && *channel.current > 0 && state.lastSceneCalled &&
      cache.scene.sceneTable[*state.lastSceneCalled].channels[getDefaultChannel()].value.value_or(0) > 0) {
    auto dimStep =
        up ? convertFromOutputChannelValue(getDefaultChannel(),
                                           cache.property[DeviceProperty::dimStepUp].value_or(DefaultDimStep))
           : -1 * convertFromOutputChannelValue(getDefaultChannel(),
                                                cache.property[DeviceProperty::dimStepDown].value_or(DefaultDimStep));
    channel.target = *channel.current + dimStep;
    if (up) {
      channel.target = limitOutputChannelValue(getDefaultChannel(), *channel.target);
    } else {
      channel.target = std::max(
          *channel.target, convertFromOutputChannelValue(
                               getDefaultChannel(), cache.property[DeviceProperty::minDim].value_or(DefaultMinDim)));
    }
    state.dimmingTimeout = now + std::chrono::milliseconds(2 * DefaultIncInterval);
    state.dimmStatus = DeviceDimmingStatus::dimming;
    state.lastDimmCommand = dimmScene;
    return true;
  }
  return false;
}

bool HalLightDevice::callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const {
  DS_REQUIRE(sceneId <= MaxSceneNumber, "Scene number is out of range.", sceneId);

  bool needsSaving = false;

  initializeStatus(device);

  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  auto sceneCache = cache.scene.sceneTable[sceneId];

  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  auto state = m_StateCache.tryLookup(device.getDSID());

  if (state.operationLocked) {
    return false;
  }

  needsSaving = updateCurrentValues(device, now, state) || needsSaving;

  double dimTime = 0;
  if (SceneHelper::isDimScene(sceneId)) {
    if (isDimmingDown(sceneId, (forced || !state.localPriority), cache)) {
      needsSaving = handleDimming(state, cache, now, sceneId, false) || needsSaving;
    } else if (isDimmingUp(sceneId, (forced || !state.localPriority), cache)) {
      needsSaving = handleDimming(state, cache, now, sceneId, true) || needsSaving;
    } else if (isStop(sceneId, (forced || !state.localPriority), cache)) {
      auto& channel = state.channels[getDefaultChannel()];
      if (channel.current) {
        state.dimmStatus = DeviceDimmingStatus::none;
        state.lastDimmCommand.reset();
        state.dimmingTimeout.reset();
        channel.target = *channel.current;
        state.standby.reset();
      }
    } else if (sceneId == SceneDimArea) {
      if (state.lastDimmCommand && (*state.lastDimmCommand == SceneDecA1 || *state.lastDimmCommand == SceneDecA2 ||
                                    *state.lastDimmCommand == SceneDecA3 || *state.lastDimmCommand == SceneDecA4)) {
        needsSaving = handleDimming(state, cache, now, *state.lastDimmCommand, false) || needsSaving;
      }
      if (state.lastDimmCommand && (*state.lastDimmCommand == SceneIncA1 || *state.lastDimmCommand == SceneIncA2 ||
                                    *state.lastDimmCommand == SceneIncA3 || *state.lastDimmCommand == SceneIncA4)) {
        needsSaving = handleDimming(state, cache, now, *state.lastDimmCommand, true) || needsSaving;
      }
    } else {
      return false;
    }
  } else {
    if (!sceneCache.flags.dontCare) {
      return false;
    }

    if (*sceneCache.flags.dontCare) {
      return false;
    }

    if (forced || *sceneCache.flags.ignoreLocalPrio || !state.localPriority || (sceneId == SceneLocalOff)) {
      bool channelChange = false;
      for (auto& channel : state.channels) {
        if (sceneCache.channels[channel.first].dontCare && *sceneCache.channels[channel.first].dontCare) {
          continue;
        }
        state.dimmStatus = DeviceDimmingStatus::none;
        state.lastDimmCommand.reset();
        state.standby.reset();
        channel.second.target = sceneCache.channels[channel.first].value;

        if (channel.second.target) {
          channel.second.target = limitOutputChannelValue(channel.first, *channel.second.target);
        }
        if (channel.second.current != channel.second.target) {
          channelChange = true;
        }
        needsSaving = true;
      }

      if (sceneId == SceneLocalOn && *sceneCache.flags.specialHandling) {
        state.localPriority = true;
      } else {
        state.localPriority = false;
      }
      state.beforeLastSceneCalled = state.lastSceneCalled;
      state.lastSceneCalled = sceneId;
      if (channelChange) {
        switch (sceneCache.flags.dimTime.value_or(DimTime::notSupported)) {
          case DimTime::notSupported:
          case DimTime::noDimTime: dimTime = 0; break;
          case DimTime::dimTime0:
            dimTime = convertDimTimeFromDevice(
                std::max(cache.property[DeviceProperty::dimTime0Up].value_or(DefaultDimTime0),
                         cache.property[DeviceProperty::dimTime0Down].value_or(DefaultDimTime0)));

            break;
          case DimTime::dimTime1:
            dimTime = convertDimTimeFromDevice(
                std::max(cache.property[DeviceProperty::dimTime1Up].value_or(DefaultDimTime1),
                         cache.property[DeviceProperty::dimTime1Down].value_or(DefaultDimTime1)));

            break;
          case DimTime::dimTime2:
            dimTime = convertDimTimeFromDevice(
                std::max(cache.property[DeviceProperty::dimTime2Up].value_or(DefaultDimTime2),
                         cache.property[DeviceProperty::dimTime2Down].value_or(DefaultDimTime2)));

            break;
        }
        if (dimTime <= 1) {
          dimTime = 0;
        }
      }
    } else {
      return false;
    }
  }
  state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(dimTime * 1000));
  state.lastChanged = now;
  needsSaving = updateCurrentValues(device, now, state) || needsSaving;
  if (state.terminatesAt &&
      std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() > 1000) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = std::make_shared<Event>(EventName::ApartmentProxyDeviceTimeout, devRef);
    event->setProperty(
        "time",
        "+" +
            intToString(
                std::chrono::duration_cast<std::chrono::seconds>(*state.terminatesAt - state.lastChanged).count() + 2));
    event->setProperty(EventProperty::Unique, "Yes");
    device.getApartment().pushEvent(event);
  } else if (state.dimmingTimeout) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = std::make_shared<Event>(EventName::ApartmentProxyDeviceDimmingTimeout, devRef);
    event->setProperty("time", "+5");
    event->setProperty(EventProperty::Unique, "Yes");
    device.getApartment().pushEvent(event);
  }

  if (needsSaving) {
    state.lastChanged = now;
    m_StateCache.update(device.getDSID(), state);
    publishToPropertyTree(device);
    return true;
  }
  return false;
} // namespace dss

bool HalLightDevice::updateOutputValue(const Device& device, uint8_t value, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  state.channels[getDefaultChannel()].current = convertFromOutputChannelValue(getDefaultChannel(), value);
  updateSwitchStatus(device, state);
  m_StateCache.update(device.getDSID(), state);
  publishToPropertyTree(device);
  return true;
}

bool HalLightDevice::setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const {
  double convertedValue = convertFromOutputChannelValue(getDefaultChannel(), value);

  return HalDevice::setOutputValue(device, getDefaultChannel(), convertedValue, now);
}

const std::vector<ChannelType> HalLightDevice::getSceneChannels(const Device& device) const {
  if (device.getOutputChannelCount() == 0 || device.hasEmulatedHardwareOutputs()) {
    return std::vector<ChannelType>({ChannelType::brightness});
  } else {
    std::vector<ChannelType> ret;
    for (auto& ch : device.getOutputChannels()) {
      if (ds::isValidEnum(ch->type)) {
        ret.push_back(ch->type);
      }
    }
    return ret;
  }
}

const std::vector<DeviceProperty> HalLightDevice::getDeviceProperties(const Device& device) const {
  if (device.isVdcDevice()) {
    return std::vector<DeviceProperty>({DeviceProperty::outputMode});
  } else {
    return std::vector<DeviceProperty>(
        {DeviceProperty::outputMode, DeviceProperty::switchedThreshold, DeviceProperty::dimTime0Up,
         DeviceProperty::dimTime0Down, DeviceProperty::dimTime1Up, DeviceProperty::dimTime1Down,
         DeviceProperty::dimTime2Up, DeviceProperty::dimTime2Down, DeviceProperty::dimStepUp,
         DeviceProperty::dimStepDown, DeviceProperty::minDim});
  }
}

std::map<DeviceProperty, std::tuple<int, int, int>> HalLightDevice::getDeviceParameters(const Device& device) const {
  if (device.getDeviceType() == DEVICE_TYPE_KL) {
    auto parameters = HalDevice::getDeviceParameters(device);
    parameters[DeviceProperty::switchedThreshold] = {3, 0x3C, 1};
    return parameters;
  } else {
    return HalDevice::getDeviceParameters(device);
  }
}

std::tuple<int, int, int> HalLightDevice::getDeviceParameter(const Device& device, DeviceProperty property) const {
  if (device.getDeviceType() == DEVICE_TYPE_KL && property == DeviceProperty::switchedThreshold) {
    return {3, 0x3C, 1};
  } else {
    return HalDevice::getDeviceParameter(device, property);
  }
}

} // namespace dss::hal
