/*
 *  Copyright (c) 2012 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "base.h"
#include "dss.h"
#include "event/event_create.h"
#include "event/event_fields.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/state.h"
#include "model/zone.h"
#include "propertysystem.h"
#include "security/security.h"
#include "system_states.h"

static ds::LogChannel dsLogChannel("handlerSystemStates");

namespace dss {

namespace StateName {
const std::string Alarm = "alarm";
const std::string Alarm2 = "alarm2";
const std::string Alarm3 = "alarm3";
const std::string Alarm4 = "alarm4";
const std::string BuildingService = "building_service";
const std::string Fire = "fire";
const std::string FireMuteEnabled = "fireMuteEnabled";
const std::string Frost = "frost";
const std::string Hibernation = "hibernation";
const std::string Panic = "panic";
const std::string Motion = "motion";
const std::string OperationLock = "operation_lock";
const std::string Presence = "presence";
const std::string Rain = "rain";
const std::string Wind = "wind";
const std::string Hail = "hail";
const std::string HeatingSystem = "heating_system";
const std::string HeatingSystemMode = "heating_system_mode";
const std::string HeatingModeControl = "heating_water_system";
} // namespace StateName

EventInterpreterPluginSystemState::EventInterpreterPluginSystemState(EventInterpreter *_pInterpreter)
    : EventInterpreterPlugin("system_state", _pInterpreter) {
  SystemState::bootstrap();
}

EventInterpreterPluginSystemState::~EventInterpreterPluginSystemState() {}

void EventInterpreterPluginSystemState::handleEvent(Event &_event, const EventSubscription &_subscription) {
  DS_DEBUG("EventInterpreterPluginSystemState::handleEvent: processing", _event.getName());
  auto state = std::make_shared<SystemState>(_event);
  addEvent(state);
}

SystemState::SystemState(const Event &event) : SystemEvent(event), m_apartment(DSS::getInstance()->getApartment()) {}

SystemState::~SystemState() {}

std::string zoneStateName(const std::string &_name, int _zoneId) {
  // set state in the zone the dsuid is logically attached
  return "zone." + intToString(_zoneId) + "." + _name;
}

std::string SystemState::formatGroupName(const std::string &_name, int _groupId) {
  DS_ASSERT(_groupId >= GroupIDAppUserMin, _groupId);
  return "cluster." + intToString(_groupId) + "." + _name;
}

std::string SystemState::formatAppartmentStateName(const std::string &_name, int _groupId) {
  DS_ASSERT(_groupId >= GroupIDAppUserMin, _groupId);
  return _name + ".group" + intToString(_groupId);
}

std::shared_ptr<State> SystemState::registerState(std::string _name, bool _persistent) {
  auto state = DSS::getInstance()->getApartment().allocateState(StateType_Service, _name, "");
  state->setPersistence(_persistent);
  return state;
}

std::shared_ptr<State> SystemState::getOrRegisterState(std::string _name) {
  std::shared_ptr<State> state;
  if (lookupState(state, _name)) {
    return state;
  }
  return registerState(_name, true);
}

std::shared_ptr<State> SystemState::loadPersistentState(eStateType _type, std::string _name) {
  auto state = m_apartment.tryGetState(_type, _name);
  if (!state) {
    state = m_apartment.allocateState(_type, _name, "");
    if (state->hasPersistentData()) {
      state->setPersistence(true);
      return state;
    } else {
      m_apartment.removeState(state);
    }
  } else {
    if (state->hasPersistentData()) {
      state->setPersistence(true);
    }
    return state;
  }
  return std::shared_ptr<State>();
}

bool SystemState::lookupState(std::shared_ptr<State> &state, const std::string &name) {
  state = m_apartment.tryGetNonScriptState(name);
  return bool(state);
}

void SystemState::callScene(int _zoneId, int _groupId, int _sceneId, CallOrigin _origin) {
  try {
    std::shared_ptr<Zone> z = m_apartment.getZone(_zoneId);
    std::shared_ptr<Group> g = z->getGroup(_groupId);
    g->callScene(_origin, SceneAccessCategory::manual, _sceneId, "", false);
  } catch (ItemNotFoundException &ex) {
  }
}

void SystemState::undoScene(int _zoneId, int _groupId, int _sceneId, CallOrigin _origin) {
  try {
    std::shared_ptr<Zone> z = m_apartment.getZone(_zoneId);
    std::shared_ptr<Group> g = z->getGroup(_groupId);
    g->undoScene(_origin, SceneAccessCategory::manual, _sceneId, "");
  } catch (ItemNotFoundException &ex) {
  }
}

void SystemState::applyFireMuteEnabled() {
  auto fireMuteEnabled = m_apartment.getNonScriptState(StateName::FireMuteEnabled)->getState() == 1;
  for (const auto &dsm : m_apartment.getDSMeters()) {
    try {
      if (dsm->isPresent() && busMemberIsDsmApiDSM(dsm->getBusMemberType())) {
        m_apartment.getBusInterface()->getStructureModifyingBusInterface()->setFireMuteEnabled(dsm->getDSID(),
                                                                                               fireMuteEnabled);
      }
    } catch (const std::exception &e) {
      DS_ERROR("Failed to set fire mute enabled flag.", dsm);
    }
  }
}

void SystemState::bootstrap() {
  auto &apartment = DSS::getInstance()->getApartment();
  std::shared_ptr<State> state;

  state = registerState(StateName::Presence, true);
  // Default presence status is "present" - set default if there is
  // no persistent storage
  if (!state->hasPersistentData()) {
    state->setState(CallOrigin::system, State_Active);
  }

  State::ValueRange_t presenceValues;
  presenceValues.push_back("unknown");
  presenceValues.push_back("present");
  presenceValues.push_back("absent");
  presenceValues.push_back("invalid");
  state->setValueRange(presenceValues);

  state = registerState(StateName::Hibernation, true);

  State::ValueRange_t sleepmodeValues;
  sleepmodeValues.push_back("unknown");
  sleepmodeValues.push_back("awake");
  sleepmodeValues.push_back("sleeping");
  sleepmodeValues.push_back("invalid");
  state->setValueRange(sleepmodeValues);

  registerState(StateName::Alarm, true);
  registerState(StateName::Alarm2, true);
  registerState(StateName::Alarm3, true);
  registerState(StateName::Alarm4, true);
  registerState(StateName::Panic, true);
  registerState(StateName::Fire, true)->setValueRange({State::INVALID, "active", "inactive", "activeMuted"});
  registerState(StateName::FireMuteEnabled, true);
  registerState(StateName::Wind, true);
  registerState(StateName::Rain, true);
  registerState(StateName::Frost, true);

  apartment.allocateState(StateType_Service, "daynight_state", "solar_computer");
  apartment.allocateState(StateType_Service, "twilight_state", "solar_computer");
  apartment.allocateState(StateType_Service, "daylight_state", "solar_computer");
  apartment.allocateState(StateType_Service, "daynight_indoors_state", "solar_computer");

  state = registerState(StateName::HeatingSystem, true);
  if (!state->hasPersistentData()) {
    state->setState(CallOrigin::systemStartup, State_Active);
  }

  state = registerState(StateName::HeatingSystemMode, true);
  if (!state->hasPersistentData()) {
    state->setState(CallOrigin::systemStartup, State_Active); // heating
  }

  state = registerState(StateName::HeatingModeControl, true);
  State::ValueRange_t heatingModeValues;
  heatingModeValues.push_back("off");
  heatingModeValues.push_back("hot water");
  heatingModeValues.push_back("cold water");
  heatingModeValues.push_back("auto");
  state->setValueRange(heatingModeValues);
  if (!state->hasPersistentData()) {
    state->setState(CallOrigin::systemStartup, "hot water");
  }
}

void SystemState::startup() {
  bool panic = false;
  bool alarm = false;
  bool sleeping = false;

  for (std::shared_ptr<Device> device : m_apartment.getDevicesVector()) {
    if (device == NULL) {
      continue;
    }

    for (auto &&input : device->getBinaryInputs()) {
      // motion
      if ((input->m_inputType == BinaryInputType::Movement) ||
          (input->m_inputType == BinaryInputType::MovementInDarkness)) {
        std::string stateName;
        if (input->m_targetGroupId >= GroupIDAppUserMin) {
          stateName = formatGroupName(StateName::Motion, input->m_targetGroupId);
        } else {
          stateName = zoneStateName(StateName::Motion, device->getZoneID());
        }
        getOrRegisterState(stateName);
      }

      // presence
      if ((input->m_inputType == BinaryInputType::Presence) ||
          (input->m_inputType == BinaryInputType::PresenceInDarkness)) {
        std::string stateName;
        if (input->m_targetGroupId >= GroupIDAppUserMin) {
          stateName = formatGroupName(StateName::Presence, input->m_targetGroupId);
        } else {
          stateName = zoneStateName(StateName::Presence, device->getZoneID());
        }
        getOrRegisterState(stateName);
      }

      // wind monitor
      if (input->m_inputType == BinaryInputType::WindDetector) {
        startupApartmentOrClusterSensors(device, input);
      }

      // rain monitor
      if (input->m_inputType == BinaryInputType::RainDetector) {
        startupApartmentOrClusterSensors(device, input);
      }

      // frost detector
      if (input->m_inputType == BinaryInputType::FrostDetector) {
        if (input->m_targetGroupId >= GroupIDAppUserMin) {
          getOrRegisterState(formatAppartmentStateName(StateName::Frost, input->m_targetGroupId));
        }
      }
    } // per device binary inputs for loop
  }   // devices for loop

  for (std::shared_ptr<Zone> zone : m_apartment.getZones()) {
    if ((zone == NULL) || (!zone->isPresent())) {
      continue;
    }

    for (std::shared_ptr<Group> group : zone->getGroups()) {
      if (isCluster(group->getID())) {
        if (group->getApplicationType() == ApplicationType::blinds) {
          registerState(formatAppartmentStateName(StateName::Wind, group->getID()), true);
        }
        continue;
      }

      switch (group->getLastCalledScene()) {
        case ScenePanic: panic = true; break;
        case SceneAlarm: alarm = true; break;
        case SceneSleeping: sleeping = true; break;
      }
    } // groups for loop
  }   // zones for loop

  // Restore apartment states from "lastCalledScene" ...
  std::shared_ptr<State> state;

  if (lookupState(state, StateName::Hibernation)) {
    if ((sleeping == true) && (state->getState() == State_Inactive)) {
      state->setState(CallOrigin::scripting, "sleeping");
    } else if (sleeping == false) {
      state->setState(CallOrigin::scripting, "awake");
    }
  } // hibernation state

  if (lookupState(state, StateName::Panic)) {
    if ((panic == true) && (state->getState() == State_Inactive)) {
      state->setState(CallOrigin::scripting, State_Active);
    } else if (panic == false) {
      state->setState(CallOrigin::scripting, State_Inactive);
    }
  } // panic state

  if (lookupState(state, StateName::Alarm)) {
    if ((alarm == true) && (state->getState() == State_Inactive)) {
      state->setState(CallOrigin::scripting, State_Active);
    } else if (alarm == false) {
      state->setState(CallOrigin::scripting, State_Inactive);
    }
  } // alarm state

  // Restore states if they have been registered before, that reads: if a persistent data file exists
  for (std::shared_ptr<Cluster> cluster : m_apartment.getClusters()) {
    if (cluster->getApplicationType() == ApplicationType::blinds) {
      loadPersistentState(StateType_Service, formatAppartmentStateName(StateName::Wind, cluster->getID()));
    }
    loadPersistentState(StateType_Group, formatAppartmentStateName(StateName::OperationLock, cluster->getID()));
  }
  loadPersistentState(StateType_Service, StateName::BuildingService);

// clear fire alarm after 6h
#define CLEAR_ALARM_URLENCODED_JSON                                                                                    \
  "%7B%20%22name%22%3A%22FireAutoClear%22%2C%20%22id%22%3A%20%22system_state_fire_alarm_reset%22%2C%22triggers%22%3A%" \
  "5B%7B%20%22type%22%3A%22state-change%22%2C%20%22name%22%3A%22fire%22%2C%20%22state%22%3A%22active%22%7D%5D%2C%"     \
  "22delay%22%3A21600%2C%22actions%22%3A%5B%7B%20%22type%22%3A%22undo-zone-scene%22%2C%20%22zone%22%3A0%2C%20%"        \
  "22group%22%3A0%2C%20%22scene%22%3A76%2C%20%22force%22%3A%22false%22%2C%20%22delay%22%3A0%20%7D%5D%2C%22conditions%" \
  "22%3A%7B%20%22enabled%22%3Anull%2C%22weekdays%22%3Anull%2C%22timeframe%22%3Anull%2C%22zoneState%22%3Anull%2C%"      \
  "22systemState%22%3A%5B%7B%22name%22%3A%22fire%22%2C%22value%22%3A%221%22%7D%5D%2C%22addonState%22%3Anull%7D%2C%"    \
  "22scope%22%3A%22system_state.auto_cleanup%22%7D"
  /*
    var t0 = 6 * 60 * 60;
    var sr = '{ "name":"FireAutoClear", "id": "system_state_fire_alarm_reset",\
      "triggers":[{ "type":"state-change", "name":"fire", "state":"active"}],\
      "delay":' + t0 + ',\
      "actions":[{ "type":"undo-zone-scene", "zone":0, "group":0, "scene":76, "force":"false", "delay":0 }],\
      "conditions":{ "enabled":null,"weekdays":null,"timeframe":null,"zoneState":null,\
        "systemState":[{"name":"fire","value":"1"}],\
        "addonState":null},\
      "scope":"system_state.auto_cleanup"\
      }';
  */
  std::shared_ptr<Event> event = std::make_shared<Event>("system-addon-scene-responder.config");
  event->setProperty("actions", "save");
  event->setProperty("value", CLEAR_ALARM_URLENCODED_JSON);
  DSS::getInstance()->getEventQueue().pushEvent(event);

  applyFireMuteEnabled();
}

std::string SystemState::getData(int *zoneId, int *groupId, int *sceneId, CallOrigin *callOrigin) {
  std::string originDSUID;

  *callOrigin = CallOrigin::unknown;
  if (m_properties.has(ef_callOrigin)) {
    *callOrigin = (CallOrigin)strToIntDef(m_properties.get(ef_callOrigin), 0);
  }

  if (!zoneId || !groupId || !sceneId) {
    return originDSUID;
  }

  *zoneId = -1;
  *groupId = -1;
  *sceneId = -1;

  if (m_properties.has("sceneID")) {
    *sceneId = strToIntDef(m_properties.get("sceneID"), -1);
  }

  if (m_properties.has(ef_originDSUID)) {
    originDSUID = m_properties.get(ef_originDSUID);
  }

  switch (m_evtRaiseLocation) {
    case erlGroup:
    case erlApartment:
      if (m_raisedAtGroup != NULL) {
        *zoneId = m_raisedAtGroup->getZoneID();
        *groupId = m_raisedAtGroup->getID();
      }
      break;
    case erlState:
      if (m_raisedAtState != NULL) {
        if (m_raisedAtState->getType() == StateType_Device) {
          std::weak_ptr<Device> weakDevice = m_raisedAtState->getProviderDevice();
          if (std::shared_ptr<Device> device = weakDevice.lock()) {
            *zoneId = device->getZoneID();
            originDSUID = dsuid2str(device->getDSID());
          }
        } else if (m_raisedAtState->getType() == StateType_Group) {
          std::weak_ptr<Group> weakGroup = m_raisedAtState->getProviderGroup();
          if (std::shared_ptr<Group> group = weakGroup.lock()) {
            *zoneId = group->getZoneID();
            *groupId = group->getID();
          }
        }
      }
      break;
    case erlDevice:
      if (m_raisedAtDevice != NULL) {
        originDSUID = dsuid2str(m_raisedAtDevice->getDSID());
      }
      break;
  }

  return originDSUID;
}

void SystemState::callscene() {
  int zoneId = -1;
  int groupId = -1;
  int sceneId = -1;
  CallOrigin callOrigin = CallOrigin::unknown;

  getData(&zoneId, &groupId, &sceneId, &callOrigin);

  if (callOrigin == CallOrigin::system) {
    // ignore scene calls originated by server generated system level events
    //  e.g. scene calls issued by state changes
    return;
  }

  std::shared_ptr<State> state;

  if (groupId == 0) {
    switch (sceneId) {
      case SceneAbsent:
        if (lookupState(state, StateName::Presence)) {
          state->setState(CallOrigin::system, "absent");
        }
        // #2561: auto-clear panic and fire
        if (lookupState(state, StateName::Panic)) {
          if (state->getState() == State_Active) {
            undoScene(0, 0, ScenePanic, CallOrigin::system);
          }
        }
        if (lookupState(state, StateName::Fire)) {
          if (state->getState() == State_Active) {
            undoScene(0, 0, SceneFire, CallOrigin::system);
          }
        }
        break;
      case ScenePresent:
        if (lookupState(state, StateName::Presence)) {
          state->setState(CallOrigin::system, "present");
        }
        break;
      case SceneSleeping:
        if (lookupState(state, StateName::Hibernation)) {
          state->setState(CallOrigin::system, "sleeping");
        }
        break;
      case SceneWakeUp:
        if (lookupState(state, StateName::Hibernation)) {
          state->setState(CallOrigin::system, "awake");
        }
        break;
      case ScenePanic:
        if (lookupState(state, StateName::Panic)) {
          state->setState(CallOrigin::system, State_Active);
        }
        break;
      case SceneFire:
        if (lookupState(state, StateName::Fire)) {
          state->setState(CallOrigin::system, State_Active);
        }
        break;
      case SceneAlarm:
        if (lookupState(state, StateName::Alarm)) {
          state->setState(CallOrigin::system, State_Active);
        }
        break;
      case SceneAlarm2:
        if (lookupState(state, StateName::Alarm2)) {
          state->setState(CallOrigin::system, State_Active);
        }
        break;
      case SceneAlarm3:
        if (lookupState(state, StateName::Alarm3)) {
          state->setState(CallOrigin::system, State_Active);
        }
        break;
      case SceneAlarm4:
        if (lookupState(state, StateName::Alarm4)) {
          state->setState(CallOrigin::system, State_Active);
        }
        break;
      case SceneWindActive:
        state = getOrRegisterState(StateName::Wind);
        state->setState(CallOrigin::system, State_Active);
        break;
      case SceneWindInactive:
        state = getOrRegisterState(StateName::Wind);
        state->setState(CallOrigin::system, State_Inactive);
        break;
      case SceneRainActive:
        state = getOrRegisterState(StateName::Rain);
        state->setState(CallOrigin::system, State_Active);
        break;
      case SceneRainInactive:
        state = getOrRegisterState(StateName::Rain);
        state->setState(CallOrigin::system, State_Inactive);
        // cluster sensor and the global rain sensor could contradict each other,
        // and we are not resolving that conflict. Currently we support either
        // global or cluster-local states, but not both at the same time.
        // In the presence of cluster rain states inactive-rain on group0 serves
        // as a broadcast command only
        for (int grp = GroupIDAppUserMin; grp <= GroupIDAppUserMax; grp++) {
          if (lookupState(state, formatAppartmentStateName(StateName::Rain, grp))) {
            state->setState(CallOrigin::system, State_Inactive);
          }
        }
        break;
    }
  } else if (isCluster(groupId)) {
    /*
     * wind/rain can be apartment wide or only to one facade
     */
    switch (sceneId) {
      case SceneWindActive:
        state = getOrRegisterState(formatAppartmentStateName(StateName::Wind, groupId));
        state->setState(CallOrigin::system, State_Active);
        break;
      case SceneWindInactive:
        state = getOrRegisterState(formatAppartmentStateName(StateName::Wind, groupId));
        state->setState(CallOrigin::system, State_Inactive);
        break;
      case SceneRainActive:
        state = getOrRegisterState(formatAppartmentStateName(StateName::Rain, groupId));
        state->setState(CallOrigin::system, State_Active);
        break;
      case SceneRainInactive:
        state = getOrRegisterState(formatAppartmentStateName(StateName::Rain, groupId));
        state->setState(CallOrigin::system, State_Inactive);
        break;
    }
  }
}

void SystemState::undoscene() {
  int zoneId = -1;
  int groupId = -1;
  int sceneId = -1;
  CallOrigin callOrigin = CallOrigin::unknown;
  std::string originDSUID;
  std::shared_ptr<State> state;

  originDSUID = getData(&zoneId, &groupId, &sceneId, &callOrigin);
  dsuid_t dsuid = str2dsuid(originDSUID);

  if (groupId != 0) {
    // panic/fire/alarm apply to broadcast group-0
    // no undoscene for wind/rain
    return;
  }

  DS_ASSERT(groupId == 0, groupId);
  switch (sceneId) {
    case ScenePanic:
      if (lookupState(state, StateName::Panic)) {
        state->setState(CallOrigin::system, State_Inactive);
      }

      // #2561: auto-reset fire if panic was reset by a button
      if (lookupState(state, StateName::Fire)) {
        if ((dsuid != DSUID_NULL) && (callOrigin == CallOrigin::dsmApi) && (state->getState() == State_Active)) {
          undoScene(0, 0, SceneFire, CallOrigin::system);
        } // valid dSID && dSM API origin && state == active
      }
      break;
    case SceneFire:
      if (lookupState(state, StateName::Fire)) {
        state->setState(CallOrigin::system, State_Inactive);
      }

      // #2561: auto-reset panic if fire was reset by a button
      if (lookupState(state, StateName::Panic)) {
        if ((dsuid != DSUID_NULL) && (callOrigin == CallOrigin::dsmApi) && (state->getState() == State_Active)) {
          undoScene(0, 0, ScenePanic, CallOrigin::system);
        } // valid dSID && dSM API origin && state == active
      }
      break;
    case SceneAlarm:
      if (lookupState(state, StateName::Alarm)) {
        state->setState(CallOrigin::system, State_Inactive);
      }
      break;
    case SceneAlarm2:
      if (lookupState(state, StateName::Alarm2)) {
        state->setState(CallOrigin::system, State_Inactive);
      }
      break;
    case SceneAlarm3:
      if (lookupState(state, StateName::Alarm3)) {
        state->setState(CallOrigin::system, State_Inactive);
      }
      break;
    case SceneAlarm4:
      if (lookupState(state, StateName::Alarm4)) {
        state->setState(CallOrigin::system, State_Inactive);
      }
      break;
  }
}

/// logical or combination of multiple binary inputs
/// if one of AKM input is high(=2) combined output is on
/// if all are low(=1) combined output is off
/// e.g. multiple presence detectors
void SystemState::stateBinaryInputGeneric(State &_state, int targetGroupId) {
  try {
    std::string stateName = _state.getName();
    auto &&nodePath = "/scripts/system_state/" + stateName + "." + intToString(targetGroupId);
    PropertyNodePtr pNode = DSS::getInstance()->getPropertySystem().getProperty(nodePath);
    if (pNode == NULL) {
      pNode = DSS::getInstance()->getPropertySystem().createProperty(nodePath);
      pNode->setIntegerValue(0);
    }

    if (m_properties.has("value")) {
      std::weak_ptr<Device> weakDevice = m_raisedAtState->getProviderDevice();
      if (std::shared_ptr<Device> pDev = weakDevice.lock()) {
        std::string val = m_properties.get("value");
        int iVal = strToIntDef(val, -1);
        if (iVal == 1) {
          pNode->setIntegerValue(pNode->getIntegerValue() + 1);
          _state.setState(CallOrigin::binaryInput, State_Active);
          _state.setOriginDeviceDSUID(pDev->getDSID());
        } else if (iVal == 2) {
          if (pNode->getIntegerValue() > 0) {
            pNode->setIntegerValue(pNode->getIntegerValue() - 1);
          }
          if (pNode->getIntegerValue() == 0) {
            _state.setState(CallOrigin::binaryInput, State_Inactive);
          }
          _state.setOriginDeviceDSUID(pDev->getDSID());
        }
      }
    } // m_properties.has("value")
  } catch (ItemNotFoundException &ex) {
  }
}

void SystemState::stateBinaryinput() {
  if (m_raisedAtState == NULL) {
    return;
  }
  std::weak_ptr<Device> weakDevice = m_raisedAtState->getProviderDevice();
  std::shared_ptr<Device> pDev = weakDevice.lock();
  if (!pDev) {
    return;
  }

  if (!m_properties.has("statename")) {
    return;
  }

  std::string statename = m_properties.get("statename");
  if (statename.empty()) {
    return;
  }

  PropertyNodePtr stateNode = DSS::getInstance()->getPropertySystem().getProperty("/usr/states/" + statename);
  if (stateNode == NULL) {
    return;
  }

  PropertyNodePtr iiNode = stateNode->getProperty("device/inputIndex");
  if (iiNode == NULL) {
    return;
  }

  uint8_t inputIndex = static_cast<uint8_t>(iiNode->getIntegerValue());
  auto &&devInput = pDev->getBinaryInput(inputIndex);

  if (devInput->m_inputId != BinaryInputId::APP_MODE) {
    return;
  }

  switch (devInput->m_inputType) {
    case BinaryInputType::AppMode:
    case BinaryInputType::RoomBrightness:
    case BinaryInputType::TwilightExternal:
    case BinaryInputType::SunRadiation:
    case BinaryInputType::BatteryLow:
    case BinaryInputType::WindowContact:
    case BinaryInputType::DoorContact:
    case BinaryInputType::WindowTilt:
    case BinaryInputType::GarageDoorContact:
    case BinaryInputType::SunProtection:
    case BinaryInputType::PowerUp:
    case BinaryInputType::Malfunction:
    case BinaryInputType::Service: return;

    case BinaryInputType::Movement:
    case BinaryInputType::MovementInDarkness:
    case BinaryInputType::Presence:
    case BinaryInputType::PresenceInDarkness: {
      if (devInput->m_targetGroupId >= GroupIDAppUserMin) {
        statename = formatGroupName(selectStateName(devInput->m_inputType), devInput->m_targetGroupId);
      } else {
        statename = zoneStateName(selectStateName(devInput->m_inputType), pDev->getZoneID());
      }
      std::shared_ptr<State> state = getOrRegisterState(statename);
      stateBinaryInputGeneric(*state, devInput->m_targetGroupId);
    } break;

    case BinaryInputType::SmokeDetector: {
      std::shared_ptr<State> state;
      if (lookupState(state, StateName::Fire)) {
        if (m_properties.has("value")) {
          std::string val = m_properties.get("value");
          int iVal = strToIntDef(val, -1);
          if (iVal == 1) {
            state->setState(CallOrigin::binaryInput, State_Active);
          }
        }
      }
    } break;

    case BinaryInputType::WindDetector:
    case BinaryInputType::RainDetector: assignApartmentOrClusterSensors(pDev, inputIndex); break;

    case BinaryInputType::RoomThermostat: {
      if (m_properties.has("value")) {
        std::string val = m_properties.get("value");
        int iVal = strToIntDef(val, -1);
        int sceneID = SceneOff;
        if (iVal == 1) {
          sceneID = Scene1;
        }
        callScene(pDev->getZoneID(), GroupIDHeating, sceneID, CallOrigin::binaryInput);
      }
    } break;

    case BinaryInputType::FrostDetector: assignGlobalSensors(pDev, inputIndex); break;

    case BinaryInputType::HeatingSystem:
    case BinaryInputType::HeatingSystemMode: {
      assignGlobalSensors(pDev, inputIndex);

      std::shared_ptr<State> heating;
      std::shared_ptr<State> heating_mode;

      if (lookupState(heating, StateName::HeatingSystem) && lookupState(heating_mode, StateName::HeatingSystemMode)) {
        // abuse auto as an invalid value
        HeatingModeSwitchValue value = HeatingModeSwitchValue::AUTO;

        if (heating->getState() == State_Inactive) {
          value = HeatingModeSwitchValue::OFF;
        } else if (heating->getState() == State_Active) {
          if (heating_mode->getState() == State_Active) {
            value = HeatingModeSwitchValue::HEATING;
          } else if (heating_mode->getState() == State_Inactive) {
            value = HeatingModeSwitchValue::COOLING;
          }
        }

        if (value != HeatingModeSwitchValue::AUTO) {
          auto event = createGenericSignalHeatingModeSwitch(value, CallOrigin::binaryInput);
          DSS::getInstance()->getEventQueue().pushEvent(event);
        }
      }
    } break;
  }
}

void SystemState::stateApartment() {
  if (!m_properties.has("statename")) {
    return;
  }

  std::string statename = m_properties.get("statename");
  if (statename.empty()) {
    return;
  }

  int zoneId;
  int groupId;
  int sceneId;
  CallOrigin callOrigin = CallOrigin::unknown;

  getData(&zoneId, &groupId, &sceneId, &callOrigin);

  groupId = 0;
  size_t groupName = statename.find(".group");
  if (groupName != std::string::npos) {
    std::string id = statename.substr(groupName + 6);
    groupId = strToIntDef(id, 0);
  }

  if (callOrigin == CallOrigin::system) {
    // ignore state change originated by server generated system level events,
    // e.g. scene calls issued by state changes, to avoid loops
    return;
  }

  int iVal = -1;
  if (m_properties.has("value")) {
    std::string val = m_properties.get("value");
    iVal = strToIntDef(val, -1);
  }

  if (statename == StateName::Fire) {
    if (iVal == 1) {
      callScene(0, 0, SceneFire, CallOrigin::system);
    } else if (iVal == 2) {
      undoScene(0, 0, SceneFire, CallOrigin::system);
    }
  }

  if (statename == StateName::FireMuteEnabled) {
    applyFireMuteEnabled();
  }

  if (statename.substr(0, 4) == StateName::Rain) {
    if (iVal == 1) {
      callScene(0, groupId, SceneRainActive, CallOrigin::system);
    } else if (iVal == 2) {
      callScene(0, groupId, SceneRainInactive, CallOrigin::system);
    }
  }

  if (statename.substr(0, 4) == StateName::Wind) {
    if (iVal == 1) {
      callScene(0, groupId, SceneWindActive, CallOrigin::system);
    } else if (iVal == 2) {
      callScene(0, groupId, SceneWindInactive, CallOrigin::system);
    }
  }

  if (statename.substr(0, 4) == StateName::Hail) {
    if (iVal == 1) {
      callScene(0, 0, SceneHailActive, CallOrigin::system);
    } else if (iVal == 2) {
      callScene(0, 0, SceneHailInactive, CallOrigin::system);
    }
  }
}

void SystemState::assignApartmentOrClusterSensors(std::shared_ptr<Device> pDev, std::size_t inputIndex) {
  auto &&devInput = pDev->getBinaryInput(inputIndex);
  std::shared_ptr<State> state;
  // create state for a user group if it does not exist (new group?)
  if (devInput->m_targetGroupId >= GroupIDAppUserMin || pDev->getGroupsCount() >= 2) {
    for (std::size_t i = 0; i < pDev->getGroupIds().size(); ++i) {
      if (pDev->getGroupIdByIndex(i) < GroupIDAppUserMin) {
        continue;
      }
      std::string statename =
          formatAppartmentStateName(selectStateName(devInput->m_inputType), pDev->getGroupIds().at(i));
      state = getOrRegisterState(statename);
      stateBinaryInputGeneric(*state, pDev->getGroupIds().at(i));
    }
  } else if (lookupState(state, selectStateName(devInput->m_inputType))) {
    stateBinaryInputGeneric(*state, devInput->m_targetGroupId);
  }
}

void SystemState::startupApartmentOrClusterSensors(std::shared_ptr<Device> device,
                                                   std::shared_ptr<DeviceBinaryInput> input) {
  if (input->m_targetGroupId >= GroupIDAppUserMin || device->getGroupsCount() >= 2) {
    for (std::size_t i = 0; i < device->getGroupIds().size(); ++i) {
      if (device->getGroupIdByIndex(i) < GroupIDAppUserMin) {
        continue;
      }
      getOrRegisterState(formatAppartmentStateName(selectStateName(input->m_inputType), device->getGroupIds().at(i)));
    }
  }
}

void SystemState::assignGlobalSensors(std::shared_ptr<Device> pDev, std::size_t inputIndex) {
  auto &&devInput = pDev->getBinaryInput(inputIndex);
  std::shared_ptr<State> state;
  if (lookupState(state, selectStateName(devInput->m_inputType))) {
    stateBinaryInputGeneric(*state, devInput->m_targetGroupId);
  }
}

std::string SystemState::selectStateName(BinaryInputType m_inputType) {
  switch (m_inputType) {
    case BinaryInputType::RainDetector: return StateName::Rain; break;

    case BinaryInputType::WindDetector: return StateName::Wind; break;

    case BinaryInputType::FrostDetector: return StateName::Frost; break;

    case BinaryInputType::SmokeDetector: return StateName::Fire; break;

    case BinaryInputType::HeatingSystem: return StateName::HeatingSystem; break;

    case BinaryInputType::HeatingSystemMode: return StateName::HeatingSystemMode; break;

    case BinaryInputType::Movement:
    case BinaryInputType::MovementInDarkness: return StateName::Motion; break;

    case BinaryInputType::Presence:
    case BinaryInputType::PresenceInDarkness: return StateName::Presence; break;

    case BinaryInputType::AppMode:
    case BinaryInputType::RoomBrightness:
    case BinaryInputType::TwilightExternal:
    case BinaryInputType::SunRadiation:
    case BinaryInputType::RoomThermostat:
    case BinaryInputType::BatteryLow:
    case BinaryInputType::WindowContact:
    case BinaryInputType::DoorContact:
    case BinaryInputType::WindowTilt:
    case BinaryInputType::GarageDoorContact:
    case BinaryInputType::SunProtection:
    case BinaryInputType::PowerUp:
    case BinaryInputType::Malfunction:
    case BinaryInputType::Service: return ""; break;

    default: return ""; break;
  }
}

void SystemState::run() {
  if (DSS::hasInstance()) {
    DSS::getInstance()->getSecurity().loginAsSystemUser("SystemState needs system rights");
  } else {
    return;
  }
  DS_INFO_ENTER(m_evtName, m_raisedAtState, m_properties);

  try {
    if (m_evtName == EventName::ModelReady) {
      startup();
    } else if (m_evtName == EventName::CallScene) {
      if ((m_evtRaiseLocation == erlGroup) && (m_raisedAtGroup != NULL)) {
        callscene();
      }
    } else if (m_evtName == EventName::UndoScene) {
      undoscene();
    } else if (m_evtName == EventName::StateChange) {
      if (m_raisedAtState->getType() == StateType_Device) {
        stateBinaryinput();
      } else if (m_raisedAtState->getType() == StateType_Service) {
        stateApartment();
      }

    } else if (m_evtName == EventName::Sunshine) {
      std::string value = m_properties.get("value");
      std::string direction = m_properties.get("direction");

      // TODO: define algorithm for system states based on sun{shine,protection} inputs

    } else if (m_evtName == EventName::FrostProtection) {
      unsigned int value = strToInt(m_properties.get("value"));
      DS_ASSERT(value == State_Active || value == State_Inactive, value);
      getOrRegisterState(StateName::Frost)->setState(CallOrigin::dsmApi, value);

    } else if (m_evtName == EventName::HeatingModeSwitch) {
      auto state = getOrRegisterState(StateName::HeatingModeControl);
      auto value = static_cast<HeatingModeSwitchValue>(strToUInt(m_properties.get("value")));
      CallOrigin origin = CallOrigin::dsmApi;
      if (m_properties.has("callOrigin")) {
        std::string s = m_properties.get("callOrigin");
        if (!s.empty()) {
          origin = (CallOrigin)strToInt(s);
        }
      }
      state->setState(origin, static_cast<int>(value));

    } else if (m_evtName == EventName::BuildingService) {
      unsigned int value = strToInt(m_properties.get("value"));
      DS_ASSERT(value == State_Active || value == State_Inactive, "");
      getOrRegisterState(StateName::BuildingService)->setState(CallOrigin::dsmApi, value);
    }
  } catch (const std::exception &ex) {
    DS_WARNING("failed to process event:", m_evtName, ex.what());
  }
}

} // namespace dss
