/*
    Copyright (c) 2009,2010,2011 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "archives.h"

#include <digitalSTROM/dsuid.h>
#include <ds/conf.h>
#include <ds/log.h>
#include <ds/reflect.h>
#include <ds/system.h>

#include <boost/filesystem.hpp>
#include <regex>
#include <vector>

#include "metering/config.h"
#include "metering/rrd-interface.h"

static ds::LogChannel dsLogChannel("dssMeteringArchives");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

namespace {

constexpr auto MAX_INTERPOLATION_INTERVAL = 7 * 24h;         // data not interpolated, if interval is longer
constexpr int DISK_FLUSH_INTERVAL = 10 * 60;                 // ten minutes

auto createArchiveConfig() {
  return std::vector{MeteringConfig{1s, 600},  MeteringConfig{60s, 720},     MeteringConfig{15min, 2976},
                     MeteringConfig{24h, 370}, MeteringConfig{7 * 24h, 260}, MeteringConfig{30 * 24h, 60}};
}

class ArchivesImpl : public Archives {
  ds::System& m_system;
  std::string m_meteringStorageLocation;

public:
  ArchivesImpl(std::shared_ptr<RRDInterface> rrdInterface, ds::System& system);
  ~ArchivesImpl() override = default;

  const std::vector<MeteringConfig> getConfig() const override { return m_Config; }

  std::string filename(const ArchiveId& archiveId) const override {
    return ds::str(m_meteringStorageLocation, std::get<Dsuid>(archiveId),
                   (std::get<Category>(archiveId) == Category::consumption) ? "" : "-production", ".rrd");
  }
  bool hasArchive(const ArchiveId& archiveId) const override { return (m_system.existsFile(filename(archiveId))); }

  void create(const ArchiveId& archiveId) override;
  void addValue(const ArchiveId& archiveId, ArchiveValue&& value) override;
  std::optional<ArchiveValue> lastValue(const ArchiveId& archiveId) override;

  ValueRange xport(time_t start, time_t end, int step, std::vector<ArchiveId> archiveIds, SeriesTypes type,
                   bool energyInWh) override;

  bool validateDBSeries(const ArchiveId& archiveId, bool& _tunePowerMaxSetting) override;
  void tuneDBPowerSettings(const ArchiveId& archiveId) override;
  bool withoutUpdates(const ArchiveId& archiveId) const override;
  bool checkDBFlush(const DateTime& _sampledAt, const ArchiveId& archiveId) const override;
  void flush(std::vector<ArchiveId> archiveIds) const override;
  void restartCached() const override;

  void setRRDCachedPath(std::string rrdCachedPath) override { m_RrdcachedPath = rrdCachedPath; }
  void setMeteringStorageLocation(std::string&& meteringStorageLocation) override {
    m_meteringStorageLocation = std::move(meteringStorageLocation);
  }

  auto limitPower(const ArchiveId& archiveId) const {
    if (archiveId.second == Category::consumption) {
      return m_electricityMeteringConfig.discontinuityDetection.maxConsumption;
    } else {
      return m_electricityMeteringConfig.discontinuityDetection.maxProduction;
    }
  }
  auto limitEnergy(const ArchiveId& archiveId) const {
    return limitPower(archiveId); // Ws value same as W
  }

private:
  std::vector<MeteringConfig> m_Config;
  std::shared_ptr<RRDInterface> m_rrdInterface;
  std::string m_RrdcachedPath;
  const ElectricityMeteringConfig m_electricityMeteringConfig;
  std::unordered_map<int, int> m_nonFiniteValueHistogram{}; // rate -> count
};

ArchivesImpl::ArchivesImpl(std::shared_ptr<RRDInterface> rrdInterface, ds::System& system)
    : m_system(system),
      m_Config(createArchiveConfig()),
      m_rrdInterface(std::move(rrdInterface)),
      m_electricityMeteringConfig(m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {})) {}

bool ArchivesImpl::validateDBSeries(const ArchiveId& archiveId, bool& _tunePowerMaxSetting) {
  static const auto rraIndexRegex = std::regex(R"(rra\[(\d+)\].rows)");
  unsigned long rrdMatchCount = 0;

  for (auto& rrdInfo : m_rrdInterface->info(filename(archiveId))) {
    DS_DEBUG(rrdInfo);
    if ((rrdInfo.key == "ds[power].index" && (rrdInfo.type == RrdInfoType::counter) &&
         (std::get<counter_t>(rrdInfo.value) == 0)) ||
        (rrdInfo.key == "ds[energy].index" && (rrdInfo.type == RrdInfoType::counter) &&
         (std::get<counter_t>(rrdInfo.value) == 1))) {
      rrdMatchCount++;
    } else {
      std::smatch match;
      if (std::regex_search(rrdInfo.key, match, rraIndexRegex)) {
        auto index = std::stoul(match[1]);
        DS_REQUIRE(index < m_Config.size(), "RRA index exceeded");
        if (std::get<counter_t>(rrdInfo.value) == m_Config[index].numberOfValues) {
          rrdMatchCount++;
        }
      }
    }

    if (rrdInfo.key == "ds[power].max") {
      DS_DEBUG(rrdInfo);
      if ((rrdInfo.type == RrdInfoType::value) && (std::get<rrd_value_t>(rrdInfo.value) != limitPower(archiveId))) {
        DS_NOTICE("validateDBSeries: invalid upper power limit", rrdInfo, archiveId);
        _tunePowerMaxSetting = true;
      }
    }
  }

  DS_DEBUG("RRD MatchCount", rrdMatchCount, _tunePowerMaxSetting);
  return (rrdMatchCount == (2 + m_Config.size()));
}

void ArchivesImpl::tuneDBPowerSettings(const ArchiveId& archiveId) {
  DS_WARNING("tuning max acceptable power value in RRD", archiveId);
  std::vector<std::string> lines;
  lines.push_back("tune");
  lines.push_back(filename(archiveId));
  lines.push_back("--maximum");
  lines.push_back(ds::str("power:", limitPower(archiveId)));

  m_rrdInterface->tune(lines);
}

void ArchivesImpl::addValue(const ArchiveId& archiveId, ArchiveValue&& value) {
  DS_CONTEXT(ds::hideArg(archiveId));
  DS_REQUIRE(archiveId != ArchiveId{}, "addValue invalid archive descriptor", archiveId);

  std::vector<std::string> lines;
  lines.push_back("update");
  if (!m_RrdcachedPath.empty()) {
    lines.push_back("--daemon");
    lines.push_back(m_RrdcachedPath);
  }
  lines.push_back(filename(archiveId));
  // ds::energy:DERIVE -- only accepts integer values
  lines.push_back(ds::str(std::chrono::system_clock::to_time_t(value.lastChanged), ":", value.power, ":",
                          gsl::narrow_cast<uint64_t>(std::ceil(value.energy))));

  DS_DEBUG(lines);
  m_rrdInterface->update(lines);
} // postEnergyMeasurement

std::optional<ArchiveValue> ArchivesImpl::lastValue(const ArchiveId& archiveId) {
  try {
    auto [lastChanged, data] = m_rrdInterface->lastValue(filename(archiveId));

    auto ret = ArchiveValue{.lastChanged = ds::DateTime{std::chrono::seconds{lastChanged}}};
    for (const auto& [name, value] : data) {
      if (name == "energy") {
        ret.energy = value;
      } else if (name == "power") {
        ret.power = value;
      }
    }
    // TODO rely on validateDBSeries to verify that energy and power data source is present
    return ret;

  } catch (const std::exception& e) {
    // probably unknown value in db, conversion to double failed
    DS_NOTICE("tryCreateCacheFromDB", e);
  }

  return std::nullopt;
}

void ArchivesImpl::create(const ArchiveId& archiveId) {
  DS_DEBUG("Creating new RRD database", archiveId);

  /* create new DB */
  std::vector<std::string> lines;
  lines.push_back(
      ds::str("DS:power:GAUGE:5:0:", limitPower(archiveId))); // ignored by unit tests on host, but enforced on platform
  lines.push_back("DS:energy:DERIVE:5:0:U");

  for (const auto& archiveConfig : m_Config) {
    std::stringstream sstream;
    sstream << "RRA:AVERAGE:0.5:" << std::chrono::seconds(archiveConfig.resolution).count() << ":"
            << archiveConfig.numberOfValues;
    lines.push_back(sstream.str());
  }

  DateTime iCurrentTimeStamp;
  time_t currentInSecs = iCurrentTimeStamp.secondsSinceEpoch();
  if (currentInSecs > 10) {
    currentInSecs -= 10;
  }

  m_rrdInterface->create(filename(archiveId), 1, currentInSecs, lines);
}

void ArchivesImpl::restartCached() const {
  if (!m_RrdcachedPath.empty()) {
    DS_NOTICE("Restarting rrdcached....");
    /* TODO sv restart rrdcached might hang */
    if (system("killall -9 rrdcached") != 0) {
      DS_WARNING("Restarting rrdcached was not successful.");
    }
  }
}

bool ArchivesImpl::withoutUpdates(const ArchiveId& archiveId) const {
  auto lastAccess = m_rrdInterface->lastTimestamp(filename(archiveId));
  if (!lastAccess) {
    // probably invalid file format, rely on later checks to fail as well
    return false;
  }
  auto lastUpdateAt = std::chrono::system_clock::from_time_t(*lastAccess);

  // If new sample time is more than MAX_INTERPOLATION_INTERVAL newer than last
  // db entry, it would take too long to roll forward. (more than 15 minutes!)
  // If the the new sample ist older than latest sample in the database,
  // rrd will silently ignore it (ntp update), and no data will be stored
  // until time is going forward again.
  // If the jump into the past is too big discard all data and start over.

  auto now = std::chrono::system_clock::now();
  if (std::chrono::abs(now - lastUpdateAt) > MAX_INTERPOLATION_INTERVAL) {
    DS_NOTICE("withoutUpdates", (now - lastUpdateAt), MAX_INTERPOLATION_INTERVAL);
    return true;
  }

  return false;
}

bool ArchivesImpl::checkDBFlush(const DateTime& sampledAt, const ArchiveId& archiveId) const {
  auto lastTimestamp = m_rrdInterface->lastTimestamp(filename(archiveId));
  if (!lastTimestamp) {
    // not accessed for a long time
    int deltaTime = sampledAt.difference(DateTime(-1));
    return (deltaTime > DISK_FLUSH_INTERVAL);
  }

  int deltaTime = sampledAt.difference(DateTime(*lastTimestamp));
  return (deltaTime > DISK_FLUSH_INTERVAL);
}

// TODO I guess single archiveId as interface would do would, design problem?
void ArchivesImpl::flush(std::vector<ArchiveId> archiveIds) const {
  if (m_RrdcachedPath.empty()) {
    DS_DEBUG("flush: no rrdcached path set");
    return;
  }

  std::vector<std::string> lines;
  lines.push_back("flushcached");
  lines.push_back("--daemon");
  lines.push_back(m_RrdcachedPath);

  for (const auto& archiveId : archiveIds) {
    lines.push_back(filename(archiveId));
  }

  DS_DEBUG("flushing cached rrd data to file");
  m_rrdInterface->flush(lines);
}

ValueRange ArchivesImpl::xport(time_t start, time_t end, int step, std::vector<ArchiveId> archiveIds, SeriesTypes type,
                               bool energyInWh) {
  std::vector<std::string> lines;
  lines.push_back("xport");
  if (!m_RrdcachedPath.empty()) {
    lines.push_back("--daemon");
    lines.push_back(m_RrdcachedPath);
  }
  lines.push_back("--start");
  lines.push_back(ds::str(start));
  lines.push_back("--end");
  lines.push_back(ds::str(end));
  lines.push_back("--step");
  lines.push_back(ds::str(step));
  lines.push_back("--maxrows");
  lines.push_back("3000");
  {
    std::stringstream sstream;
    for (size_t i = 0; i < archiveIds.size(); ++i) {
      auto archiveId = archiveIds[i];
      sstream << "DEF:raw" << i << "=" << filename(archiveId);
      if ((step == 1) && ((type == SeriesTypes::power))) {
        sstream << ":power";
      } else {
        sstream << ":energy";
      }

      sstream << ":AVERAGE";
      lines.push_back(sstream.str());

      sstream.str(std::string());
      sstream << "CDEF:data" << i << "=raw" << i << ",UN,0,raw" << i << ",IF";
      lines.push_back(sstream.str());
      sstream.str(std::string());
    }
    sstream << "CDEF:sum=data0";
    for (size_t i = 1; i < archiveIds.size(); ++i) {
      sstream << ",data" << i << ",+";
    }
    lines.push_back(sstream.str());
  }
  switch (type) {
    case SeriesTypes::energy:
    case SeriesTypes::energyDelta: {
      lines.push_back(ds::str("CDEF:ratePerStep=sum,", step, ",*"));
      if (energyInWh) {
        lines.push_back("CDEF:adj=ratePerStep,3600,/");
        lines.push_back("XPORT:adj");
      } else {
        lines.push_back("XPORT:ratePerStep");
      }
      break;
    }
    case SeriesTypes::power:
    default: {
      lines.push_back("XPORT:sum");
      break;
    }
  }

  auto results = m_rrdInterface->xport(lines);
  // DS_REQUIRE(start, end, step) -> move to rrd-interface

  for (auto& [value, timestamp] : results.values) {
    if (std::isfinite(value)) {
      continue;
    }
    auto total = m_nonFiniteValueHistogram[step]++;
    DS_NOTICE("NaN/Infinity value detected", timestamp, value, step, total);
    value = 0.0;
  }

  DS_REQUIRE(std::none_of(std::cbegin(results.values), std::cend(results.values),
                          [](const auto& data) { return !std::isfinite(data.value); }),
             "values contain non-finite values");

  return results;
}

} // namespace

namespace dss::metering {

std::shared_ptr<Archives> makeArchives(std::shared_ptr<RRDInterface> rrdInterface, ds::System& system) {
  return std::make_shared<ArchivesImpl>(rrdInterface, system);
};

} // namespace dss::metering
