/*
    Copyright (c) 2009,2010,2011 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "metering.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <digitalSTROM/dsuid.h>
#include <ds/asio/io-service.h>
#include <ds/asio/timer.h>
#include <ds/log.h>
#include <sys/types.h>
#include <time.h>

#include <algorithm>
#include <boost/filesystem.hpp>
#include <boost/range/algorithm/find.hpp>
#include <boost/range/algorithm/find_if.hpp>
#include <boost/range/algorithm/replace_if.hpp>
#include <cmath>
#include <functional>
#include <iterator>
#include <numeric>

#include "businterface.h"
#include "dss.h"
#include "metering/archives.h"
#include "metering/config.h"
#include "metering/power-states.h"
#include "metering/rrd-interface.h"
#include "metering/socket-server.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "propertysystem.h"
#include "security/security.h"

static ds::LogChannel dsLogChannel("dssMeteringMetering");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

using system_clock = std::chrono::system_clock;

namespace {

class Direction {
public:
  Direction(Archives* archives, Category category, ElectricityMeteringConfig& electricityMeteringConfig)
      : m_archives(archives), m_directionType(category), m_config(electricityMeteringConfig) {}

  std::vector<dsuid_t> getActiveMeters() const { return m_meters; }
  bool isActiveMeter(const dsuid_t& meterId) const {
    boost::mutex::scoped_lock lock(m_cachedSeries_mutex);
    return (boost::find(m_meters, meterId) != m_meters.end());
  }

  std::optional<ArchiveValue> getInstantValue(const dsuid_t& meterId) const;
  std::optional<ArchiveValue> getEnergy(const dsuid_t& meterId) const;
  std::optional<ArchiveValue> tryCreateCacheFromDB(const dsuid_t& meterId) const;

  bool tryGetValidArchiveFilename(const dsuid_t& meterId) const;
  void getOrCreateCachedSeries(const dsuid_t& meterId);

  void addEnergy(const dsuid_t& meterId, EnergyMeasurement&& energyMeasurement);

  ValueRange getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                       bool _energyInWh, const TimeDuration& timeDuration, int _valueCount);
  unsigned int getTotalPower() const;

  double getInstantPower(const dsuid_t& meterId) const {
    auto value = getInstantValue(meterId);
    return (value) ? value->power : 0;
  }
  double getLastEnergy(const dsuid_t& meterId) const {
    // keep latest energy value even when sensor value timeout
    auto value = getEnergy(meterId);
    return (value) ? value->energy : 0;
  }

  void pollTimestamp(ds::DateTime timepoint) { m_lastPoll = timepoint; }

  double get15MinuteAverage(std::vector<Dsuid> meterIds, metering::SeriesTypes type, Unit unit) {
    auto range = getSeries(meterIds, 15min, type, unit == Unit::Wh, {}, 1);
    auto& values = range.values;
    if (values.size() != 1) {
      DS_NOTICE("unexpected value count received", values.size(), range);
      return 0;
    }
    DS_REQUIRE(values.size() == 1);
    DS_REQUIRE(std::isfinite(values[0].value));
    return values[0].value;
  };

private:
  mutable boost::mutex m_cachedSeries_mutex;
  mutable std::vector<dsuid_t> m_meters;

  std::optional<unsigned long> deltaLast(const dsuid_t& meterId, EnergyMeasurement energyMeasurement);

  mutable std::vector<std::pair<dsuid_t, ArchiveValue>> m_energyCache;
  std::vector<std::pair<dsuid_t, EnergyMeasurement>> m_lastMeasurements;

  Archives* m_archives;
  Category m_directionType;
  const ElectricityMeteringConfig& m_config;
  ds::DateTime m_lastPoll;
};

std::optional<unsigned long> Direction::deltaLast(const dsuid_t& meterId, EnergyMeasurement newMeasurement) {
  auto it = boost::find_if(m_lastMeasurements, [&meterId](const auto& elt) { return elt.first == meterId; });
  if (it == m_lastMeasurements.end()) {
    // the first value sets only the baseline, e.g. startup
    DS_NOTICE("deltaEnergy: starting new baseline, assume system startup");
    m_lastMeasurements.push_back(std::make_pair(meterId, std::move(newMeasurement)));
    return {};
  }

  const auto [__, newEnergy, newTimestamp] = newMeasurement;
  auto [_, lastEnergy, lastTimestamp] = std::exchange(it->second, newMeasurement);

  // TODO check dsm reboot counter instead
  auto limit = [config = m_config.discontinuityDetection](Category category) -> unsigned {
    if (category == Category::consumption) {
      return config.maxConsumption;
    } else {
      return config.maxProduction;
    }
  };

  // sanity check, avoid big jumps: #10705
  if (newTimestamp <= lastTimestamp) {
    DS_NOTICE("deltaEnergy: timestamp not increasing, starting new baseline:", newTimestamp, lastTimestamp);
    return {};
  }
  if (newEnergy < lastEnergy) {
    DS_NOTICE("deltaEnergy: energy counter is going backwards:", newEnergy, lastEnergy);
    return {};
  }
  auto deltaTime = std::chrono::duration_cast<std::chrono::milliseconds>(newTimestamp - lastTimestamp);
  if (deltaTime > 5s) {
    DS_NOTICE("deltaEnergy: too long delay between updates, setting new baseline:", deltaTime);
    return {};
  }

  if (newEnergy - lastEnergy < std::ceil(deltaTime.count() / 1000.0) * limit(m_directionType)) {
    return (newEnergy - lastEnergy);
  }

  DS_NOTICE("deltaEnergy: sensor value marked as outlier:", newEnergy, lastEnergy, deltaTime, limit(m_directionType),
            "not adding to db");
  return {};
}

unsigned int Direction::getTotalPower() const {
  return std::accumulate(std::cbegin(m_energyCache), std::cend(m_energyCache), 0,
                         [sensorValueTimeout = m_config.sensorValueTimeout](unsigned int sum, const auto& pair) {
                           const auto& value = std::get<ArchiveValue>(pair);
                           return sum +
                                  (((system_clock::now() - value.lastChanged) > sensorValueTimeout) ? 0 : value.power);
                         });
}

std::optional<ArchiveValue> Direction::getInstantValue(const dsuid_t& meterId) const {
  const auto cache = boost::find_if(m_energyCache, [&meterId](auto& x) { return x.first == meterId; });
  if (cache == m_energyCache.end()) {
    return std::nullopt;
  }

  const auto& energy = cache->second;
  if ((system_clock::now() - energy.lastChanged) > m_config.sensorValueTimeout) {
    DS_WARNING("no update from", meterId, "for", ds::hideArg(m_config.sensorValueTimeout));
    return std::nullopt;
  }

  return energy;
}

bool Direction::tryGetValidArchiveFilename(const dsuid_t& meterId) const {
  if (isActiveMeter(meterId)) {
    return true;
  }

  auto archiveId = ArchiveId{meterId, m_directionType};
  if (m_archives->hasArchive(archiveId)) {
    boost::mutex::scoped_lock lock(m_cachedSeries_mutex);
    DS_NOTICE("add known archive", archiveId);
    m_meters.push_back(meterId);
    return true;
  }
  return false;
}

void Direction::getOrCreateCachedSeries(const dsuid_t& meterId) {
  if (tryGetValidArchiveFilename(meterId)) {
    return;
  }

  auto archiveId = ArchiveId{meterId, m_directionType};
  m_archives->create(archiveId);
  {
    boost::mutex::scoped_lock lock(m_cachedSeries_mutex);
    DS_NOTICE("add known archive", archiveId);
    m_meters.push_back(meterId);
  }
}

std::optional<ArchiveValue> Direction::getEnergy(const dsuid_t& meterId) const {
  DS_CONTEXT("getEnergy", ds::hideArg(m_directionType), ds::hideArg(meterId));
  if (auto cache = boost::find_if(m_energyCache, [&meterId](auto& x) { return x.first == meterId; });
      cache != m_energyCache.end()) {
    return cache->second;
  }
  if (auto fromDB = tryCreateCacheFromDB(meterId)) {
    DS_NOTICE(*fromDB);
    m_energyCache.push_back(std::make_pair(meterId, *fromDB));
    return fromDB;
  }
  return {};
}

void Direction::addEnergy(const dsuid_t& meterId, EnergyMeasurement&& energyMeasurement) {
  DS_CONTEXT("addEnergy", ds::hideArg(m_directionType), ds::hideArg(meterId));
  auto deltaEnergy = deltaLast(meterId, energyMeasurement);
  DS_DEBUG("sensor:", ds::hideArg(energyMeasurement), deltaEnergy);
  if (!deltaEnergy) {
    // the first value sets only the baseline or delta against previous is too high to be real
    return;
  }

  auto energyMetering = getEnergy(meterId);
  if (!energyMetering) {
    energyMetering.emplace(ArchiveValue{.power = gsl::narrow_cast<double>(energyMeasurement.power),
                                        .energy = gsl::narrow_cast<double>(*deltaEnergy),
                                        .lastChanged = energyMeasurement.timestamp});
    m_energyCache.push_back(std::make_pair(meterId, *energyMetering));
    DS_NOTICE("new entry", energyMetering);
  } else {
    energyMetering->energy += *deltaEnergy;
    energyMetering->power = energyMeasurement.power;
    energyMetering->lastChanged = energyMeasurement.timestamp;
    boost::replace_if(
        m_energyCache, [&meterId](auto& x) { return x.first == meterId; }, std::make_pair(meterId, *energyMetering));
  }

  getOrCreateCachedSeries(meterId);
  DS_ASSERT(boost::find(m_meters, meterId) != m_meters.end());

  // TODO why can't we leave it as double, why do we need ceil
  auto archiveValue = ArchiveValue{
      .power = gsl::narrow_cast<double>(energyMeasurement.power),
      .energy = std::ceil(energyMetering->energy),
      .lastChanged = energyMeasurement.timestamp,
  };

  try {
    m_archives->addValue({meterId, m_directionType}, std::move(archiveValue));
  } catch (const ds::Exception& e) {
    DS_WARNING(m_lastPoll, archiveValue.lastChanged, e.what());
  }
}

std::optional<ArchiveValue> Direction::tryCreateCacheFromDB(const dsuid_t& meterId) const {
  // after restart addConsumed energy will skip this and start from 0 [Ws] each time
  if (!tryGetValidArchiveFilename(meterId)) {
    DS_NOTICE("tryCreateCacheFromDB: no valid archive file for", meterId);
    return std::nullopt;
  }

  return m_archives->lastValue({meterId, m_directionType});
}

ValueRange Direction::getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, SeriesTypes _type,
                                bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) {
  std::vector<ArchiveId> archiveIds;
  for (auto it = std::begin(meterIds); it != std::end(meterIds);) {
    if (tryGetValidArchiveFilename(*it)) {
      archiveIds.push_back({*it, m_directionType});
      it++;
    } else {
      DS_DEBUG("no data file", *it);
      it = meterIds.erase(it);
    }
  }

  DS_DEBUG(meterIds);
  if (archiveIds.empty()) {
    return {};
  }

  auto& archiveConfigs = m_archives->getConfig();
  auto it = std::find_if(std::rbegin(archiveConfigs), std::rend(archiveConfigs),
                         [resolution](const auto& element) { return element.resolution == resolution; });
  DS_REQUIRE(it != std::rend(archiveConfigs), "unsupported resolution");
  int numberOfValues = it->numberOfValues;

  // adjust limits by start/end/values
  _valueCount = std::min(_valueCount, numberOfValues);

  long unsigned int step = resolution.count();
  auto align = [step](time_t value) { return value - value % step; };

  // supported range by rra at given resolution, can't go further into the past
  auto now = system_clock::now();
  time_t maxEnd = align(system_clock::to_time_t(now));
  time_t maxStart = align(system_clock::to_time_t(now) - (step * numberOfValues));
  time_t start{}, end{};

  if (timeDuration == TimeDuration{}) {
    end = maxEnd;
    if (_valueCount == 0) {
      start = maxStart;
    } else {
      start = end - (step * _valueCount);
    }
  } else if (!timeDuration.start) {
    end = align(system_clock::to_time_t(*timeDuration.end));
    if (_valueCount == 0) {
      start = maxStart;
    } else {
      start = end - (step * _valueCount);
    }
  } else if (!timeDuration.end) {
    start = align(system_clock::to_time_t(*timeDuration.start));
    if (_valueCount == 0) {
      end = maxEnd;
    } else {
      end = start + (step * _valueCount);
    }
  } else {
    start = align(system_clock::to_time_t(*timeDuration.start));
    end = align(system_clock::to_time_t(*timeDuration.end));
  }
  end += step;

  m_archives->flush(archiveIds);
  auto series = m_archives->xport(start, end, step, archiveIds, _type, _energyInWh);
  auto returnVector = std::move(series.values);

  bool lastValueEmpty = false;
  if (!returnVector.empty() && returnVector.back().value == 0) { // -> unknown
    DS_DEBUG("last value is unknown:", returnVector.back());
    lastValueEmpty = true;
  }

  if (_type == SeriesTypes::energy) {
    double currentCounter = 0.0;
    for (auto meterId : meterIds) {
      if (auto energyMetering = getEnergy(meterId)) {
        currentCounter += energyMetering->energy;
      }
    }
    if (_energyInWh) {
      currentCounter /= 3600;
    }
    auto lastZeroValue = Value{0, DateTime((time_t)0)};
    for (auto& data : returnVector) {
      if (data.value == 0) {
        lastZeroValue = data;
      } else {
        break;
      }
    }
    for (auto iter = returnVector.rbegin(); iter < returnVector.rend(); ++iter) {
      if (iter->timeStamp == lastZeroValue.timeStamp) {
        break;
      }
      double val = std::exchange(iter->value, currentCounter);
      currentCounter -= val;
    }
  }

  if (lastValueEmpty) {
    // delete the last value, if it is Unknown (zero)
    // assumption: avoid return to zero of graph if current value did not arrive yet
    returnVector.pop_back();
  }

  return {.start = series.start, .end = series.end, .step = series.step, .values = returnVector};
}

class MeteringImpl : public MeteringSubsystem {
  ds::System& m_system;
  ElectricityMeteringConfig m_electricityMeteringConfig;
  std::string m_MeteringStorageLocation{"/invalid-store/"};
  MeteringBusInterface* m_pMeteringBusInterface = NULL;

  std::shared_ptr<Archives> m_archives;
  std::shared_ptr<PowerStates> m_powerStates;
  Direction m_consumption;
  Direction m_production;
  ds::Clock::time_point m_nextPollAt;

public:
  // callbacks from Direction
  std::string createArchiveFilename(const ArchiveId& archiveId) const { return m_archives->filename(archiveId); }

  void checkDBConsistency() override;
  std::vector<ArchiveId> searchArchivesInFolder() const override;
  bool recreateDeadArchives(const std::vector<ArchiveId>& archiveIds) const override;
  bool validateAndConvertArchive(const ArchiveId& archiveId) override;

  metering::Archives& getArchives() override { return *m_archives; }

  void doStart() override;
  void execute() override;
  void initialize() override;
  void pollLoop(boost::asio::io_service& io, ds::asio::Timer& timer);

public:
  MeteringImpl(DSS* _pDSS, std::shared_ptr<Archives>&& archives, std::shared_ptr<PowerStates> powerStates,
               ds::System& system);
  ~MeteringImpl() override = default;

  const std::vector<MeteringConfig> getConfig() const override { return m_archives->getConfig(); }
  void setMeteringBusInterface(MeteringBusInterface* _value) override { m_pMeteringBusInterface = _value; }

  std::optional<unsigned long> getConsumedPower(const dsuid_t& meterId) const override {
    return gsl::narrow<unsigned int>(m_consumption.getInstantPower(meterId));
  }
  unsigned int getConsumedPowerApartment() const override { return m_consumption.getTotalPower(); }

  std::optional<unsigned long> getProducedPower(const dsuid_t& meterId) const override {
    return gsl::narrow<unsigned int>(m_production.getInstantPower(meterId));
  }
  unsigned int getProducedPowerApartment() const override { return m_production.getTotalPower(); }

  std::optional<ArchiveValue> getEnergy(const dsuid_t& meterId) const override {
    return m_consumption.getEnergy(meterId);
  }
  std::optional<ArchiveValue> getProducedEnergy(const dsuid_t& meterId) const override {
    return m_production.getEnergy(meterId);
  }

  double get15minConsumed(std::vector<Dsuid> meterIds, metering::SeriesTypes type) override;
  double get15minProduced(std::vector<Dsuid> meterIds, metering::SeriesTypes type) override;

  TotalEnergy getLastEnergy(const dsuid_t& meterId) const override {
    return {m_production.getLastEnergy(meterId), m_consumption.getLastEnergy(meterId)};
  }

  std::shared_ptr<State> createState(MeteringSystem meteringSystem, PowerStateConfig&& powerStateConfig) override {
    return m_powerStates->createState(meteringSystem, std::move(powerStateConfig));
  }
  void removeState(MeteringSystem meteringSystem, const PowerStateConfig& powerStateConfig) override {
    return m_powerStates->removeState(meteringSystem, powerStateConfig);
  }
  void triggerState(MeteringSystem meteringSystem, TotalPower&& totalPower) override {
    m_powerStates->trigger(meteringSystem, std::move(totalPower));
  }

  void addConsumedEnergy(DSMeter& meter, EnergyMeasurement&& energyEvent) override;
  void addProducedEnergy(DSMeter& meter, EnergyMeasurement&& energyMeasurement) override {
    m_production.addEnergy(meter.getDSID(), std::move(energyMeasurement));
  }

  ValueRange getSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution, metering::SeriesTypes _type,
                       bool _energyInWh, const TimeDuration& timeDuration, int _valueCount) override {
    return m_consumption.getSeries(meterIds, resolution, _type, _energyInWh, timeDuration, _valueCount);
  }

  ValueRange getProducedSeries(std::vector<Dsuid> meterIds, std::chrono::seconds resolution,
                               metering::SeriesTypes _type, bool _energyInWh, const TimeDuration& timeDuration,
                               int _valueCount) override {
    return m_production.getSeries(meterIds, resolution, _type, _energyInWh, timeDuration, _valueCount);
  }

}; // MeteringImpl

MeteringImpl::MeteringImpl(DSS* _pDSS, std::shared_ptr<Archives>&& archives, std::shared_ptr<PowerStates> powerStates,
                           ds::System& system)
    : MeteringSubsystem(_pDSS),
      m_system(system),
      m_archives(std::move(archives)),
      m_powerStates(std::move(powerStates)),
      m_consumption(m_archives.get(), Category::consumption, m_electricityMeteringConfig),
      m_production(m_archives.get(), Category::production, m_electricityMeteringConfig) {}

void MeteringImpl::addConsumedEnergy(DSMeter& meter, EnergyMeasurement&& energyMeasurement) {
  auto meteringCache = MeteringCache{
      .power = gsl::narrow<int>(energyMeasurement.power), // property tree exported
      .lastSensorValue = energyMeasurement.energy,
      .lastChanged = dss::DateTime{energyMeasurement.timestamp},
  };

  m_consumption.addEnergy(meter.getDSID(), std::move(energyMeasurement));

  if (auto energyMetering = m_consumption.getEnergy(meter.getDSID())) {
    meteringCache.energyWs = energyMetering->energy;
    meteringCache.energyWh = energyMetering->energy / 3600;

    // backend data for property tree, etc
    meter.updateMeteringCache(std::move(meteringCache));
  }
}

double MeteringImpl::get15minConsumed(std::vector<Dsuid> meterIds, metering::SeriesTypes type) {
  return m_consumption.get15MinuteAverage(meterIds, type, Unit::Wh);
}

double MeteringImpl::get15minProduced(std::vector<Dsuid> meterIds, metering::SeriesTypes type) {
  return m_production.get15MinuteAverage(meterIds, type, Unit::Wh);
}

// subsystem interface
void MeteringImpl::initialize() {
  Subsystem::initialize();
  auto config = getDSS().getPropertySystem().createProperty(getConfigPropertyBasePath());
  m_MeteringStorageLocation =
      config->getOrCreateChildValue<std::string>("storageLocation", getDSS().getDataDirectory() + "metering/");
  m_MeteringStorageLocation = addTrailingBackslash(m_MeteringStorageLocation);
  DS_NOTICE(m_MeteringStorageLocation);
  m_archives->setRRDCachedPath(
      config->getOrCreateChildValue<std::string>("rrdDaemonAddress", "unix:/var/run/rrdcached.sock"));

  if (isEnabled()) {
    if (!boost::filesystem::is_directory(m_MeteringStorageLocation)) {
      throw std::runtime_error("Metering directory " +
                               boost::filesystem::system_complete(m_MeteringStorageLocation).string() +
                               " does not exist!");
    }

    if (!rwAccess(m_MeteringStorageLocation)) {
      throw std::runtime_error("Metering directory " +
                               boost::filesystem::system_complete(m_MeteringStorageLocation).string() +
                               " is not readable/writable!");
    }
  }

  m_electricityMeteringConfig = m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {});
  DS_NOTICE("config", ds::hideArg(m_electricityMeteringConfig));
  m_archives->setMeteringStorageLocation(std::string(m_MeteringStorageLocation));
}

bool MeteringImpl::recreateDeadArchives(const std::vector<ArchiveId>& archiveIds) const {
  bool restartCached{false};
  for (const auto& archiveId : archiveIds) {
    if (m_archives->withoutUpdates(archiveId)) {
      DS_WARNING("Metering DB ", m_archives->filename(archiveId), " is out of date. removing ");
      m_archives->create(archiveId); // bogus, file is not removed, TODO add unit test to verify
      restartCached = true;
    }
  }
  return restartCached;
}

std::vector<ArchiveId> MeteringImpl::searchArchivesInFolder() const {
  constexpr auto len = "1111111111111111111111111111111100"sv.length();

  std::vector<ArchiveId> archiveIds{};

  for (auto file : m_system.lsRegularFiles(m_MeteringStorageLocation, ".rrd"s)) {
    auto stem = std::string(boost::filesystem::path(file).stem().c_str());
    DS_DEBUG(stem);

    dsuid_t dsuid;
    try {
      if (stem.size() <= 24) {
        dsid_t dsid = str2dsid(stem);
        dsuid = dsuid_from_dsid(&dsid);

        auto oldFile = std::exchange(
            file, ds::str(m_MeteringStorageLocation, createArchiveFilename({dsuid, Category::consumption})));

        if (m_system.existsFile(file)) {
          DS_WARNING("Failed to rename old dsid based archive, new file already exists");
          // m_system.remove(file);
        } else {
          m_system.rename(oldFile, file);
          DS_NOTICE("Migrated metering data from ", oldFile, " to ", file);
        }
      } else if (stem.length() >= len) {
        dsuid = str2dsuid(stem.substr(0, len));
      } else {
        DS_NOTICE("invalid archive file", file);
        continue;
      }

    } catch (std::exception& e) {
      DS_NOTICE("failed to migrate", file, e.what());
      continue;
    }

    Category category =
        (stem.rfind("-production"sv) != std::string::npos) ? Category::production : Category::consumption;
    DS_NOTICE(dsuid, category);
    archiveIds.push_back(ArchiveId{dsuid, category});
  }
  return archiveIds;
}

void MeteringImpl::checkDBConsistency() {
  DateTime now;

  auto archiveIds = std::vector<ArchiveId>();

  auto fillMeterCategory = [&archiveIds](const auto& meters, Category category) {
    std::transform(meters.begin(), meters.end(), std::back_inserter(archiveIds), [category](const auto& dsuid) {
      return ArchiveId{dsuid, category};
    });
  };

  fillMeterCategory(m_consumption.getActiveMeters(), Category::consumption);
  fillMeterCategory(m_production.getActiveMeters(), Category::production);

  auto restartCached = recreateDeadArchives(archiveIds);
  if (restartCached) {
    // no need to flush
    m_archives->restartCached();
    return;
  }

  auto needsFlush = std::any_of(std::cbegin(archiveIds), std::cend(archiveIds),
                                [&, now](const auto& archiveId) { return m_archives->checkDBFlush(now, archiveId); });

  if (needsFlush) {
    m_archives->flush(archiveIds);
  }
}

void MeteringImpl::pollLoop(boost::asio::io_service& io, ds::asio::Timer& timer) {
  if (m_Terminated) {
    DS_NOTICE("pollLoop: terminating");
    return;
  }

  auto startedAt = ds::Clock::now();
  if (m_nextPollAt != std::chrono::floor<std::chrono::seconds>(startedAt)) {
    DS_NOTICE("pollLoop: missed wakeup", startedAt, m_nextPollAt, ds::hideArg(startedAt - m_nextPollAt));
  }

  DS_ASSERT(m_pMeteringBusInterface != NULL, "");
  try {
    auto now = system_clock::now();
    m_consumption.pollTimestamp(now);
    m_production.pollTimestamp(now);
    m_pMeteringBusInterface->requestMeterData();
  } catch (dss::BusApiError& apiError) {
    log("Metering::execute: Couldn't get metering data: " + std::string(apiError.what()));
  }

  triggerState(MeteringSystem::apartmentSummation,
               {.production = m_production.getTotalPower(), .consumption = m_consumption.getTotalPower()});
  checkDBConsistency();

  auto now = ds::Clock::now();
  if (now - startedAt > 1s) {
    DS_WARNING("pollLoop: exceeded 1s budget", now - startedAt);
  }

  m_nextPollAt = std::chrono::ceil<std::chrono::seconds>(now);
  timer.expiresFromNow(m_nextPollAt - now, [this, &io, &timer] { pollLoop(io, timer); });
}

bool MeteringImpl::validateAndConvertArchive(const ArchiveId& archiveId) {
  DS_ASSERT(m_archives->hasArchive(archiveId));
  try {
    bool tunePowerMaxSetting = false;
    bool validate = m_archives->validateDBSeries(archiveId, tunePowerMaxSetting);
    if (!validate) {
      DS_NOTICE("remove invalid archive", archiveId)
      m_system.remove(createArchiveFilename(archiveId));
      return false;
    } else if (tunePowerMaxSetting) {
      DS_NOTICE("archive needs tuning", archiveId);
      m_archives->tuneDBPowerSettings(archiveId); // non-const
    }
    return true;
  } catch (const ds::Exception& e) {
    DS_WARNING(e.what());
    return false;
  }
  return true;
}

void MeteringImpl::execute() {
  while (DSS::getInstance()->getModelMaintenance().isInitializing()) {
    sleepSeconds(1);
  }
  DS_NOTICE("pollLoop: modelReady received");

  auto archiveIds = searchArchivesInFolder();
  DS_NOTICE("found", ds::hideArg(archiveIds.size()), "archives");

  auto changes = recreateDeadArchives(archiveIds); // TODO remove file not re-create
  if (changes) {
    // file was removed/recreated
    archiveIds = searchArchivesInFolder();
  }

  for (auto it = archiveIds.cbegin(); it != archiveIds.end(); it++) {
    const auto& archiveId = *it;
    DS_NOTICE("checking:", archiveId);
    if (!validateAndConvertArchive(*it)) {
      DS_NOTICE("skip archive", *it);
      archiveIds.erase(it);
    }
  }

  ds::asio::IoService ioService; // `ds::asio::IoService` requires its construct to run in the same thread
  boost::asio::io_service::work work(ioService); // block in `run()` until `m_stopping`

  ds::asio::Timer timer(ioService);
  pollLoop(ioService, timer);
  auto socketServer = metering::makeSocketServer(ioService, m_electricityMeteringConfig.socketPath, *this);
  socketServer->start();

  while (!m_Terminated) {
    try {
      ioService.run();
      break; // run() exited normally
    } catch (std::exception& e) {
      DS_ERROR("io_service", e.what());
    }
  }
}

void MeteringImpl::doStart() {
  log("Writing files to: " + m_MeteringStorageLocation);
  run();
} // start

} // namespace

namespace dss {

std::shared_ptr<MeteringSubsystem> makeMetering(DSS* dss, Apartment& apartment) {
  return std::make_shared<MeteringImpl>(dss, makeArchives(makeRRDInterface(), ds::system()),
                                          makePowerStates(apartment), ds::system());
}

std::shared_ptr<MeteringSubsystem> makeMockedMetering(Apartment& apartment,
                                                        std::shared_ptr<metering::Archives> archives,
                                                        ds::System& system) {
  return std::make_shared<MeteringImpl>(nullptr, std::move(archives), makePowerStates(apartment), system);
}

} // namespace dss
