/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include "model-features.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include <algorithm>
#include <boost/range/algorithm/find_if.hpp>

static ds::LogChannel dsLogChannel("modelFeaturesTest");
static const char* TAGS = "[device][api]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

using namespace std::literals;

namespace {

const std::vector<ModelFeatureId> MF_GE_KM220 = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,        ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,     ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const std::vector<ModelFeatureId> MF_GE_KM2 = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,        ModelFeatureId::outmode,
    ModelFeatureId::outvalue8,     ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const std::vector<ModelFeatureId> MF_GE_KL200 = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,         ModelFeatureId::identification,
    ModelFeatureId::ledauto,       ModelFeatureId::transt,        ModelFeatureId::outmodeswitch,
    ModelFeatureId::outvalue8,     ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,
    ModelFeatureId::pushbarea,     ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch,
    ModelFeatureId::impulseconfig, ModelFeatureId::blinkconfig};

const std::vector<ModelFeatureId> MF_GE_TKM210 = {
    ModelFeatureId::dontcare,      ModelFeatureId::blink,           ModelFeatureId::identification,
    ModelFeatureId::transt,        ModelFeatureId::outmode,         ModelFeatureId::outvalue8,
    ModelFeatureId::pushbutton,    ModelFeatureId::pushbdevice,     ModelFeatureId::pushbarea,
    ModelFeatureId::pushbadvanced, ModelFeatureId::outconfigswitch, ModelFeatureId::impulseconfig,
    ModelFeatureId::blinkconfig};

const std::vector<ModelFeatureId> MF_RT_KM2 = {
    ModelFeatureId::dontcare, ModelFeatureId::blink,   ModelFeatureId::identification, ModelFeatureId::ledauto,
    ModelFeatureId::transt,   ModelFeatureId::outmode, ModelFeatureId::outvalue8};

const std::vector<ModelFeatureId> MF_RT_TKM2 = {ModelFeatureId::dontcare,       ModelFeatureId::blink,
                                                ModelFeatureId::identification, ModelFeatureId::transt,
                                                ModelFeatureId::outmode,        ModelFeatureId::outvalue8};

SCENARIO("model features", TAGS) {
  GIVEN("model") {
    const auto instance = makeModelFeatures();

    SECTION("Required Models") {
      struct X {
        int colorId;
        const std::vector<std::string> requiredModels;
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{
          {ColorIDYellow,
           {
               "KL:200"s,  "KM:2"s,    "KM:220"s,  "KM:300"s,  "SDM:20"s,  "SDM:300"s, "SDM:301"s,
               "SDS:2"s,   "SDS:20"s,  "SDS:210"s, "SDS:22"s,  "TKM:2"s,   "TKM:210"s, "TKM:220"s,
               "TKM:230"s, "TKM:300"s, "UMV:200"s, "UMV:204"s, "UMV:210"s, "ZWD:300"s, "ZWS:2"s,
           }},
          {ColorIDGray,
           {
               "KL:2"s,
               "KL:210"s,
               "KL:220"s,
               "KL:230"s,
               "KL:300"s,
               "KL:330"s,
               "TKM:2"s,
           }},
          {ColorIDBlue,
           {
               "KL:2"s,
               "KM:200"s,
               "KM:300"s,
               "SDS:2"s,
               "SK:204"s,
               "UMV:200"s,
           }},
          {ColorIDCyan, {"TKM:2"s}},
          {ColorIDViolet, {}},
          {ColorIDRed,
           {
               "KM:2"s,
               "SDM:2"s,
               "TKM:2"s,
           }},
          {ColorIDGreen,
           {
               "KM:2"s,
               "TKM:2"s,
           }},
          {ColorIDBlack,
           {
               "AKM:2"s,
               "KL:2"s,
               "KL:213"s,
               "KL:214"s,
               "SDS:20"s,
               "SDS:22"s,
               "SK:20"s,
               "SKM:300"s,
               "TKM:2"s,
               "TNY:2"s,
               "UMR:200"s,
               "ZWS:2"s,
               "ZWS:205"s,
           }},
          {ColorIDWhite, {}},

      }));

      THEN("supported models exist") {
        INFO(ds::str(x.colorId));
        auto modelsFeatures = instance->getFeatures(x.colorId);
        auto supportedModels = std::vector<std::string>{};
        std::transform(std::begin(modelsFeatures), std::end(modelsFeatures), std::back_inserter(supportedModels),
                       [](const auto& pair) -> std::string { return pair.first; });
        std::sort(std::begin(supportedModels), std::end(supportedModels));
        CHECK(supportedModels == x.requiredModels);
      }
    }

    SECTION("Smoke test some model feature values") {
      struct X {
        int colorId;
        std::string model;
        const std::vector<ModelFeatureId>& modelFeatures;
      };

      const auto& x = GENERATE(values(std::initializer_list<X>{
          {ColorIDYellow, "KM:220", MF_GE_KM220},
          {ColorIDYellow, "KM:2", MF_GE_KM2},
          {ColorIDYellow, "KL:200", MF_GE_KL200},
          {ColorIDYellow, "TKM:210", MF_GE_TKM210},
          {ColorIDRed, "KM:2", MF_RT_KM2},
          {ColorIDRed, "TKM:2", MF_RT_TKM2},
      }));

      THEN("features exist and match") {
        INFO(ds::str(x.model));
        auto modelsFeatures = instance->getFeatures(x.colorId);
        const auto features = std::find_if(std::begin(modelsFeatures), std::end(modelsFeatures),
                                           [key = x.model](const auto& pair) { return pair.first == key; });
        CHECK(features != modelsFeatures.end());
        CHECK(features->second == ModelFeatures::features_t{x.modelFeatures});
        DS_NOTICE(*features);
      }
    }
    SECTION("dynamic extend feature matrix") {
      WHEN("adding features for OEM device") {
        const auto featuresPtr = std::vector{ModelFeatureId::pushbutton, ModelFeatureId::outconfigswitch};
        instance->setFeatures(ColorIDViolet, "sampleDevice"s, std::vector(featuresPtr));
        THEN("adding features for OEM device") {
          auto modelsFeatures = instance->getFeatures(ColorIDViolet);
          auto modelFeature =
              *boost::find_if(modelsFeatures, [](const auto& pair) { return pair.first == "sampleDevice"s; });
          CHECK(modelFeature.second == ModelFeatures::features_t{featuresPtr});
        }
      }
      WHEN("adding features for the same model 2nd time") {
        auto modelsFeatures = instance->getFeatures(ColorIDYellow);
        instance->setFeatures(ColorIDYellow, "KL:200", std::vector(MF_GE_TKM210)); // try to assign wrong features
        THEN("builtin features are not modified") {
          auto modelsFeatures = instance->getFeatures(ColorIDYellow);
          auto modelFeature = *boost::find_if(modelsFeatures, [](const auto& pair) { return pair.first == "KL:200"s; });
          CHECK(modelFeature.second == ModelFeatures::features_t{MF_GE_KL200});
        }
      }
      WHEN("adding features for standard device") {
        instance->setFeatures(ColorIDYellow, "KM:220"s, {ModelFeatureId::pushbutton, ModelFeatureId::outconfigswitch});
        THEN("builtin features are not modified") {
          auto modelsFeatures = instance->getFeatures(ColorIDYellow);
          auto modelFeature = *boost::find_if(modelsFeatures, [](const auto& pair) { return pair.first == "KM:220"s; });
          CHECK(modelFeature.second == ModelFeatures::features_t{MF_GE_KM220});
        }
      }
    }
    SECTION("check color strings") {
      WHEN("valid colorId") {
        struct X {
          int colorId;
          std::string colorString;
        };
        const auto& x = GENERATE(values(std::initializer_list<X>{{ColorIDYellow, "GE"},
                                                                 {ColorIDGray, "GR"},
                                                                 {ColorIDBlue, "BL"},
                                                                 {ColorIDCyan, "TK"},
                                                                 {ColorIDViolet, "MG"},
                                                                 {ColorIDRed, "RT"},
                                                                 {ColorIDGreen, "GN"},
                                                                 {ColorIDBlack, "SW"},
                                                                 {ColorIDWhite, "WE"}}));
        CHECK(instance->colorToString(x.colorId) == x.colorString);
      }

      WHEN("invalid colorId") { CHECK_THROWS(instance->colorToString(99)); }
    }

    SECTION("available model features") {
      static const std::vector<ModelFeatureId> availableFeatures = {ModelFeatureId::dontcare,
                                                                    ModelFeatureId::blink,
                                                                    ModelFeatureId::identification,
                                                                    ModelFeatureId::ledauto,
                                                                    ModelFeatureId::leddark,
                                                                    ModelFeatureId::transt,
                                                                    ModelFeatureId::outmode,
                                                                    ModelFeatureId::outmodeswitch,
                                                                    ModelFeatureId::outvalue8,
                                                                    ModelFeatureId::pushbutton,
                                                                    ModelFeatureId::pushbdevice,
                                                                    ModelFeatureId::pushbarea,
                                                                    ModelFeatureId::pushbadvanced,
                                                                    ModelFeatureId::pushbcombined,
                                                                    ModelFeatureId::shadeprops,
                                                                    ModelFeatureId::shadeposition,
                                                                    ModelFeatureId::motiontimefins,
                                                                    ModelFeatureId::optypeconfig,
                                                                    ModelFeatureId::shadebladeang,
                                                                    ModelFeatureId::highlevel,
                                                                    ModelFeatureId::consumption,
                                                                    ModelFeatureId::jokerconfig,
                                                                    ModelFeatureId::akmsensor,
                                                                    ModelFeatureId::akminput,
                                                                    ModelFeatureId::akmdelay,
                                                                    ModelFeatureId::twowayconfig,
                                                                    ModelFeatureId::outputchannels,
                                                                    ModelFeatureId::heatinggroup,
                                                                    ModelFeatureId::heatingoutmode,
                                                                    ModelFeatureId::heatingprops,
                                                                    ModelFeatureId::pwmvalue,
                                                                    ModelFeatureId::valvetype,
                                                                    ModelFeatureId::extradimmer,
                                                                    ModelFeatureId::umvrelay,
                                                                    ModelFeatureId::blinkconfig,
                                                                    ModelFeatureId::impulseconfig,
                                                                    ModelFeatureId::umroutmode,
                                                                    ModelFeatureId::pushbsensor,
                                                                    ModelFeatureId::locationconfig,
                                                                    ModelFeatureId::windprotectionconfigawning,
                                                                    ModelFeatureId::windprotectionconfigblind,
                                                                    ModelFeatureId::outmodegeneric,
                                                                    ModelFeatureId::outconfigswitch,
                                                                    ModelFeatureId::temperatureoffset,
                                                                    ModelFeatureId::apartmentapplication,
                                                                    ModelFeatureId::ftwtempcontrolventilationselect,
                                                                    ModelFeatureId::ftwdisplaysettings,
                                                                    ModelFeatureId::ftwbacklighttimeout,
                                                                    ModelFeatureId::ventconfig,
                                                                    ModelFeatureId::fcu,
                                                                    ModelFeatureId::pushbdisabled,
                                                                    ModelFeatureId::consumptioneventled,
                                                                    ModelFeatureId::consumptiontimer,
                                                                    ModelFeatureId::jokertempcontrol,
                                                                    ModelFeatureId::dimtimeconfig,
                                                                    ModelFeatureId::outmodeauto,
                                                                    ModelFeatureId::dimmodeconfig,
                                                                    ModelFeatureId::setumr200config,
                                                                    ModelFeatureId::extendedvalvetypes,
                                                                    ModelFeatureId::customtransitiontime,
                                                                    ModelFeatureId::outmodetempcontrol,
                                                                    ModelFeatureId::outmodeenoceanvalve,
                                                                    ModelFeatureId::operationlock,
                                                                    ModelFeatureId::grkl387workaround,
                                                                    ModelFeatureId::customactivityconfig};
      CHECK(instance->getAvailableFeatures() == availableFeatures);
    }
  }
}

} // namespace
