/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include "model/buttonmanipulator.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "model/apartment-mock.h"
#include "model/device-mock.h"
#include "structuremanipulator.h"

static ds::LogChannel dsLogChannel("dssModelButtonManipulatorTest");
static const char* TAGS = "[model][device][manipulator]";

using namespace dss;
using namespace dss::model;
using namespace trompeloeil;
using namespace std::literals;
using trompeloeil::_;

namespace {

SCENARIO("setButtonInputMode", TAGS) {
  auto busInterface = BusInterfaceMock{};
  ApartmentMock apartment(nullptr, nullptr, &busInterface);

  auto structureManipulator = StructureManipulator(apartment);

  auto buttonManipulator = makeButtonManipulator(apartment, structureManipulator);

  GIVEN("SW-TKM210") {
    auto& dev = *createDevice(apartment, {.hardwareType = HardwareType::SW_TKM200});
    dev.setActiveGroup(ApplicationType::lights);
    THEN("has no partner device") { CHECK_THROWS(buttonManipulator->setButtonInputMode(dev, "2way_up")); }
  }

  GIVEN("GR-KL300") {
    auto& dev = *createDevice(apartment, DeviceMockConf{.hardwareType = HardwareType::GR_KL300,
                                                        .buttonInputMode = ButtonInputMode::STANDARD});

    REQUIRE(!dev.is2WayMaster());
    REQUIRE(!dev.getPartnerDevice()->is2WaySlave());

    WHEN("setting 2way_up") {
      REQUIRE_CALL(busInterface.device,
                   setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2)));
      REQUIRE_CALL(busInterface.device,
                   setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1)));

      buttonManipulator->setButtonInputMode(dev, "2way_up");
      THEN("buttons are paired") {
        REQUIRE(dev.is2WayMaster());
        REQUIRE(dev.getPartnerDevice()->is2WaySlave());
      }
    }
    WHEN("setting 2way_down") {
      REQUIRE_CALL(busInterface.device,
                   setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_DW_WITH_INPUT2)));
      REQUIRE_CALL(busInterface.device,
                   setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_UP_WITH_INPUT1)));

      buttonManipulator->setButtonInputMode(dev, "2way_down");
      THEN("button are paired") {
        REQUIRE(dev.is2WayMaster());
        REQUIRE(dev.getPartnerDevice()->is2WaySlave());
      }

      WHEN("setting 1way") {
        REQUIRE_CALL(busInterface.device, setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::STANDARD)));
        REQUIRE_CALL(busInterface.device,
                     setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::AKM_STANDARD)));

        REQUIRE_CALL(busInterface.structureModify, removeFromGroup(_, _, _));
        REQUIRE_CALL(busInterface.structureModify, addToGroup(_, _, _));
        REQUIRE_CALL(busInterface.device, setDeviceConfig(_, CfgClassDevice, 0x40, _));
        REQUIRE_CALL(busInterface.device, setDeviceButtonActiveGroup(_, _));
        REQUIRE_CALL(busInterface.device, setDeviceConfig(_, CfgClassFunction, CfgFunction_ButtonMode, _));

        buttonManipulator->setButtonInputMode(dev, "1way");

        THEN("button are unpaired") {
          auto partnerDevice = dev.getPartnerDevice();
          REQUIRE(!dev.is2WayMaster());
          REQUIRE(!partnerDevice->is2WaySlave());
          REQUIRE(partnerDevice->getGroupsCount() == 1);
          REQUIRE(partnerDevice->isInGroup(GroupIDBlack)); // dropped from Gray
        }
      }
    }
  }

  SECTION("AKM") {
    GIVEN("AKM devices") {
      struct X {
        HardwareType hardwareType;
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{{HardwareType::SW_AKM200},
                                                               {HardwareType::SW_AKM210},
                                                               {HardwareType::GR_KL300},
                                                               {HardwareType::SW_SKM300},
                                                               {HardwareType::GE_TKM300},
                                                               {HardwareType::SW_UMR200}}));

      INFO(ds::str(x.hardwareType));
      auto& device = *createDevice(apartment, {.hardwareType = x.hardwareType});
      if (device.getDeviceType() == DEVICE_TYPE_SKM && device.getDeviceNumber() == 300) {
        device.setButtonInputMode(ButtonInputMode::AKM_ON_FALLING_EDGE);
      }

      THEN("isAKMDevice succeeds") { CHECK(isAkmDevice(device)); }
    }

    SECTION("setAKMInputProperty") {
      GIVEN("AKM device") {
        auto& device = *createDevice(apartment, {.hardwareType = HardwareType::SW_UMR200});

        THEN("valid mode is accepted") {
          REQUIRE_CALL(busInterface.device,
                       setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::AKM_ON_FALLING_EDGE)));
          CHECK(buttonManipulator->setAKMInputProperty(device, AutomationInputMode::on_falling_edge) ==
                ErrorCode{true, ""s});
        }

        THEN("invalid akm modes are rejected") {
          auto expected = buttonManipulator->setAKMInputProperty(device, AutomationInputMode{});
          REQUIRE(!expected.success);
          REQUIRE(expected.message == ds::str("Unsupported mode: 'none'"));
        }

        GIVEN("paired AKM device") {
          ALLOW_CALL(busInterface.device,
                     setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2)));
          ALLOW_CALL(busInterface.device,
                     setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1)));
          auto& partnerDevice = *device.getPartnerDevice();

          device.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2);
          partnerDevice.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1);

          THEN("setAKMInputProperty throws exception") {
            CHECK_THROWS(buttonManipulator->setAKMInputProperty(device, AutomationInputMode::standard));
          }
        }
      }
    }

    GIVEN("non AKM device") {
      auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL200});

      THEN("isAKMDevice fails") { CHECK(!isAkmDevice(device)); }
      THEN("setAKMInputProperty fails") {
        auto errorCode = buttonManipulator->setAKMInputProperty(device, AutomationInputMode::standard);
        CHECK(errorCode.success == false);
      }
    }
  }
}

}; // namespace
