#include "device-property-tree.h"

#include <ds/log.h>

#include "dss.h"
#include "hal/hardware-abstraction.h"
#include "model/apartment.h"
#include "model/buttonmanipulator.h"
#include "model/device.h"
#include "model/dsmeter.h"

static ds::LogChannel dsLogChannel("modelDevicePropertyTree");

using namespace dss::model;

namespace dss {

struct DevicePropertyTreeImpl final : public DevicePropertyTree {
  DevicePropertyTreeImpl(Apartment& apartment) : m_apartment(apartment) {}

  void publishToPropertyTree(Device& device) override;
  void publishVdcToPropertyTree(Device& device) override;
  void publishValveTypeToPropertyTree(Device& device) override;
  void publishAliasInZoneGroups(const Device& device) const override;
  void publishDeviceSpecification(Device& device) const override;
  void republishModelFeaturesToPropertyTree(Device& device) override;
  void removeFromPropertyTree(Device& device) override;
  void updateAKMNode(Device& device) override;

  void publishGroupNode(Device& device, int groupId) override;
  void removeGroupNode(Device& device, int groupId) override;

  void publishShadeAttributes(PropertyNode& deviceRoot, model::ShadeAttributes& shadeAttributes) override {
    deviceRoot.createProperty("CardinalDirection")
        ->linkToProxy(PropertyProxyToString<CardinalDirection_t>(shadeAttributes.cardinalDirection));
    deviceRoot.createProperty("WindProtectionClass")
        ->linkToProxy(PropertyProxyReference<int, WindProtectionClass_t>(shadeAttributes.windProtectionClass));
  }

  void removeShadeAttributes(PropertyNode& deviceRoot) override {
    if (auto cardinalDirectionNode = deviceRoot.getProperty("CardinalDirection"); cardinalDirectionNode) {
      deviceRoot.removeChild(cardinalDirectionNode);
    }
    if (auto windProtectionNode = deviceRoot.getProperty("WindProtectionClass"); windProtectionNode) {
      deviceRoot.removeChild(windProtectionNode);
    }
  }

  Apartment& m_apartment;
};

void DevicePropertyTreeImpl::publishVdcToPropertyTree(Device& device) {
  if (device.m_isVdcDevice && (device.m_pPropertyNode != NULL)) {
    PropertyNodePtr propNode = device.m_pPropertyNode->createProperty("properties");

    propNode->createProperty("DisplayId")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.displayId, false));
    propNode->createProperty("HardwareModelGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.hardwareModelGuid, false));
    propNode->createProperty("ModelUID")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.modelUID, false));
    propNode->createProperty("ModelVersion")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.modelVersion, false));
    propNode->createProperty("VendorGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.vendorGuid, false));
    propNode->createProperty("OemGuid")->linkToProxy(
        PropertyProxyReference<std::string>(device.m_vdsdSpec.oemGuid, false));
    propNode->createProperty("OemModelGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.oemModelGuid, false));
    propNode->createProperty("ConfigURL")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.configURL, false));
    propNode->createProperty("HardwareGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.hardwareGuid, false));
    propNode->createProperty("HardwareInfo")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.model, false));
    propNode->createProperty("HardwareVersion")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_vdsdSpec.hardwareVersion, false));
  }
}

void DevicePropertyTreeImpl::publishValveTypeToPropertyTree(Device& device) {
  if ((device.m_pPropertyNode != NULL)) {
    if (device.isValveDevice()) {
      PropertyNodePtr valveType = device.m_pPropertyNode->createProperty("ValveType");
      valveType->linkToProxy(
          PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getValveTypeAsString));
    }
    PropertyNodePtr activeCoolingMode = device.m_pPropertyNode->createProperty("ActiveCoolingMode");
    activeCoolingMode->linkToProxy(
        PropertyProxyMemberFunction<Device, bool, false>(device, &Device::getActiveCoolingMode));
  }
}

void DevicePropertyTreeImpl::publishAliasInZoneGroups(const Device& device) const {
  for (auto&& g : device.m_groupIds) {
    std::string gPath = "zones/zone" + intToString(device.getGroupZoneID(g)) + "/groups/group" + intToString(g) +
                        "/devices/" + dsuid2str(device.m_DSID);
    PropertyNodePtr gnode = m_apartment.getPropertyNode()->createProperty(gPath);
    if (gnode) {
      gnode->alias(device.m_pPropertyNode);
    }
  }
}

void DevicePropertyTreeImpl::republishModelFeaturesToPropertyTree(Device& device) {
  if (!device.m_pPropertyNode) {
    return;
  }

  // remove old nodes
  {
    auto&& node = device.m_pPropertyNode->getProperty("modelFeatures");
    if (node) {
      device.m_pPropertyNode->removeChild(node);
    }
  }
  // add new nodes
  auto&& modelFeaturesNode = device.m_pPropertyNode->createProperty("modelFeatures");
  for (auto&& modelFeaturePair : device.m_modelFeatures) {
    const auto& id = modelFeaturePair.first;
    if (const auto& name = ds::tryFromEnum<const char*>(id)) {
      auto&& node = modelFeaturesNode->createProperty(*name);
      node->setValue(modelFeaturePair.second);
    } else {
      DS_WARNING("Cannot publish model feature ", id);
    }
  }
}

void DevicePropertyTreeImpl::publishToPropertyTree(Device& device) {
  if (device.m_pPropertyNode != NULL) {
    // already published
    return;
  }

  if (m_apartment.getPropertyNode() == NULL) {
    DS_NOTICE("No root node set"); // unit tests, without PropertySystem
    return;
  }

  if (!device.m_visible) {
    DS_NOTICE("publishToPropertyTree", device.m_visible, device.m_DSID, "skipping");
    return;
  }

  DS_NOTICE("publishToPropertyTree", device.getDSID(), device.getName());
  device.m_pPropertyNode =
      m_apartment.getPropertyNode()->createProperty("zones/zone0/devices/" + dsuid2str(device.m_DSID));

  dsid_t dsid;
  if (dsuid_to_dsid(device.m_DSID, &dsid)) {
    device.m_pPropertyNode->createProperty("dSID")->setStringValue(dsid2str(dsid));
  } else {
    device.m_pPropertyNode->createProperty("dSID")->setStringValue("");
  }
  device.m_pPropertyNode->createProperty("DisplayID")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getDisplayID));

  device.m_pPropertyNode->createProperty("dSUID")->setStringValue(dsuid2str(device.m_DSID));
  device.m_pPropertyNode->createProperty("present")->linkToProxy(
      PropertyProxyMemberFunction<Device, bool>(device, &Device::isPresent));
  device.m_pPropertyNode->createProperty("name")->linkToProxy(
      PropertyProxyMemberFunction<Device, std::string>(device, &Device::getName, &Device::setName));
  if (!device.m_DSMeterDSIDstr.empty()) {
    device.m_pPropertyNode->createProperty("DSMeterDSID")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_DSMeterDSIDstr, false));
  } else {
    device.m_pPropertyNode->createProperty("DSMeterDSID")->setStringValue("");
  }
  device.m_pPropertyNode->createProperty("DSMeterDSUID")
      ->linkToProxy(PropertyProxyReference<std::string>(device.m_DSMeterDSUIDstr, false));

  device.m_pPropertyNode->createProperty("ZoneID")->linkToProxy(PropertyProxyReference<int>(device.m_ZoneID, false));
  device.m_pPropertyNode->createProperty("functionID")
      ->linkToProxy(PropertyProxyReference<int>(device.m_FunctionID, false));
  device.m_pPropertyNode->createProperty("revisionID")
      ->linkToProxy(PropertyProxyReference<int>(device.m_RevisionID, false));
  device.m_pPropertyNode->createProperty("productID")
      ->linkToProxy(PropertyProxyReference<int>(device.m_ProductID, false));
  device.m_pPropertyNode->createProperty("vendorID")
      ->linkToProxy(PropertyProxyReference<int>(device.m_VendorID, false));

  device.m_pPropertyNode->createProperty("HWInfo")->linkToProxy(
      PropertyProxyReference<std::string>(device.m_HWInfo, false));
  device.m_pPropertyNode->createProperty("GTIN")->linkToProxy(
      PropertyProxyReference<std::string>(device.m_GTIN, false));

  PropertyNodePtr oemNode = device.m_pPropertyNode->createProperty("productInfo");
  oemNode->createProperty("ProductState")
      ->linkToProxy(
          PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getOemProductInfoStateAsString));
  oemNode->createProperty("ProductName")
      ->linkToProxy(PropertyProxyReference<std::string>(device.m_OemProductName, false));
  oemNode->createProperty("ProductIcon")
      ->linkToProxy(PropertyProxyReference<std::string>(device.m_OemProductIcon, false));
  oemNode->createProperty("ProductURL")
      ->linkToProxy(PropertyProxyReference<std::string>(device.m_OemProductURL, false));
  oemNode->createProperty("ConfigLink")
      ->linkToProxy(PropertyProxyReference<std::string>(device.m_OemConfigLink, false));
  oemNode->createProperty("State")->linkToProxy(
      PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getOemStateAsString));
  oemNode->createProperty("EAN")->linkToProxy(
      PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getOemEanAsString));
  oemNode->createProperty("SerialNumber")
      ->linkToProxy(PropertyProxyReference<int, uint16_t>(device.m_OemSerialNumber, false));
  oemNode->createProperty("PartNumber")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(device.m_OemPartNumber, false));
  oemNode->createProperty("isIndependent")->linkToProxy(PropertyProxyReference<bool>(device.m_OemIsIndependent, false));
  oemNode->createProperty("InternetState")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getOemInetStateAsString));
  oemNode->createProperty("configurationLocked")
      ->linkToProxy(PropertyProxyReference<bool>(device.m_IsConfigLocked, false));

  device.m_pPropertyNode->createProperty("isVdcDevice")
      ->linkToProxy(PropertyProxyReference<bool>(device.m_isVdcDevice, false));
  device.m_pPropertyNode->createProperty("hasActions")
      ->linkToProxy(PropertyProxyReference<bool>(device.m_hasActions, false));
  publishVdcToPropertyTree(device);

  device.m_pPropertyNode->createProperty("lastKnownZoneID")
      ->linkToProxy(PropertyProxyReference<int>(device.m_LastKnownZoneID, false));
  device.m_pPropertyNode->createProperty("shortAddress")
      ->linkToProxy(PropertyProxyReference<int, uint16_t>(device.m_ShortAddress, false));
  device.m_pPropertyNode->createProperty("lastKnownShortAddress")
      ->linkToProxy(PropertyProxyReference<int, uint16_t>(device.m_LastKnownShortAddress, false));
  if (!device.m_LastKnownMeterDSIDstr.empty()) {
    device.m_pPropertyNode->createProperty("lastKnownMeterDSID")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_LastKnownMeterDSIDstr, false));
  } else {
    device.m_pPropertyNode->createProperty("lastKnownMeterDSID")->setStringValue("");
  }
  device.m_pPropertyNode->createProperty("lastKnownMeterDSUID")
      ->linkToProxy(PropertyProxyReference<std::string>(device.m_LastKnownMeterDSUIDstr, false));
  device.m_pPropertyNode->createProperty("firstSeen")
      ->linkToProxy(PropertyProxyMemberFunction<DateTime, std::string, false>(device.m_FirstSeen, &DateTime::toString));
  device.m_pPropertyNode->createProperty("lastDiscovered")
      ->linkToProxy(
          PropertyProxyMemberFunction<DateTime, std::string, false>(device.m_LastDiscovered, &DateTime::toString));
  device.m_pPropertyNode->createProperty("inactiveSince")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getInactiveSinceStr));
  device.m_pPropertyNode->createProperty("locked")->linkToProxy(
      PropertyProxyReference<bool>(device.m_IsLockedInDSM, false));
  device.m_pPropertyNode->createProperty("ignoreOperationLock")
      ->linkToProxy(PropertyProxyReference<bool>(device.m_IgnoreOperationLock, false));
  device.m_pPropertyNode->createProperty("outputMode")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(device.m_OutputMode, false));
  device.m_pPropertyNode->createProperty("button/id")
      ->linkToProxy(PropertyProxyReference<int>(device.m_ButtonID, false));
  device.m_pPropertyNode->createProperty("button/inputMode")
      ->linkToProxy(PropertyProxyReference<int, ButtonInputMode>(device.m_ButtonInputMode, false));
  device.m_pPropertyNode->createProperty("button/inputIndex")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(device.m_ButtonInputIndex, false));
  device.m_pPropertyNode->createProperty("button/inputCount")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(device.m_ButtonInputCount, false));
  device.m_pPropertyNode->createProperty("button/activeGroup")
      ->linkToProxy(PropertyProxyReference<int>(device.m_ButtonActiveGroup, false));
  device.m_pPropertyNode
      ->createProperty("button/groupMembership") // deprecated
      ->linkToProxy(PropertyProxyReference<int>(device.m_ButtonActiveGroup, false));
  device.m_pPropertyNode->createProperty("button/setsLocalPriority")
      ->linkToProxy(PropertyProxyReference<bool>(device.m_ButtonSetsLocalPriority));
  device.m_pPropertyNode->createProperty("button/callsPresent")
      ->linkToProxy(PropertyProxyReference<bool>(device.m_ButtonCallsPresent));
  device.m_pPropertyNode->createProperty("button/usage")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(device, &Device::getButtonUsageAsString));
  if (!device.m_pPropertyNode->getProperty("sensorEvents")) {
    PropertyNodePtr sensorNode = device.m_pPropertyNode->createProperty("sensorEvents");
  }
  PropertyNodePtr binaryInputNode = device.m_pPropertyNode->createProperty("binaryInputs");
  PropertyNodePtr sensorInputNode = device.m_pPropertyNode->createProperty("sensorInputs");
  PropertyNodePtr outputChannelNode = device.m_pPropertyNode->createProperty("outputChannels");

  device.m_pPropertyNode->createProperty("isValveType")
      ->linkToProxy(PropertyProxyMemberFunction<Device, bool>(device, &Device::isValveDevice));

  publishShadeAttributes(*device.getPropertyNode(), device.m_shadeAttributes);

  device.m_pPropertyNode->createProperty("Floor")->linkToProxy(PropertyProxyReference<int>(device.m_floor));

  device.m_pPropertyNode->createProperty("supportedBasicScenes")
      ->linkToProxy(PropertyProxyFunction<std::string>(
          [&device] { return serializeBasicScenes(device.m_supportedBasicScenes); }));

  publishValveTypeToPropertyTree(device);

  device.m_TagsNode = device.m_pPropertyNode->createProperty("tags");
  device.m_TagsNode->setFlag(PropertyNode::Archive, true);

  if (device.m_ZoneID != 0) {
    std::string basePath = "zones/zone" + intToString(device.m_ZoneID) + "/devices";
    if (device.m_pAliasNode == NULL) {
      PropertyNodePtr node = m_apartment.getPropertyNode()->getProperty(basePath + "/" + dsuid2str(device.m_DSID));
      if ((node == NULL) || ((node != NULL) && (node->size() == 0))) {
        device.m_pAliasNode = m_apartment.getPropertyNode()->createProperty(basePath + "/" + dsuid2str(device.m_DSID));
        device.m_pAliasNode->alias(device.m_pPropertyNode);
      }
    } else {
      PropertyNodePtr base = m_apartment.getPropertyNode()->getProperty(basePath);
      if (base != NULL) {
        base->addChild(device.m_pAliasNode);
      }
    }
  }

  publishAliasInZoneGroups(device);
  for (auto&& g : device.m_groupIds) {
    PropertyNodePtr gsubnode = device.m_pPropertyNode->createProperty("groups/group" + intToString(g));
    gsubnode->createProperty("id")->setIntegerValue(g);
  }

  if (device.m_DSMeterDSID != DSUID_NULL) {
    device.setDSMeter(m_apartment.getDSMeterByDSID(device.m_DSMeterDSID));
  }

  republishModelFeaturesToPropertyTree(device);
  if (m_apartment.getDss() && m_apartment.getDss()->getDeviceCacheDatabase()) {
    try {
      m_apartment.getDss()->getHal()->publishToPropertyTree(device.getDSID());
    } catch (const ds::Exception& e) {
    }
  }
} // publishToPropertyTree

void DevicePropertyTreeImpl::publishDeviceSpecification(Device& device) const {
  if (m_apartment.getPropertyNode() == NULL) {
    DS_NOTICE("No root node set"); // unit tests, without PropertySystem
    return;
  }

  if (!device.m_pPropertyNode) {
    // invisible device?
    DS_NOTICE("publishDeviceSpecification", device.m_visible, device.m_DSID, "skipping");
    return;
  }

  auto& deviceSpecification = *device.m_pPropertyNode->createProperty("DeviceSpecification");
  deviceSpecification.createProperty("ActiveGroup")
      ->linkToProxy(PropertyProxyReference<int>(device.m_ActiveGroup, false));
  deviceSpecification.createProperty("DefaultGroup")
      ->linkToProxy(PropertyProxyReference<int>(device.m_DefaultGroup, false));

  deviceSpecification.createProperty("Groups")->setStringValue(ds::str(device.m_specification.Groups));
}

void DevicePropertyTreeImpl::removeFromPropertyTree(Device& device) {
  DS_NOTICE("removeFromPropertyTree", device.getDSID(), device.getName());
  if (device.m_pPropertyNode != NULL) {
    try {
      if ((device.m_DSMeterDSID != DSUID_NULL) && (device.m_DSID != DSUID_NULL)) {
        std::string devicePath = "devices/" + dsuid2str(device.m_DSID);
        PropertyNodePtr dev =
            m_apartment.getDSMeterByDSID(device.m_DSMeterDSID)->getPropertyNode()->getProperty(devicePath);
        dev->alias(PropertyNodePtr());
        dev->getParentNode()->removeChild(dev);
      } else {
        DS_NOTICE("removeFromPropertyTree, not removed from dsm node:", device.m_DSMeterDSID, device.m_DSID);
      }
    } catch (std::runtime_error& err) {
      DS_ERROR(err.what());
    }

    if (m_apartment.getPropertyNode() != NULL) {
      for (auto&& g : device.m_groupIds) {
        std::string gPath = "zones/zone" + intToString(device.getGroupZoneID(g)) + "/groups/group" + intToString(g) +
                            "/devices/" + dsuid2str(device.m_DSID);
        PropertyNodePtr gnode = m_apartment.getPropertyNode()->getProperty(gPath);
        if (gnode != NULL) {
          DS_NOTICE("removeFromPropertyTree:", gPath);
          gnode->alias(PropertyNodePtr());
          gnode->getParentNode()->removeChild(gnode);
        } else {
          DS_NOTICE("removeFromPropertyTree: not not found", gPath);
        }

        PropertyNodePtr gsubnode = device.m_pPropertyNode->getProperty("groups/group" + intToString(g));
        if (gsubnode != NULL) {
          DS_NOTICE("removeFromPropertyTree: remove group subnode", intToString(g));
          gsubnode->getParentNode()->removeChild(gsubnode);
        }
      }
    }

    device.m_pPropertyNode->unlinkProxy(true);
    PropertyNode* parent = device.m_pPropertyNode->getParentNode();
    if (parent != NULL) {
      DS_NOTICE("removeFromPropertyTree: remove node itself");
      parent->removeChild(device.m_pPropertyNode);
    } else {
      DS_NOTICE("removeFromPropertyTree: remove node itself: no parent");
    }
    device.m_pPropertyNode.reset();
  }
  if (device.m_pAliasNode != NULL) {
    device.m_pAliasNode->alias(PropertyNodePtr());
    PropertyNode* parent = device.m_pAliasNode->getParentNode();
    if (parent != NULL) {
      DS_NOTICE("removeFromPropertyTree: remove alias node");
      parent->removeChild(device.m_pAliasNode);
    } else {
      DS_NOTICE("removeFromPropertyTree: remove alias itself: no parent");
    }
    device.m_pAliasNode.reset();
  }
}

void DevicePropertyTreeImpl::updateAKMNode(Device& device) {
  if ((device.m_pPropertyNode != NULL) && isAkmDevice(device)) {
    device.m_pPropertyNode->createProperty("AKMInputProperty")
        ->linkToProxy(PropertyProxyReference<std::string>(device.m_AKMInputProperty, false));
  }
}

void DevicePropertyTreeImpl::publishGroupNode(Device& device, int groupId) {
  // create group-n property
  auto gsubnode = device.m_pPropertyNode->createProperty(ds::str("groups/group", groupId));
  gsubnode->createProperty("id")->setIntegerValue(groupId);
}

void DevicePropertyTreeImpl::removeGroupNode(Device& device, int groupId) {
  auto gsubnode = device.m_pPropertyNode->getProperty(ds::str("groups/group", groupId));
  if (gsubnode) {
    gsubnode->getParentNode()->removeChild(gsubnode);
  }
}

std::unique_ptr<DevicePropertyTree> makeDevicePropertyTree(Apartment& apartment) {
  return std::make_unique<DevicePropertyTreeImpl>(apartment);
}

} // namespace dss
