/*
// Copyright (c) 2016 Intel Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
*/

#include <ds/json.h>
#include <ds/log.h>

#include <boost/asio.hpp>
//#include <boost/circular_buffer.hpp>
#include <iostream>

#include "config-file.h"
#include "parser.h"
#include "socket-writer.h"

static ds::LogChannel dsLogChannel("p1Main");

namespace dss::p1 {

struct Fuf {
  Fuf(boost::asio::io_service& io, std::chrono::seconds seconds,
      std::function<void(const boost::system::error_code& error_code)>&& callback)
      : m_timer(boost::asio::steady_timer(io, seconds)), m_seconds(seconds), m_callback(std::move(callback)) {}
  ~Fuf() = default;

#if 0
  Fuf(const Fuf&) = default;
  Fuf(Fuf&&) = default;
#endif
  void start() {
    m_timer.expires_at(m_timer.expiry() + m_seconds);
    m_timer.async_wait([this](const auto& ec) -> void { (*this)(ec); });
  }

  void operator()(const boost::system::error_code& error_code) {
    DS_NOTICE("retry timer", m_timer.expiry(), error_code);
    m_callback(error_code);
    start();
  }

  boost::asio::steady_timer m_timer;
  std::chrono::seconds m_seconds;
  std::function<void(const boost::system::error_code& error_code)> m_callback;
};

struct SenderQueue {
  SenderQueue(boost::asio::io_context& io, std::string socketPath) : m_io(io), m_socketPath(std::move(socketPath)) {}

  void send(MeterReading&& meterReading) {
    // we don't re-use socket client yet
    std::make_shared<dss::p1::SocketClient>(m_io, m_socketPath)->send(ds::toJson(meterReading));
  }

  boost::asio::io_context& m_io;
  std::string m_socketPath;
};

struct FrameBuffer {
  // FrameBuffer(int capacity) : m_data{boost::circular_buffer<char>(capacity)} {}
  FrameBuffer(SenderQueue& senderQueue, size_t capacity)
      : m_senderQueue(senderQueue), m_frame(std::vector<char>(capacity)) {
    m_frame.clear(); // m_frame.size() == capacity
  }

  void processFrame() {
    DS_DEBUG(m_frame.size(), m_sum, "frame:\n", ds::hideArg(std::string_view(m_frame.data(), m_frame.size())));
    if (auto reading = parseFrame(m_frame)) {
      DS_NOTICE(reading);
      m_senderQueue.send(std::move(*reading));
    };
  }

  using data_t = std::vector<char>;
  void push(data_t::const_iterator begin, data_t::const_iterator end) {
    for (auto current = begin; current != end; current++) {
      if (*current == '/') {
        processFrame();
        m_frame.clear();
        m_sum = 0;
      }
      m_frame.push_back(*current); // OOM problem if frame end is never found !!!
      m_sum++;
    }
  }

  // boost::circular_buffer<char> m_data;
  SenderQueue& m_senderQueue;
  std::vector<char> m_frame;
  int m_sum{};
};

struct SerialConnection {
  SerialConnection(boost::asio::io_context& io, SerialPortConfig&& config, FrameBuffer& frameBuffer)
      : m_io(io),
        m_config(config),
        m_readBuffer(std::vector<char>(config.readBufferCapacity)),
        m_frameBuffer(frameBuffer) {}

  void start() {
    if (m_isReading) {
      DS_NOTICE("serial console is readable");
      return;
    }

    try {
      m_serialPort = std::make_shared<boost::asio::serial_port>(m_io, m_config.serialDevice);
    } catch (std::exception& exception) {
      DS_WARNING("serialConsole", exception.what(), m_config.serialDevice);
      // std::rethrow_exception(std::current_exception());
      return;
    }

    m_serialPort->set_option(boost::asio::serial_port_base::baud_rate(m_config.baudRate));
    m_serialPort->set_option(boost::asio::serial_port_base::character_size(8));
    m_serialPort->set_option(boost::asio::serial_port_base::stop_bits(boost::asio::serial_port_base::stop_bits::one));
    m_serialPort->set_option(boost::asio::serial_port_base::parity(boost::asio::serial_port_base::parity::none));
    m_serialPort->set_option(
        boost::asio::serial_port_base::flow_control(boost::asio::serial_port_base::flow_control::none));

    doRead();
  }

  void doRead() {
    m_isReading = true;
    m_serialPort->async_read_some(boost::asio::buffer(m_readBuffer), [this](const auto& error_code, size_t size) {
      m_isReading = false; // single threaded
      if (!error_code) {
        // DS_NOTICE(size, m_readBuffer.size());
        m_frameBuffer.push(m_readBuffer.begin(), m_readBuffer.begin() + size);
        doRead();
      }
      // DS_NOTICE("doRead (frame-split)", error_code, size);
      // std::cerr << std::string_view(m_readBuffer.data(), size) << "\n\n";
    });
  }

  boost::asio::io_context& m_io;
  SerialPortConfig m_config;
  std::vector<char> m_readBuffer{};
  std::shared_ptr<boost::asio::serial_port> m_serialPort{};

  bool m_isReading{false};
  FrameBuffer& m_frameBuffer;
};

} // namespace dss::p1

int main(int argc, char** argv) {
  DS_NOTICE("foo");

  // auto context = ds::ResourceContext{contextData, io, ds::system(), std::this_thread::get_id()};

  auto io = boost::asio::io_service{};

  using dss::p1::loadConfigFile;
  auto config = loadConfigFile(ds::system());

  dss::p1::SenderQueue senderQueue(io, "/var/run/dss-metering.sock");
  // dss::p1::SenderQueue senderQueue(io, "/tmp/socket-server");
  dss::p1::FrameBuffer frameBuffer{senderQueue, config.serialPortConfig.readBufferCapacity};
  dss::p1::SerialConnection serialConnection{io, std::move(config.serialPortConfig), frameBuffer};
  serialConnection.start();

  dss::p1::Fuf reconnectTimer(io, std::chrono::seconds(10), [&serialConnection](const boost::system::error_code& ec) {
    DS_NOTICE("callback", ec);
    serialConnection.start();
  });
  reconnectTimer.start();

  io.run();

  return 0;
}
