/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "circuitrequesthandler.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "metering/metering-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance-mock.h"
#include "web/restful.h"

static ds::LogChannel dsLogChannel("webHandlerCircuitRequestHandlerTest");

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;
using namespace std::literals;

namespace {

static const char* TAGS = "[web][handler][circuitrequesthandler]";

TEST_CASE("dssWebHandlerCircuitRequest", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apartment = *pApartment;
  auto meteringMock = MeteringMock{};
  apartment.setMetering(&meteringMock);

  auto busInterface = BusInterfaceMock{};
  auto circuitRequestHandler =
      CircuitRequestHandler(apartment, &busInterface.structureModify, &busInterface.structureQuery);

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  SECTION("MeteringConfiguration") {
    std::stringstream ss;
    controller1->setCapability_HasMetering(true);

    SECTION("getMeteringConfiguration") {
      SECTION("producer") {
        REQUIRE_CALL(busInterface.structureQuery, getMeteringConfiguration(_))
            .RETURN(MeteringConfigurationSpec_t{MeteringProducerCategory::producer, 3});
        ss << circuitRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getMeteringConfiguration", ds::str("dsuid=", controllerId1)}, nullptr, nullptr);
        CHECK(ss.str() == R"({"result":{"category":"producer","scaleFactor":3},"ok":true})");
      }

      SECTION("consumer") {
        REQUIRE_CALL(busInterface.structureQuery, getMeteringConfiguration(_))
            .RETURN(MeteringConfigurationSpec_t{MeteringProducerCategory::consumer});
        ss << circuitRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/getMeteringConfiguration", ds::str("dsuid=", controllerId1)}, nullptr, nullptr);
        CHECK(ss.str() == R"({"result":{"category":"consumer"},"ok":true})");
      }
    }

    SECTION("setMeteringConfiguration") {
      SECTION("producer") {
        REQUIRE_CALL(busInterface.structureQuery, getMeteringConfiguration(_))
            .RETURN(MeteringConfigurationSpec_t{MeteringProducerCategory::consumer});

        CHECK(!controller1->getCapability_HasMeteringProducerEnabled());
        REQUIRE_CALL(busInterface.structureModify,
                     setMeteringConfiguration(Dsuid(controllerId1),
                                              MeteringConfigurationSpec_t{MeteringProducerCategory::producer, 3}));

        ss << circuitRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/setMeteringConfiguration",
                           ds::str("dsuid=", controllerId1, "&scaleFactor=3", "&category=producer")},
            nullptr, nullptr);
        CHECK(ss.str() == R"({"ok":true})");
        CHECK(controller1->getCapability_HasMeteringProducerEnabled());
      }
      SECTION("consumer") {
        REQUIRE_CALL(busInterface.structureQuery, getMeteringConfiguration(_))
            .RETURN(MeteringConfigurationSpec_t{MeteringProducerCategory::consumer});

        REQUIRE_CALL(busInterface.structureModify,
                     setMeteringConfiguration(Dsuid(controllerId1),
                                              MeteringConfigurationSpec_t{MeteringProducerCategory::consumer}));

        ss << circuitRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/setMeteringConfiguration", ds::str("dsuid=", controllerId1, "&category=consumer")},
            nullptr, nullptr);
        CHECK(ss.str() == R"({"ok":true})");
        CHECK(!controller1->getCapability_HasMeteringProducerEnabled());
      }

      SECTION("missing category argument") {
        REQUIRE_CALL(busInterface.structureQuery, getMeteringConfiguration(_))
            .RETURN(MeteringConfigurationSpec_t{MeteringProducerCategory::consumer});

        ss << circuitRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/setMeteringConfiguration", ds::str("dsuid=", controllerId1)}, nullptr, nullptr);
        CHECK(ss.str() == R"({"ok":false,"message":"Metering category value invalid"})");
        CHECK(!controller1->getCapability_HasMeteringProducerEnabled());
      }
      // scaleFactor is optional -> see consumer above
    }
  }

  SECTION("getEnergyMeterValue") {
    std::stringstream ss;

    SECTION("controller has metering") {
      controller1->setCapability_HasMetering(true);
      REQUIRE_CALL(meteringMock, getLastEnergy(_)).RETURN(TotalEnergy{.consumption = 77});
      ss << circuitRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getEnergyMeterValue", ds::str("dsuid=", controllerId1)}, nullptr, nullptr);
      CHECK(ss.str() == R"({"result":{"meterValue":77},"ok":true})");
    }
    SECTION("conroller without metering") {
      ss << circuitRequestHandler.jsonHandleRequest(
          RestfulRequest{"foo/getEnergyMeterValue", ds::str("dsuid=", controllerId1)}, nullptr, nullptr);
      CHECK(ss.str() == R"({"ok":false,"message":"Metering not supported on this device"})");
    }
  }
}

} // namespace
