/*
    Copyright (c) 2009, 2010 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include "webserver.h"

#include <arpa/inet.h>
#include <dirent.h>
#include <ds-api/rest.h>
#include <ds/http.h>
#include <ds/tracing.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/filesystem.hpp>
#include <boost/range/numeric.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/tuple/tuple.hpp>
#include <iostream>
#include <sstream>
#include "backend-vdcs.h"
#include "businterface.h"
#include "cache/apartment-proxy.h"
#include "ds-api/service.h"
#include "ds485types.h"
#include "dss.h"
#include "logger.h"
#include "model/apartment.h"
#include "model/device.h"
#include "propertysystem.h"
#include "security/security.h"
#include "session.h"
#include "sessionmanager.h"
#include "web/handler/apartmentrequesthandler.h"
#include "web/handler/circuitrequesthandler.h"
#include "web/handler/databaserequesthandler.h"
#include "web/handler/devicerequesthandler.h"
#include "web/handler/eventrequesthandler.h"
#include "web/handler/internalrequesthandler.h"
#include "web/handler/meteringrequesthandler.h"
#include "web/handler/propertyrequesthandler.h"
#include "web/handler/setrequesthandler.h"
#include "web/handler/staterequesthandler.h"
#include "web/handler/structurerequesthandler.h"
#include "web/handler/subscriptionrequesthandler.h"
#include "web/handler/systemrequesthandler.h"
#include "web/handler/zonerequesthandler.h"
#include "web/restful.h"
#include "web/webrequests.h"
#include "web/websocket/jsonapiserver.h"

static ds::LogChannel dsLogChannel("webServer");
namespace fs = boost::filesystem;

namespace dss {

  static const char* httpCodeToMessage(const int _code) {
    if(_code == 400) {
      return "Bad Request";
    } else if(_code == 401) {
      return "Unauthorized\r\nWWW-Authenticate: Basic realm=\"dSS\"";
    } else if(_code == 403) {
      return "Forbidden";
    } else if(_code == 500) {
      return "Internal Server Error";
    } else {
      return "OK";
    }
  }

  static std::string strprintf_HTTPHeader(int _code, int length, const std::string& _contentType,
                                          const std::string& _setCookie, const dsApi::HttpHeaders& headers = {}) {
    std::ostringstream sstream;
    sstream << "HTTP/1.1 " << _code << ' ' << httpCodeToMessage(_code) << "\r\n";
    sstream << "Content-Type: " << _contentType << "; charset=utf-8\r\n";
    if(!_setCookie.empty()) {
      sstream << "Set-Cookie: " << _setCookie << "; HttpOnly";
#ifndef WITH_INSECURE_COOKIE
      sstream << "; Secure";
#endif
      sstream << "\r\n";
    }
    if (ds::system().isInternalRelease()) {
      sstream << "Access-Control-Allow-Origin: *\r\n";
    }
    sstream << "Content-Length: " << intToString(length) << "\r\n";
    for (const auto& header : headers) {
      for (const auto& headerValue : header.second) {
        sstream << header.first << ": " << headerValue << "\r\n";
      }
    }
    sstream << "\r\n";
    std::string header = sstream.str();
    return header;
  }

  static void emitHTTPPacket(struct mg_connection* _connection, int _code, const std::string& _contentType,
                             const std::string& _setCookie, const std::string& content,
                             const dsApi::HttpHeaders& headers = {}) {
    std::string packet = strprintf_HTTPHeader(_code, content.length(), _contentType, _setCookie, headers);
    packet += content;
    mg_write(_connection, packet.c_str(), packet.length());
  }

  static int emitHTTPJsonPacket(mg_connection* connection, int code, const std::string& setCookie,
                                const std::string& content = {}, const dsApi::HttpHeaders& headers = {}) {
    emitHTTPPacket(connection, code, "application/json", setCookie, content, headers);
    return code;
  }

  static int emitHTTPJsonForbidden(mg_connection* connection, const std::string& setCookie) {
    DS_ERROR("Forbidden access.");
    return emitHTTPJsonPacket(connection, 403, setCookie, JSONWriter::failure("not logged in"));
  }

  static void emitHTTPTextPacket(struct mg_connection* _connection, int _code,
                                 const std::string& _setCookie,
                                 const std::string& content) {
    return emitHTTPPacket(_connection, _code, "text/html", _setCookie,
                          content);
  }

  //============================================= WebServer

  WebServer::WebSocketServers createWebSocketServers() {
    WebServer::WebSocketServers servers;
    servers.push_back(std::make_unique<websocket::SimpleServer>("/websocket", 128));
    servers.push_back(std::make_unique<websocket::JsonApiServer>("/bus", 128));
    return servers;
  }

  WebServer::WebServer(DSS* _pDSS)
      : Subsystem(_pDSS, "WebServer"), m_mgContext(0), m_TrustedPort(0), m_WebSocketServers(createWebSocketServers()) {}

  WebServer::~WebServer() {
    if (m_mgContext) {
      mg_stop(m_mgContext);
    }
  } // dtor

  void WebServer::initialize() {
    Subsystem::initialize();

    getDSS().getPropertySystem().setStringValue(getConfigPropertyBasePath() + "webroot", getDSS().getWebrootDirectory(), true, false);
    getDSS().getPropertySystem().setStringValue(getConfigPropertyBasePath() + "bindip", "0.0.0.0", true, false);
    getDSS().getPropertySystem().setStringValue(getConfigPropertyBasePath() + "listen", "8080", true, false);
    getDSS().getPropertySystem().setIntValue(getConfigPropertyBasePath() + "trustedPort", 0, true, false);
    getDSS().getPropertySystem().setIntValue(getConfigPropertyBasePath() + "announcedport", 8080, true, false);
    getDSS().getPropertySystem().setIntValue(getConfigPropertyBasePath() + "webSocketTimeoutSeconds", WEB_SOCKET_TIMEOUT_S, true, false);
    getDSS().getPropertySystem().setStringValue(getConfigPropertyBasePath() + "files/apartment.xml", getDSS().getDataDirectory() + "apartment.xml", true, false);
    getDSS().getPropertySystem().setStringValue(getConfigPropertyBasePath() + "sslcert", getDSS().getPropertySystem().getStringValue("/config/configdirectory") + "dsscert.pem" , true, false);

    std::vector<std::string> portList = splitString(DSS::getInstance()->getPropertySystem().getStringValue(getConfigPropertyBasePath() + "listen"), ',', true);
    std::string configPorts, protocol;
    std::string bindIP = DSS::getInstance()->getPropertySystem().getStringValue(getConfigPropertyBasePath() + "bindip");
    bool useSsl = false;
    for (std::string port : portList) {
      protocol = "s";   // defaults to https
      if(port.at(port.length() - 1) == 'h') {
        protocol = "";  // use plain http on this port
        port.erase(port.length() - 1);
      }
      if(port.find(":") == std::string::npos) {
        configPorts += bindIP + ":" + port;
      } else {
        configPorts += port;
      }
      configPorts += protocol + ",";
      if (protocol == "s") {
        useSsl = true;
      }
    }
    configPorts.erase(configPorts.length() - 1);
    m_TrustedPort = getDSS().getPropertySystem().getIntValue(getConfigPropertyBasePath() + "trustedPort");

    std::string configAliases = DSS::getInstance()->getPropertySystem().getStringValue(getConfigPropertyBasePath() + "webroot");

    getDSS().getPropertySystem().setIntValue(getConfigPropertyBasePath() + "sessionTimeoutMinutes", WEB_SESSION_TIMEOUT_MINUTES, true, false);
    m_SessionManager->setTimeout(getDSS().getPropertySystem().getIntValue(getConfigPropertyBasePath() + "sessionTimeoutMinutes")*60);

    getDSS().getPropertySystem().setIntValue(getConfigPropertyBasePath() + "sessionLimit", WEB_SESSION_LIMIT, true, false);
    m_SessionManager->setMaxSessionCount(getDSS().getPropertySystem().getIntValue(getConfigPropertyBasePath() + "sessionLimit"));


    getDSS().getPropertySystem().setIntValue(getConfigPropertyBasePath() + "webSocketTimeoutSeconds", WEB_SOCKET_TIMEOUT_S, true, false);

    auto wsTimeout = ds::str(
        DSS::getInstance()->getPropertySystem().getIntValue(getConfigPropertyBasePath() + "webSocketTimeoutSeconds") *
        1000); // s to ms

    publishJSLogfiles();
    instantiateHandlers();

    std::vector<const char*> mgOptions{
        "document_root",        configAliases.c_str(), //
        "listening_ports",      configPorts.c_str(),   //
        "websocket_timeout_ms", wsTimeout.c_str(),     //
        "enable_keep_alive",    "yes",                 //
    };
    std::string sslCert;
    if (useSsl) {
      sslCert = DSS::getInstance()->getPropertySystem().getStringValue(getConfigPropertyBasePath() + "sslcert");
      if (!boost::filesystem::exists(sslCert)) {
        throw std::runtime_error("Could not find SSL certificate file: " + sslCert);
      } else if (boost::filesystem::is_empty(sslCert)) {
        throw std::runtime_error("SSL certificate file is empty: " + sslCert);
      }
      mgOptions.push_back("ssl_certificate");
      mgOptions.push_back(sslCert.c_str());
    }
    mgOptions.push_back(nullptr);
    DS_INFO(mgOptions);

    struct mg_callbacks callbacks = {};
    callbacks.log_message = [](const struct mg_connection*, const char* message) -> int {
      DS_CHANNEL_ERROR(::dsLogChannel, ds::hideArg(message));
      return !0;
    };

    m_mgContext = mg_start(&callbacks, 0, mgOptions.data());
    if (m_mgContext == NULL) {
      throw std::runtime_error("Could not start webserver");
    }

    for (auto& s : m_WebSocketServers) {
      s->registerToCivetWebServer(m_mgContext, m_TrustedPort);
    }

    mg_set_request_handler(m_mgContext, "**", &httpRequestCallback, 0);

    log("Webserver started", lsInfo);
  } // initialize

  void WebServer::setSessionManager(std::shared_ptr<SessionManager> _pSessionManager) {
    m_SessionManager = _pSessionManager;
  } // setSessionManager

  void WebServer::publishJSLogfiles() {
    // pre-create the property as the gui accesses it on startup and gets upset if the node's not there
    getDSS().getPropertySystem().createProperty("/system/js/logfiles/");

    std::string logDir = getDSS().getJSLogDirectory();
    try {
      fs::directory_iterator end_iter;
      for(fs::directory_iterator dir_itr(logDir);
          dir_itr != end_iter;
          ++dir_itr)  {
        if(fs::is_regular(dir_itr->status())) {
          std::string fileName = dir_itr->path().filename().string();
          if(endsWith(dir_itr->path().filename().string(), ".log") ||
             endsWith(dir_itr->path().filename().string(), ".LOG")) {
            std::string abspath = dir_itr->path().string();
            getDSS().getPropertySystem().setStringValue("/config/subsystems/WebServer/files/" + std::string(fileName), abspath, true, false);
            getDSS().getPropertySystem().setStringValue("/system/js/logfiles/" + std::string(fileName), abspath, true, false);
          }
        }
      }
    } catch(std::exception& e) {
      throw std::runtime_error("Directory for js logfiles does not exist: '" + logDir + "'");
    }
  } // publishJSLogfiles

  void WebServer::doStart() { } // start

  const char* kHandlerApartment = "apartment";
  const char* kHandlerZone = "zone";
  const char* kHandlerDevice = "device";
  const char* kHandlerCircuit = "circuit";
  const char* kHandlerSet = "set";
  const char* kHandlerProperty = "property";
  const char* kHandlerEvent = "event";
  const char* kHandlerSystem = "system";
  const char* kHandlerStructure = "structure";
  const char* kHandlerMetering = "metering";
  const char* kHandlerSubscription = "subscription";
  const char* kHandlerDatabase = "database";
  const char* kHandlerState = "state";
  const char* kHandlerInternal = "internal";

  void WebServer::instantiateHandlers() {
    m_Handlers[kHandlerApartment] = new ApartmentRequestHandler(getDSS().getApartment(), getDSS().getModelMaintenance());
    m_Handlers[kHandlerZone] = new ZoneRequestHandler(getDSS().getApartment());
    m_Handlers[kHandlerDevice] = new DeviceRequestHandler( getDSS().getApartment());
    m_Handlers[kHandlerCircuit] = new CircuitRequestHandler(
            getDSS().getApartment(),
            getDSS().getBusInterface().getStructureModifyingBusInterface(),
            getDSS().getBusInterface().getStructureQueryBusInterface());
    m_Handlers[kHandlerSet] = new SetRequestHandler(getDSS().getApartment());
    m_Handlers[kHandlerProperty] = new PropertyRequestHandler(getDSS().getPropertySystem());
    m_Handlers[kHandlerEvent] = new EventRequestHandler(getDSS().getEventInterpreter());
    m_Handlers[kHandlerSystem] = new SystemRequestHandler(m_SessionManager);
    m_Handlers[kHandlerStructure] =
        new StructureRequestHandler(getDSS().getApartment(), getDSS().getModelMaintenance());
    m_Handlers[kHandlerMetering] =
        new MeteringRequestHandler(getDSS().getApartment(), getDSS().getApartment().getMetering());
    m_Handlers[kHandlerSubscription] = new SubscriptionRequestHandler(getDSS().getEventInterpreter());
    m_Handlers[kHandlerDatabase] = new DatabaseRequestHandler();
    m_Handlers[kHandlerState] = new StateRequestHandler(getDSS().getApartment());
    m_Handlers[kHandlerInternal] = new InternalRequestHandler(getDSS().getApartment(), getDSS().getEventInterpreter());
  } // instantiateHandlers

  /**
   * extractToken
   * http://tools.ietf.org/html/rfc6265#section-4.1.1
   */
  std::string extractToken(const char *_cookie) {
    if (!_cookie) {
      return "";
    }

    // according spec there has to be a space
    std::string cookie_s = "; " + std::string(_cookie);
    size_t start = cookie_s.find("; token=");
    if (std::string::npos == start) {
      return "";
    }
    start += strlen("; token=");

    size_t end = cookie_s.find(';', start);
    return (end == std::string::npos) ?
      cookie_s.substr(start) :
      cookie_s.substr(start, end - start);
  }

  std::string generateCookieString(const std::string& token) {
    return "token=" + token + "; path=/";
  }

  /**
   * emit 'Set-Cookie' header with same name but 'expires' in the past
   * http://tools.ietf.org/html/rfc6265#section-4.1.2
   */
  std::string generateRevokeCookieString() {
    return "token=; path=/; expires=Wed, 29 April 1970 12:00:00 GMT";
  }

  /**
   * extractAuthenticatedUser
   * http://tools.ietf.org/html/rfc2617
   */
  std::string extractAuthenticatedUser(const char *_header) {
    if (!_header) {
      return "";
    }

    // header:= Digest SP [user | qopc | etc]+ comma separated
    // user:= "username" "=" quoted-string
    static const char key[] = " username=\"";

    std::string header_s(_header);
    size_t start = header_s.find(key);
    if (std::string::npos == start) {
      return "";
    }
    start += strlen(key);
    size_t end = header_s.find("\"", start);
    if (std::string::npos == end) {
      return "";
    }
    return header_s.substr(start, end - start);
  }

  int WebServer::jsonHandler(struct mg_connection* _connection, RestfulRequest& request,
      const std::string& trustedSetCookie, std::shared_ptr<Session> _session, bool authenticated) {
    std::string result, setCookieHeader = trustedSetCookie;
    int returnCode = 500;

    request.setActiveCallback(boost::bind(&mg_connection_active, _connection));
    if (m_Handlers.find(request.getClass()) != m_Handlers.end()) {
      try {
        if (!authenticated && (request.getClass() != kHandlerSystem)) {
          return emitHTTPJsonForbidden(_connection, setCookieHeader);
        }
        WebServerResponse response =
          m_Handlers[request.getClass()]->jsonHandleRequest(request, _session, _connection);
        std::string callback = request.getParameter("callback");
        if (callback.empty()) {
          result = response.getResponse();
        } else {
          result = callback + "(" + response.getResponse() + ")";
        }
        if (response.isRevokeSessionToken()) {
          setCookieHeader = generateRevokeCookieString();
        } else if (response.isPublishSessionToken()) {
          // CAUTION: the session is new, the _session argument
          // of this function is NULL
          setCookieHeader = generateCookieString(response.getNewSessionToken());
          // NOTE, this way we might override the trusted login session
          // IGNORE, it doesn't happen and if so, we don't care either
        }
        log("JSON request returned with 200: " + result.substr(0, 120), lsInfo);
        returnCode = 200;
        emitHTTPJsonPacket(_connection, returnCode, setCookieHeader, result);
      } catch (const ds::Exception& e) {
        result = JSONWriter::failure(e);
        if (e.userMessage().empty()) {
          returnCode = 500;
          log("JSON request returned with 500: " + result.substr(0, 120), lsInfo);
        } else {
          // some clients can not parse the response body when they receive a 500 error
          // in cases when we have a user visible error message we return the error with
          // a 202 error instead
          returnCode = 202;
          log("JSON request returned with 202: " + result.substr(0, 120), lsInfo);
        }
        emitHTTPJsonPacket(_connection, returnCode, setCookieHeader, result);
      } catch(const std::exception& e) {
        result = JSONWriter::failure(e.what());
        returnCode = 500;
        log("JSON request returned with 500: " + result.substr(0, 120), lsInfo);
        emitHTTPJsonPacket(_connection, returnCode, setCookieHeader, result);
      }
    } else {
      log("Unknown function '" + request.getUrlPath() + "'", lsError);
      result = JSONWriter::failure("Call to unknown function");
      returnCode = 404;
      emitHTTPJsonPacket(_connection, returnCode, setCookieHeader, result);
    }
    return returnCode;
  } // jsonHandler

  int WebServer::iconHandler(struct mg_connection* _connection,
                             RestfulRequest &request,
                             const std::string &setCookieHeader) {
    std::string result;
    int returnCode = 0;
    try {
      if (request.getClass() != "getDeviceIcon") {
        throw std::runtime_error("unhandled function");
      }

      std::string dsuidStr = request.getParameter("dsuid");
      std::string dsidStr = request.getParameter("dsid");
      if (dsuidStr.empty() && dsidStr.empty()) {
        throw std::invalid_argument("missing device dsuid parameter");
      }

      dsuid_t deviceDSUID = dsidOrDsuid2dsuid(dsidStr, dsuidStr);
      std::shared_ptr<Device> result;
      result = getDSS().getApartment().getDeviceByDSID(deviceDSUID);
      if (result == NULL) {
        throw std::runtime_error("device with id " + dsuid2str(deviceDSUID) + " not found");
      }

      std::string icon = result->getIconPath();
      if (icon.empty()) {
        throw std::runtime_error("icon for device " + dsuid2str(deviceDSUID) + " not found");
      }

      std::string iconBasePath =
          DSS::getInstance()->getPropertySystem().getStringValue(
                            "/config/subsystems/Apartment/iconBasePath");
      if (!endsWith(iconBasePath, "/")) {
        iconBasePath += "/";
      }

      icon = iconBasePath + icon;
      log("Using icon file: " + icon);
      if (boost::filesystem::exists(icon)) {
        returnCode = 200;
        mg_send_file(_connection, icon.c_str());
      } else {
        throw std::runtime_error("icon file " + icon + " for device " + dsuid2str(deviceDSUID) + " not found");
      }
    } catch (const std::exception& e) {
      result = JSONWriter::failure(e.what());
      returnCode = 500;
      emitHTTPJsonPacket(_connection, returnCode, setCookieHeader, result);
    }

    return returnCode;
  } // iconHandler

  int WebServer::httpBrowseProperties(struct mg_connection* _connection,
                                      RestfulRequest &request,
                                      const std::string &trustedSetCookie) {
    std::string path = request.getUrlPath();
    if (path.empty()) {
      path = "/";
    }

    std::ostringstream stream;
    stream << "<html><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"><body>";
    stream << "<h1>" << path << "</h1>";
    stream << "<ul>";
    PropertyNodePtr node;
    node = DSS::getInstance()->getPropertySystem().getProperty(path);
    if(node != NULL) {
      for(int iNode = 0; iNode < node->getChildCount(); iNode++) {
        PropertyNodePtr cnode = node->getChild(iNode);

        stream << "<li>";
        if(cnode->getChildCount() != 0) {
          stream << "<a href=\"/browse" << path << "/" << cnode->getDisplayName() << "\">" << cnode->getDisplayName() << "</a>";
        } else {
          stream << cnode->getDisplayName();
        }
        stream << " : ";
        stream << getValueTypeAsString(cnode->getValueType()) << " : ";
        switch (cnode->getValueType()) {
          case vTypeBoolean:
            stream << cnode->getBoolValue();
            break;
          case vTypeInteger:
            stream << cnode->getIntegerValue();
            break;
          case vTypeNone:
            break;
          case vTypeString:
            stream << cnode->getStringValue();
            break;
          case vTypeFloating:
          case vTypeUnsignedInteger:
          default:
            stream << "unknown value";
        }
      }
    }

    stream << "</ul></body></html>";
    std::string tmp = stream.str();
    emitHTTPTextPacket(_connection, 200, trustedSetCookie, tmp);
    return 200;
  } // httpBrowseProperties

  int WebServer::notifyHandler(struct mg_connection* connection, RestfulRequest& request) {
    try {
      const auto& path = request.getUrlPath();
      const auto& notification = [&]() -> ResourceNotification {
        const auto& method = gsl::cstring_span{mg_get_request_info(connection)->request_method};
        const auto& jsonPayload = [&] {
          const auto& payload = request.query().substr(1); // RestfulRequest preppended with '&' :(
          return ds::fromJson<ds::JsonFragment>(payload);
        };
        if (method == "PUT") {
          return PutResource{jsonPayload()};
        } else if (method == "DELETE") {
          return DeleteResource{};
        }
        DS_FAIL_REQUIRE("Unsupported http method.", method);
      };
      if (boost::starts_with(path, "/vdcs")) {
        getDSS().backendVdcs().notify(path, notification());
      } else {
        DS_FAIL_REQUIRE("Unsupported notify path.", path);
      }
      emitHTTPTextPacket(connection, 200, {}, "{}\n");
      return 1;
    } catch (const std::exception& e) {
      DS_WARNING(e);
      emitHTTPTextPacket(connection, 500, {}, JSONWriter::failure(e));
      return 1;
    }
  }

  int WebServer::serviceHandler(struct mg_connection* connection, std::string&& path, const std::string& setCookie,
                                std::string toplevel) {
    auto tracer = toplevel == "/api" ? ds::ThreadLocalTracer("api") : ds::ThreadLocalTracer("internalApi");
    auto result = [&] {
      try {
        auto info = mg_get_request_info(connection);
        dsApi::HttpArgs args{.method = ds::requireEnum<dsApi::HttpMethod>(info->request_method),
                             .content = {},
                             .query = (info->query_string ? info->query_string : "")};
        std::array<char, 256 * 1024> buffer;
        int count = mg_read(connection, buffer.data(), buffer.size() - 1);
        if (count > 0) {
          buffer[count] = 0;
          args.content = ds::fromJson<ds::JsonFragment>(buffer.data());
        }
        if (toplevel == "/api") {
          DS_INFO("Invoke api", path, args);
          return getDSS().api().invoke(dsApi::HttpRequest{.path = std::move(path), .httpArgs = std::move(args)});
        } else {
          DS_INFO("Invoke internal", path, args);
          return getDSS().internalApi().invoke(dsApi::HttpRequest{.path = std::move(path), .httpArgs = std::move(args)});
        }
      } catch (const std::exception& e) {
        DS_WARNING(e);
        return dsApi::errorResult(e);
      }
    }();
    const auto trace = tracer.consume();
    result.headers["DS-Tracing"].push_back(ds::toJson(trace));
    DS_INFO_LEAVE(result);
    emitHTTPJsonPacket(connection, result.status, setCookie, result.content ? ds::toJson(result.content) : "",
                       result.headers);
    return 1;
  }

  RestfulRequest WebServer::extractRequest (struct mg_connection* _connection,
                                            std::string _sublevel,
                                            const struct mg_request_info *_info) {
    if (strcmp(_info->request_method, "POST") == 0 || strcmp(_info->request_method, "PUT") == 0) {
      static const size_t size = 64 * 1024; // 64k
      char post_data[size];
      int post_data_len = mg_read(_connection, post_data, sizeof(post_data));
      std::string requestData;
      if (post_data_len > 0) {
        requestData.assign(post_data, post_data_len);
        return RestfulRequest(_sublevel, requestData);
      }
    }
    // #11669: current apps send POST but actually mean GET so the post_data is empty
    // and the query string is filled. Fall back to the GET case.
    return RestfulRequest(_sublevel, _info->query_string ?: "");
  } // extractRequest

  int WebServer::httpRequestCallback(struct mg_connection* _connection, void*) {
    auto& dsLogChannel = ::dsLogChannel;
    try {
      bool authenticated = false;
      std::string trustedLoginCookie;
      const struct mg_request_info* _info = mg_get_request_info(_connection);
      const auto& cookie = mg_get_header(_connection, "Cookie");

      DS_INFO_ENTER(_info->request_method, _info->local_uri, _info->query_string, _info->remote_addr, cookie);
      if (!_info->local_uri || strlen(_info->remote_addr) < 1) {
        return 0;
      }

      std::string uri_path(_info->local_uri);
      size_t offset = uri_path.find('/', 1);
      if (std::string::npos == offset) {
        return 0;
      }
      std::string toplevel = uri_path.substr(0, offset);
      std::string sublevel = uri_path.substr(offset);

      WebServer& self = DSS::getInstance()->getWebServer();
      self.m_SessionManager->getSecurity()->signOff(); // protect log call below

      const auto& httpPath = uri_path;
      DS_CONTEXT(httpPath);

      std::shared_ptr<Session> session;
      std::string token = extractToken(cookie);
      if (token.empty()) {
        // ... used when not going through lighttpd
        token = ds::tryHttpQueryArg(_info->query_string, "token").value_or(std::string{});
      }
      if (!token.empty()) {
        session = self.m_SessionManager->getSession(token);
      }

      // lighttpd is proxying to our trusted port. We trust that it
      // properly authenticated the user and just filter the HTTP
      // Authorization header for the username
      if (((session == NULL) || (session->getUser() == NULL)) && (_info->local_port == self.m_TrustedPort)) {
        DS_DEBUG("Request over trusted port.");
        const char* auth_header = mg_get_header(_connection, "Authorization");
        std::string userName = extractAuthenticatedUser(auth_header);
        authenticated = true;

        if (!userName.empty()) {
          self.log("Logging-in from a trusted port as '" + userName + "'");
          if (!self.m_SessionManager->getSecurity()->impersonate(userName)) {
            self.log("Unknown user", lsInfo);
            return 0;
          }
          if ((session == NULL) && (toplevel != "/icons")) {
            std::string newToken = self.m_SessionManager->registerSession();
            if (newToken.empty()) {
              self.log("Session limit reached: " + httpPath, lsError);
              return 0;
            }
            self.log("Registered new JSON session for trusted port (" + newToken + ")");
            session = self.m_SessionManager->getSession(newToken);
            session->inheritUserFromSecurity();

            // this is the first call, piggy back the set-cookie header
            trustedLoginCookie = generateCookieString(session->getID());
          }
        }
      }

      if (session != NULL) {
        session->touch();
        self.m_SessionManager->getSecurity()->authenticate(session);
        authenticated = true;
      }

      if (toplevel == "/api" || toplevel == "/internal") {
        const char* authorization = mg_get_header(_connection, "Authorization");
        if (authorization && !authenticated) {
          authenticated = self.m_SessionManager->getSecurity()->isAuthorizationHeaderValid(authorization);
        }
        if (!authenticated) {
          DS_ERROR("Forbidden access.");
          return emitHTTPJsonPacket(_connection, 401, trustedLoginCookie);
        }
        return self.serviceHandler(_connection, std::move(sublevel), trustedLoginCookie, toplevel);
      }

      RestfulRequest request = extractRequest(_connection, sublevel, _info);
      if (toplevel == "/browse") {
        return self.httpBrowseProperties(_connection, request, trustedLoginCookie);
      } else if (toplevel == "/json") {
        return self.jsonHandler(_connection, request, trustedLoginCookie, session, authenticated);
      } else if (toplevel == "/icons") {
        return self.iconHandler(_connection, request, trustedLoginCookie);
      } else if (toplevel == "/notify") {
        if (!authenticated) {
          return emitHTTPJsonForbidden(_connection, trustedLoginCookie);
        }
        return self.notifyHandler(_connection, request);
      }
    } catch (const std::exception& e) {
      DS_ERROR("httpRequestCallback failed.", e);
    }

    return 0; // civetweb handles the request on its own
  }

  void WebServer::sendToWebSockets(const Event& e) {
    for (auto& s : m_WebSocketServers) {
      s->sendToWebSockets(e);
    }
  }
}
