/*
    Copyright (c) 2017 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "web/websocket/identifydevice.h"

#include <device-info.h>

#include <boost/locale/generator.hpp>

#include "ds/locale/ds-locale.h"
#include "ds/system.h"
#include "event.h"
#include "model/apartment.h"
#include "model/devicereference.h"
#include "model/dsmeter.h"
#include "model/state.h"
#include "web/webrequests.h"

namespace dss {

namespace websocket {

static ds::LogChannel dsLogChannel("webWebsocketIdentifydevice");

IdentifyDevice::IdentifyDevice() {
}

bool IdentifyDevice::checkIfForMe(const Event& e) const {
  if(e.isContextSet()) {
    return e.getContext() == "identifyDevice" && e.getName() == EventName::IdentifyDevice;
  }
  return false;
}

ds::JsonFragment IdentifyDevice::toJson(const Event& e) const {
  if (checkIfForMe(e)) {
    IdentifyDeviceEvent eventMsg;
    std::shared_ptr<const DeviceReference> device = e.getRaisedAtDevice();

    eventMsg.description = buildDescription(device, e);

    return ds::toJsonFragment(eventMsg);
  }
  return ds::JsonFragment();
}

std::string IdentifyDevice::buildDescription(
    std::shared_ptr<const DeviceReference> device, const Event& e) const {
  std::string desc = "";
  int buttonIndex = device->getDevice()->getButtonInputIndex();
  int buttonCount = device->getDevice()->getButtonInputCount();
  int buttonId = device->getDevice()->getButtonID();

  std::string meterName = device->getDevice()->getApartment().getDSMeterByDSID(device->getDevice()->getDSMeterDSID())->getName();

  if (device->getDevice()->isVdcDevice()) {
    auto&& dss = *DSS::getInstance();
    std::string gtin = device->getDevice()->getOemEanAsString();
    auto&& propSys = DSS::getInstance()->getPropertySystem();
    auto locale = dss.localeGenerator().generate(ds::locale::sanitizeLanguage(propSys.getStringValue("/system/language/locale")));
    auto prefer = DeviceInfo::Prefer();
    JSONWriter json;
    auto db = vdcDb::make(dss);
    auto query = dss.getApartment().getBusInterface()->getStructureQueryBusInterface();
    DeviceInfo vdcInfo(prefer, *query, *db, device->getDevice().get(), gtin, locale, json, ds::system());

    std::vector<std::pair<std::string, std::string>> vdcData = {
      {"dSMID", ds::str(device->getDevice()->getDSMeterDSID())},
      {"Room ID", ds::str(device->getDevice()->getZoneID())},
      {"Controller", meterName},
      {"dSUID", dsuid2str(device->getDevice()->getDSID())},
      {"Device ID", device->getDevice()->getDisplayID()},
      {"HW info", device->getDevice()->getHWInfo()},
      {"----------", ""}};

    auto states = device->getDevice()->getStates();
    for (auto it=states.begin(); it != states.end(); it++) { // states
      if(it->second->getPrevState() != it->second->getState()) {
        vdcData.push_back(make_pair(vdcInfo.getTranslatedStateName(ds::str(it->first)),
                                    vdcInfo.getTranslatedStateValue(ds::str(it->first), ds::str(it->second->toString()))));
      }
    }

    for (int i = 0; i < device->getDevice()->getSensorCount(); i++) { // sensor values
      auto sensorType = static_cast<int>(device->getDevice()->getSensor(i).m_sensorType);
      auto sensorId = device->getDevice()->getSensor(i).m_sensorId;
      auto oldSensorValue = device->getDevice()->getSensor(i).m_sensorPrevValueFloat;
      auto sensorValue = device->getDevice()->getDeviceSensorValueEx(i).value;
      if(oldSensorValue != sensorValue) {
        vdcData.push_back(make_pair(ds::str("sensortype"), ds::str(sensorType)));
        vdcData.push_back(make_pair(vdcInfo.getTranslatedSensorName(sensorId), ds::str(sensorValue)));
      }
    }

    std::vector<std::shared_ptr<DeviceBinaryInput>> binaryInputs = device->getDevice()->getBinaryInputs();
    for (uint i = 0; i < binaryInputs.size(); i++) { // binary input states
      auto state = statusFieldValueName(static_cast<StatusFieldValue>(binaryInputs[i]->getState().getState() - 1));
      auto prevState = statusFieldValueName(static_cast<StatusFieldValue>(binaryInputs[i]->getState().getPrevState() - 1));
      if (state != std::nullopt && (state != prevState)) {
        vdcData.push_back(make_pair(ds::str("input", i + 1), ds::str(state)));
      }
    }

    if (e.hasPropertySet("clickType") && e.hasPropertySet("keyType")) {
      vdcData.push_back({"Pushbutton", displayButton(strToInt(e.getPropertyByName("keyType")), buttonIndex, buttonCount)});
    }

    desc += ds::toJson(vdcData);

  } else {
    dsid_t dsid;
    dsuid_to_dsid(device->getDevice()->getDSID(), &dsid);
    std::vector<std::pair<std::string, std::string>> deviceData = {
      {"dSMID", ds::str(device->getDevice()->getDSMeterDSID())},
      {"Room ID", ds::str(device->getDevice()->getZoneID())},
      {"Meter", meterName},
      {"dSUID", dsuid2str(device->getDevice()->getDSID())},
      {"dSID", dsid2str(dsid)},
      {"Device ID", device->getDevice()->getDisplayID()},
      {"HW info", device->getDevice()->getHWInfo()},
      {"Pushbutton", displayButton(buttonId, buttonIndex, buttonCount)}};

    desc += ds::toJson(deviceData);
  }
  return desc;
}

std::string IdentifyDevice::displayButton(int id, int index, int count) const {
  std::string ret;
  int displayLine = 0;

  enum {
    ROOM_BUTTON_1_WAY = 0,
    ROOM_BUTTON_2_WAY_DOWN = 1,
    ROOM_BUTTON_2_WAY_UP = 2,
    LOCAL_BUTTON_1_WAY = 4,
    LOCAL_BUTTON_2_WAY_DOWN = 5,
    LOCAL_BUTTON_2_WAY_UP = 6
  };

  switch (id) {
    default:
    case ROOM_BUTTON_1_WAY:
    case LOCAL_BUTTON_1_WAY:
      displayLine = 1;
      break;
    case ROOM_BUTTON_2_WAY_DOWN:
    case LOCAL_BUTTON_2_WAY_DOWN:
      ret = ds::str(dsMessage("identifyDevice", "2-Way down"));
      break;
    case ROOM_BUTTON_2_WAY_UP:
    case LOCAL_BUTTON_2_WAY_UP:
      ret = ds::str(dsMessage("identifyDevice", "2-Way up"));
      break;
  }

  if ((count > 1) && (displayLine > 0)) {
    if (!ret.empty()) {
      ret = ds::str(dsMessage("identifyDevice", "Line"), " ", index + 1, ", ", ret);
    } else {
      ret = ds::str(dsMessage("identifyDevice", "Line"), " ", index + 1);
    }
  }
  return ret;
}

} // namespace websocket

} // namespace dss
