/*
    Copyright (c) 2009, 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN

#include <boost/filesystem.hpp>
#include <boost/test/unit_test.hpp>

#include "config.h"
#include "event.h"
#include "eventinterpreterjsplugins.h"
#include "eventinterpreterplugins.h"

using namespace dss;

BOOST_AUTO_TEST_SUITE(JsEvents)

class NonRunningFixture {
public:
  NonRunningFixture() : m_tz("Europe/Zurich") {
    m_pEventInterpreter.reset(new EventInterpreter(NULL));
    m_pQueue.reset(new EventQueue(m_pEventInterpreter.get(), 2));
    m_pRunner.reset(new EventRunner(m_pEventInterpreter.get()));
    m_propSys.reset(new PropertySystem());
    m_pEventInterpreter->setEventQueue(m_pQueue.get());
    m_pEventInterpreter->setEventRunner(m_pRunner.get());
    m_pQueue->setEventRunner(m_pRunner.get());
    m_pRunner->setEventQueue(m_pQueue.get());
    m_pEventInterpreter->initialize();
  }

protected:
  std::shared_ptr<EventQueue> m_pQueue;
  std::shared_ptr<EventRunner> m_pRunner;
  std::shared_ptr<EventInterpreter> m_pEventInterpreter;
  std::shared_ptr<PropertySystem> m_propSys;
  TZSwitcher m_tz;
};

BOOST_AUTO_TEST_CASE(testEventHandlerJavascriptDoesntLeakExceptionsWithNonexistingFile) {
  EventInterpreter interpreter(NULL);
  interpreter.initialize();
  EventInterpreterPluginJavascript* plugin = new EventInterpreterPluginJavascript(&interpreter);
  interpreter.addPlugin(plugin);

  std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
  opts->setParameter("filename1", "idontexistandneverwill.js");
  EventSubscription subscription("testEvent", "javascript", interpreter, opts);
  Event evt("testEvent");
  plugin->handleEvent(evt, subscription);
} // testEventHandlerJavascriptDoesntLeakExceptionsWithNonexistingFile

BOOST_FIXTURE_TEST_CASE(testRemovingSubscription, NonRunningFixture) {
  EventInterpreterPlugin* plugin = new EventInterpreterPluginRaiseEvent(m_pEventInterpreter.get());
  m_pEventInterpreter->addPlugin(plugin);

  std::shared_ptr<Event> pEvent = std::make_shared<Event>("my_event");

  std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
  opts->setParameter("event_name", "event1");
  auto subscription = std::make_shared<EventSubscription>("my_event", "raise_event", *m_pEventInterpreter, opts);
  m_pEventInterpreter->subscribe(subscription);

  m_pQueue->pushEvent(pEvent);

  m_pEventInterpreter->executePendingEvent();
  m_pEventInterpreter->executePendingEvent();

  BOOST_CHECK_EQUAL(m_pEventInterpreter->getEventsProcessed(), 2);

  m_pEventInterpreter->unsubscribe(subscription->getID());

  m_pQueue->pushEvent(pEvent);

  m_pEventInterpreter->executePendingEvent();
  m_pEventInterpreter->executePendingEvent();

  BOOST_CHECK_EQUAL(m_pEventInterpreter->getEventsProcessed(), 3);
} // testRemovingSubscription

// TODO: in case a script does not run through, the exception is caught in the
// event interpreter and our test can not evaluate the result; figure out a way
// to test
BOOST_AUTO_TEST_CASE(testMultipleScriptFiles) {
  std::string fileName1 = getTempDir() + "s1.js";
  std::string fileName2 = getTempDir() + "s2.js";

  std::ofstream ofs1(fileName1.c_str());
  ofs1 << "function niceprint(text) { print('[NICE]: ' + text); }\n";
  ofs1.close();

  std::ofstream ofs2(fileName2.c_str());
  ofs2 << "niceprint('kraaah!');\n";
  ofs2.close();

  EventInterpreter interpreter(NULL);
  interpreter.initialize();
  EventInterpreterPluginJavascript* plugin = new EventInterpreterPluginJavascript(&interpreter);
  interpreter.addPlugin(plugin);

  std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
  opts->setParameter("filename1", fileName1);
  opts->setParameter("filename2", fileName2);
  EventSubscription subscription("testEvent", "javascript", interpreter, opts);
  Event evt("testEvent");
  plugin->handleEvent(evt, subscription);

  boost::filesystem::remove(fileName1);
  boost::filesystem::remove(fileName2);
} // testMultipleScriptFiles

// TODO: we need assertions in javascript sometime to make this test useful
BOOST_AUTO_TEST_CASE(testEventSourceGetsPassed) {
  std::string fileName1 = getTempDir() + "testeventsourcegetspassed.js";

  std::ofstream ofs1(fileName1.c_str());
  ofs1 << "print('isZone:      ' + raisedEvent.source.isZone);\n";
  ofs1 << "print('isDevice:    ' + raisedEvent.source.isDevice);\n";
  ofs1 << "print('isApartment: ' + raisedEvent.source.isApartment);\n";
  ofs1.close();

  EventInterpreter interpreter(NULL);
  interpreter.initialize();
  EventInterpreterPluginJavascript* plugin = new EventInterpreterPluginJavascript(&interpreter);
  interpreter.addPlugin(plugin);

  std::shared_ptr<SubscriptionOptions> opts = std::make_shared<SubscriptionOptions>();
  opts->setParameter("filename1", fileName1);
  EventSubscription subscription("testEvent", "javascript", interpreter, opts);
  Event evt("testEvent");
  plugin->handleEvent(evt, subscription);

  boost::filesystem::remove(fileName1);
}

BOOST_AUTO_TEST_SUITE_END()
